#include "entityload.h"
#include "iotools.h"
#include "mathem.h"
#include <math.h>
#include <string.h>
#include <ctype.h>



/**
  This constructor initializes data to the zero values
*/
entityload::entityload()
{
  ncomp = 0L;
  nlc   = 0L;
  ft    = stat;
  lgcs  = 0L;
  val   = NULL;
  func  = NULL;
  tfunc = NULL;
}



/**
  This destructor deallocates used memory
*/
entityload::~entityload()
{
  long i;
  if (func)
  {
    for (i=0; i <ncomp; i++)
    {
      delete func[i];
      delete [] tfunc[i];
    }
    delete [] func;
    delete [] tfunc;
  }
  delete [] val;
}



/**
  This function reads data about entity load from the text file given by the parameter in
  Parameters :
  @param in - pointer to the opned text file
  @param lc - total number of load cases
  @param slc - pointer to array with number of subloadcases for particular load cases, default value is NULL

  Returns :
  @retval 0 - on succes
  @retval 1 - unsupported function type is required
  @retval 2 - invalid type of load coordinate system
  @retval 3 - function has insufficient number of parameters
  @retval 4 - wrong name of function parameters
*/
long entityload::read(XFILE *in, long lc, long *slc=NULL)
{
  Parser par;
  long var_x, var_y, var_z;

  xfscanf(in, "%k%ld", "lc_id", &nlc);
  if ((nlc < 1) || (nlc > lc))
    return(1);
  if (slc)
  {
    xfscanf(in, "%k%ld", "slc_id", &nslc);
    if ((nslc < 1) || (nslc > slc[nlc-1]))
      return(1);
  }
  long i;
  xfscanf(in, "%k%ld%k%m", "ncomp", &ncomp, "func_type", &generalfunct_kwdset, &ft);
  xfscanf(in, "%k%ld", "coord_sys", &lgcs);
  if ((lgcs < 1) || (lgcs > 2))
    return 2;
  switch (ft)
  {
    case stat:
      val = new double [ncomp];
      memset (val, 0, sizeof(*val)*ncomp);
      xfscanf(in, "%k", "load_comp");
      for (i=0; i<ncomp; i++)
        xfscanf(in, "%le", val+i);
      break;
    case pars:
      func  = new Equation*[ncomp];      
      tfunc = new char*[ncomp];
      memset (func, 0, sizeof(*func)*ncomp);
      memset (tfunc, 0, sizeof(*tfunc)*ncomp);
      for (i=0; i<ncomp; i++)
      {
        tfunc[i] = new char[1001];
        xfscanf(in, "%1000s", tfunc[i]);
        func[i] = par.TextToTree(tfunc[i]);
        // check number of variables
        if (func[i]->Variables.count() != 3)
          return 3;
        var2coord(func[i], var_x, var_y, var_z);
        if ((var_x < 0) || (var_y < 0) || (var_z < 0))
          return 4;
      }
      break;
    default:
      return 5;
  }
  return(0);
}



/**
  The function returns indeces of 'x' parameter, 'y' parameter and 'z' parameter
  of the function defined in eq. Performed comparison is case insesitiv.

  Parameters:
  @param eq - pointer to the parsed function
  @param var_x - index of x-coordinate (used by eq->Variables.at function)
  @param var_y - index of y-coordinate (used by eq->Variables.at function)
  @param var_z - index of z-coordinate (used by eq->Variables.at function)
  
  Returns:
    Indeces are returned via parameters var_x, var_y, var_z.
    Negative value of the index is returned in case that it is not found.
*/
void entityload::var2coord(Equation *eq, long &var_x, long &var_y, long &var_z)
{
  variable *var;

  var_x = var_y = var_z = -1;

  var = eq->Variables.at(0);
  if (tolower(var->Name[0]) == 'x')
    var_x = 0;
  else
  {
    if (tolower(var->Name[0]) == 'y')
      var_y = 0;
    else
    {
      if (tolower(var->Name[0]) == 'z')
        var_z = 0;
    }
  }

  var = eq->Variables.at(1);
  if (tolower(var->Name[0]) == 'x')
    var_x = 1;
  else
  {
    if (tolower(var->Name[0]) == 'y')
      var_y = 1;
    else
    {
      if (tolower(var->Name[0]) == 'z')
        var_z = 1;
    }
  }

  var = eq->Variables.at(2);
  if (tolower(var->Name[0]) == 'x')
    var_x = 2;
  else
  {
    if (tolower(var->Name[0]) == 'y')
      var_y = 2;
    else
    {
      if (tolower(var->Name[0]) == 'z')
        var_z = 2;
    }
  }
}   



/**
  This function approximates values on the loaded entity to the element node which
  lay on it. It is used linear approximation.
  Parameters :
  @param tn - node where the load should be approximated
  @param nv - array with approximated nodal values

  Returns :
  @retval 0 - on succes
  @retval 1 - in case of unknown type of load function
*/
long entityload::getval(snode &tn, double *nv)
{
  long i, var_x, var_y, var_z;
  variable *var;
    
  switch (ft)
  {
    case stat:
      for (i=0; i<ncomp; i++)
        nv[i] = val[i];
      break;
    case pars:
      for (i=0; i<ncomp; i++)
      {
        var2coord(func[i], var_x, var_y, var_z);
        // setting of x coordinate
        var = func[i]->Variables.at(var_x);
        var->Value = tn.x;
        // setting of y coordinate
        var = func[i]->Variables.at(var_y);
        var->Value = tn.y;
        // setting of z coordinate
        var = func[i]->Variables.at(var_z);
        var->Value = tn.z;
        nv[i] = func[i]->Evaluate();
      }
      break;
    default:
      return 1;
  }
  return(0);
}



/**
  The function creates new object of the loadel type 
  which contains edge loads generated for given element and 
  its edges with given property.

  Paremeters:
    @param el - element whose edge load will be generated
    @param prop - edge property id of the element el where the edge load will be applied
    @param nodes - array with node structures
  Returns:
    @retval The function returns pointer to the new allocated object of the loadel type.

  Created by Tomas Koudelka, 06.2009
*/
loadel *entityload::edge2loadel(selement &el, long prop, snode *nodes)
{
  long i, j, k, id, tnne;
  loadel *ret=new loadel;
  double *nv;

  for (i=0, tnne=0; i<el.ned; i++)
    tnne += el.nned[i];
  ret->tel  = edge;
  ret->ned  = el.ned;
  ret->nnve = tnne*ncomp;
  ret->nlc  = nlc;
  ret->le = new long[ret->ned];
  memset(ret->le, 0, sizeof(*ret->le)*ret->ned);
  ret->nodvale = new double[ret->nnve];
  memset(ret->nodvale, 0, sizeof(*ret->nodvale)*ret->nnve);
  nv = new double[ncomp];  
  id = 0;

  for(i=0; i<el.ned; i++)
  {
    if (el.propedg[i] == prop)
    {
      ret->le[i] = lgcs;
      for(j=0; j<el.nned[i]; j++)
      {
        if (getval(nodes[el.nodes[el.edgenod[i][j]]], nv))
        {
          delete ret;
          delete nv;
          return NULL;
        }
        
        for(k=0; k<ncomp; k++)
        {
          ret->nodvale[id] += nv[k];
          id++;
        }
      }
    }
  }
  delete [] nv;
  return ret;
}



/**
  The function creates new object of the loadel type 
  which contains surface loads generated for given element and 
  its surfaces with given property.

  Paremeters:
    @param el - element whose surface load will be generated
    @param prop - surface property id of the element el where the surface load will be applied
    @param nodes - array with node structures
  Returns:
    @retval The function returns pointer to the new allocated object of the loadel type.

  Created by Tomas Koudelka, 06.2009
*/
loadel *entityload::surface2loadel(selement &el, long prop, snode *nodes)
{
  long i, j, k, id, tnns;
  loadel *ret=new loadel;
  double *nv;

  for (i=0, tnns=0; i<el.nsurf; i++)
    tnns += el.nnsurf[i];
  ret->tel    = surface;
  ret->nsurf  = el.nsurf;
  ret->nnvs   = tnns*ncomp;
  ret->nlc    = nlc;
  ret->ls = new long[ret->nsurf];
  memset(ret->ls, 0, sizeof(*ret->ls)*ret->nsurf);
  ret->nodvals = new double[ret->nnvs];
  memset(ret->nodvals, 0, sizeof(*ret->nodvals)*ret->nnvs);
  nv = new double[ncomp];  
  id = 0;

  for(i=0; i<el.nsurf; i++)
  {
    if (el.propsurf[i] == prop)
    {
      ret->ls[i] = lgcs;
      for(j=0; j<el.nnsurf[i]; j++)
      {
        if (getval(nodes[el.nodes[el.surfnod[i][j]]], nv))
        {
          delete ret;
          delete nv;
          return NULL;
        }
        
        for(k=0; k<ncomp; k++)
        {
          ret->nodvals[id] += nv[k];
          id++;
        }
      }
    }
  }
  delete [] nv;
  return ret;
}



/**
  The function creates new object of the loadel type 
  which contains volume loads generated for given element and 
  its volume with given property.

  Paremeters:
    @param el - element whose volume load will be generated
    @param prop - surface property id of the element el where the volume load will be applied
    @param nodes - array with node structures
  Returns:
    @retval The function returns pointer to the new allocated object of the loadel type.

  Created by Tomas Koudelka, 06.2009
*/
loadel *entityload::vol2loadel(selement &el, long prop, snode *nodes)
{
  long i, k, id;
  loadel *ret=new loadel;
  double *nv;

  nv = new double[ncomp];  
  if (el.prop == prop)
  {
    ret->nnvv = el.nne*ncomp;
    ret->tel    = volume;
    ret->nlc    = nlc;
    ret->nodvalv = new double[ret->nnvv];
    memset(ret->nodvalv, 0, sizeof(*ret->nodvalv)*ret->nnvv);
     
    id = 0;
    for (i=0; i<el.nne; i++)       
    {
      if (getval(nodes[el.nodes[i]], nv))
      {
        delete ret;
        delete nv;
        return NULL;
      }
      for(k=0; k<ncomp; k++)
      {
        ret->nodvalv[id] += nv[k];
        id++;
      }
    }
  }
  delete [] nv;
  return ret;
}
