#include <math.h>
#include "dloadcase.h"
#include "loadcase.h"
#include "global.h"
#include "gtopology.h"
#include "matrix.h"
#include "vector.h"
#include "globmat.h"
#include "dloadn.h"
#include "dloadpd.h"
#include "element.h"
#include "node.h"
#include "elemswitch.h"



dloadcase::dloadcase (void)
{
  nslc=0;
  nln = 0;  nle = 0;  npd = 0;
  lon = NULL;  pd = NULL;
  slc = NULL;  gf=NULL;
}

dloadcase::~dloadcase (void)
{
  delete [] lon;  
  delete [] pd;
  delete [] slc;  
  delete [] gf;
}

/**
   function reads load case characteristics
   
   @param in - input stream
   
   24.7.2001
*/
void dloadcase::read (XFILE *in)
{
  long i;
  
  //  type of time-dependent load
  xfscanf (in,"%k%m","type_of_time_load",&dynload_kwdset,(int*)&tdl);
  
  switch (tdl){
  case timeindload:{
    //  time-independent load case multiplied by a time function

    //  number of subload cases
    xfscanf (in,"%ld",&nslc);
    
    slc = new loadcase [nslc];
    gf = new gfunct [nslc];
    
    for (i=0;i<nslc;i++){
      //  subload cases
      slc[i].read (in);
      //  time function
      gf[i].read (in);
    }
    
    break;
  }
  case seismicload:{
    //  seismic load
    stool.read (in);
    break;
  }
  case responsespectrum:{
    //  
    stool.read (in);
    break;
  }
  case timedepload:{
    //  time-dependent load case
    //  all nodes, elements, prescribed displacements and other
    //  quantities are equiped with their own time functions
    
    //  loaded nodes
    xfscanf (in,"%ld",&nln);
    lon = new dloadn [nln];
    for (i=0;i<nln;i++){
      lon[i].read (in);
    }
    
    
    /*  
    //  loaded elements
    fscanf (in,"%ld",&nle);
    loe = new dloadel [nle];
    for (i=0;i<nle;i++){
    loe[i].read (in);
    }*/
    
    
    
    //  prescribed displacements
    xfscanf (in,"%ld",&npd);
    pd = new dloadpd [npd];
    for (i=0;i<npd;i++){
      pd[i].read (in);
    }
    
    break;
  }
    
  default:{
    fprintf (stderr,"\n\n unknown type of load case is required in function read (file %s, line %d).\n",__FILE__,__LINE__);
  }
  }
  
}



/**
   Function assembles vector of right hand side.
   It is created by prescribed displacements, forces, moments, eigenstrains, etc.   
   In case that the flv is not NULL, %vector of load caused by forces only is stored in flv. 
   
   @param lcid - load case id
   @param rhs  - right hand side vector
   @param flv  - array of load %vector caused by forces only
   @param t    - actual time

   26.10.2001
   modified by TKo 7.2008
*/
void dloadcase::assemble (long lcid,double *rhs,double *flv,long n,double t)
{
  long i,j,k,ndofn,ndofe,*cn;
  vector r,f;
  matrix sm;
  double a,*aux,*auxf;
  
  //  arrays rhs and flv have to be cleaned in the caller (above) function
  
  switch (tdl){
  case timeindload:{
    //  load is defined by time independent load case, where all components
    //  are multiplied by scale factor which depends on time

    aux  = new double [n];
    auxf = new double [n];
    
    for (i=0;i<nslc;i++){
      nullv (aux,n);
      nullv (auxf,n);
      //  scale factor
      a=gf[i].getval(t);
      
      //  assembling of vector of contributions
      slc[i].assemble (0,aux,auxf,a);
      //  updating of rhs array
      for (j=0;j<n;j++){
	rhs[j]+=aux[j];
      }
      // assembling of load vector caused by forces only    
      if (flv)
      {
        for (j=0;j<n;j++)
          flv[j]+=auxf[j];
      }
    }
    
    delete [] aux;
    delete [] auxf;
    
    break;
  }
  case seismicload:{
    //  assembles values for seismic load
    
    
    //  auxiliary array
    aux = new double [n];
    
    stool.assemble (rhs,t);

    Mmat->gmxv (rhs,aux);
    
    copyv (aux,rhs,n);
  
    // assembling of load vector caused by forces only    
    if (flv)
      copyv (aux,flv,n);
    
    delete [] aux;
    
    break;
  }
  case responsespectrum:{
    //stool.assemble (rhs,t);
    break;
  }
  case timedepload:{
    //  load is defined by fully time dependent load case
    
    //  contributions from nodes
    for (i=0;i<nln;i++){
      ndofn=Mt->give_ndofn (lon[i].idn);
      for (j=0;j<ndofn;j++){
	k=Mt->give_dof (lon[i].idn,j);
	if (k<=0)  continue;
	if (k>0)  rhs[k-1]+=lon[i].getval(t,j);
      }
    }
    
    
    /*
    //  contributions from elements
    for (i=0;i<nle;i++){
    ndofe=Mt->elements[loe[i].ide].ndofe;
    cn = new long [ndofe];
    Mt->give_code_numbers (loe[i].ide,cn);
    
    for (j=0;j<ndofe;j++){
    k=cn[j]-1;
    if (k<0)  continue;
    else  rhs[k]+=loe[i].fe[j];
    }
    }
    */

    // assembling of load vector caused by forces only    
    if (flv)
      copyv(rhs,flv,n);

    //  contributions from prescribed displacements
    if (npd>0){
      long ne = Mt->ne;
      for (i=0;i<ne;i++){
	if (Mt->elements[i].prescdispl==1){
	  ndofe=Mt->give_ndofe (i);
	  allocm (ndofe,ndofe,sm);
	  stiffmat (i,lcid,sm);
	  
	  allocv (ndofe,r);
	  allocv (ndofe,f);
	  cn = new long [ndofe];
	  Mt->give_code_numbers (i,cn);
	  elprdispl (lcid,i,r.a,cn,ndofe);
	  mxv (sm,r,f);
	  cmulv (-1.0,f);
	  locglob (rhs,f.a,cn,ndofe);
	  destrm (sm);
	  delete [] cn;
	  destrv (r);
	  destrv (f);
	}
      }
    }
    
    break;
  }
    
  default:{
    print_err("unknown type of load case is required",__FILE__,__LINE__,__func__);
  }
  }
  
}

/**
   function assembles %vector of right hand side with respect of left hand side
   node forces cause by the attained displacements
   
   @param rhs - right hand side %vector
   @param lhs - left hand side
   @param flv   - array of load %vector caused by forces only
   
   TKo, 26.10.2001
   modified by TKo 7.2008
*/
void dloadcase::assemble (double *rhs, double *lhs, double *flv)
{
  long i,j,k,ndofn;

  //  array rhs has to be cleaned in the caller (above) function

  //  contributions from nodes
  for (i=0;i<nln;i++){
    ndofn=Mt->give_ndofn (lon[i].idn);
    for (j=0;j<ndofn;j++){
      k=Mt->give_dof (lon[i].idn,j);
      if (k<=0)  continue;
      if (k>0)  rhs[k-1]+=lon[i].getval(lhs[k-1], j);
    }
  }
  // assembling of load vector caused by forces only    
  if (flv)
    copyv(rhs,flv,Ndofm);
}


/**
   function evaluates reactions
   
   @param lcid - load case id
   
*/
void dloadcase::compute_reactions (long lcid)
{
  long i,j,k,ii,nne,ndofn,ne,te,ndofe;
  vector f,tf;
  ivector nod;

  for (i=0;i<Mt->nn;i++){
    if (Mt->nodes[i].react==1){
      ndofn=Mt->give_ndofn (i);
      for (k=0;k<ndofn;k++){
	Mt->nodes[i].r[k]=0.0;
      }
    }
  }
  
  ne=Mt->ne;
  for (i=0;i<ne;i++){
    if ((Mt->elements[i].react==1) && (Gtm->leso[i]==1)){
      te = Mt->give_elem_type (i);
      ndofe=Mt->give_ndofe (i);
      allocv(ndofe, f);
      elem_internal_forces (i,lcid,f);
      /*      if (Mt->elements[i].presctemp==1){
        //  type of eigenstrain is temperature strain
        Mm->est=tempstrain;
	allocv (ndofe,tf);
	elem_eigstrain_forces (lcid,i,tf);
	subv (f,tf,f);
	destrv (tf);
	}*/
      nne=Mt->give_nne (i);
      
      allocv(nne, nod);
      Mt->give_elemnodes (i,nod);
      ii=0;
      for (j=0;j<nne;j++){
	ndofn=Mt->give_ndofn (nod(j));
	if (Mt->nodes[nod(j)].react==1){
	  for (k=0;k<ndofn;k++){
	    Mt->nodes[nod(j)].r[k]+=f[ii];
	    ii++;
	  }
	}
	else  ii+=ndofn;
      }
      destrv(f);      
      destrv(nod);
    }
  }

  /*  if (Mp->eigstrains==1){
    
    for (i=0;i<ne;i++){
      if ((Mt->elements[i].react==1) && (Gtm->leso[i]==1)){
	te = Mt->give_elem_type (i);
	ndofe=Mt->give_ndofe (i);
	allocv (ndofe,f);
	nne=Mt->give_nne (i);
	allocv (nne,nod);
	Mt->give_elemnodes (i,nod);
        //  type of eigenstrain is eigenstrain
        Mm->est=eigstrain;
	elem_eigstrain_forces (lcid,i,f);
	ii=0;
	for (j=0;j<nne;j++){
	  ndofn=Mt->give_ndofn (nod.a[j]);
	  if (Mt->nodes[nod.a[j]].react==1){
	    for (k=0;k<ndofn;k++){
	      Mt->nodes[nod.a[j]].r[k]-=f[ii];
	      ii++;
	    }
	  }
	  else  ii+=ndofn;
	}
	destrv (f);
      }
    }
  }
  */
}

/**
   function returns prescribed displacement
   
   @param time - actual time
   @param dof - DOF id
   
   TKo, 14.11.2007
*/
double dloadcase::get_pd(double time, long dof)
{
  double ret = 0.0;
  double a;
  long i;
  
  switch (tdl){
  case timeindload:{
    for (i=0; i<nslc; i++)
      {
        //  scale factor
        a=gf[i].getval(time);
        //  assembling of contributions
        ret += slc[i].pd[0-dof-1]*a;
      }
    break;
  }
  case timedepload:{
    ret = pd[0-dof-1].getval(Mp->time);
    break;
  }
  default:{
    fprintf(stderr, "\n\n This type of load case is not supported\n");
    fprintf(stderr, " in function dloadcase::get_pd.(file %s, line %d)\n", __FILE__, __LINE__);
    abort();
  }
  }
  return ret;
}

/**
   function computes contributions from temperature changes
   to the nodal forces

   @param lcid - load case id
   @param rhs - pointer to the right hand side
   @param n - number of components of the array rhs
   @param t - actual time
   
   21.11.2007
*/
void dloadcase::tempercontrib (long lcid,double *rhs,long n,double t)
{
  long i;
  double a;
  double *aux;

  switch (tdl){
  case timeindload:{
    aux = new double [n];
    for (i=0;i<nslc;i++){
      nullv (aux,n);
      //  scale factor
      a=gf[i].getval(t);
      
      //  assembling of vector of contributions
      slc[i].tempercontrib (lcid,aux,a);
      addv (rhs,aux,n);
    }
    delete [] aux;
    break;
  }
  default:{
    fprintf (stderr,"\n\n unknown type of load case is required in function tempercontrib (file %s, line %d).\n",
	     __FILE__,__LINE__);
  }
  }

}
