#include <stdlib.h>
#include <math.h>
#include "linhex.h"
#include "global.h"
#include "globmat.h"
#include "genfile.h"
#include "intpoints.h"
#include "node.h"
#include "element.h"
#include "loadcase.h"




linhex::linhex (void)
{
  long i;

  nne=8;  ndofe=24;  tnip=9;  tncomp=6;  napfun=3;
  intordmm=2;  ned=12;  nned=2;
  ssst=spacestress;
  
  nb=1;
  
  ncomp = new long [nb];
  ncomp[0]=6;

  cncomp = new long [nb];
  cncomp[0]=0;
  
  nip = new long* [nb];
  intordsm = new long* [nb];
  for (i=0;i<nb;i++){
    nip[i] = new long [nb];
    intordsm[i] = new long [nb];
  }
  
  nip[0][0]=8;
  
  intordsm[0][0]=2;
}

linhex::~linhex (void)
{
  long i;
  
  for (i=0;i<nb;i++){
    delete [] nip[i];
    delete [] intordsm[i];
  }
  delete [] nip;
  delete [] intordsm;
  
  delete [] cncomp;
  delete [] ncomp;
}

void linhex::eleminit (long eid)
{
  long ii,jj;
  Mt->elements[eid].nb=nb;
  Mt->elements[eid].intordsm = new long* [nb];
  Mt->elements[eid].nip = new long* [nb];

  for (ii=0;ii<nb;ii++){
    Mt->elements[eid].intordsm[ii] = new long [nb];
    Mt->elements[eid].nip[ii] = new long [nb];
    for (jj=0;jj<nb;jj++){
      Mt->elements[eid].intordsm[ii][jj]=intordsm[ii][jj];
      Mt->elements[eid].nip[ii][jj]=nip[ii][jj];
    }
  }
}

/**
   function approximates function defined by nodal values
   
   @param xi,eta,zeta - natural coordinates
   @param nodval - nodal values
   
   20.8.2001
*/
double linhex::approx (double xi,double eta,double zeta,vector &nodval)
{
  double f;
  vector bf(nne);
  
  bf_lin_hex_3d (bf.a,xi,eta,zeta);
  
  scprd (bf,nodval,f);

  return f;
}

/**
   function assembles matrix of base functions
   
   @param n - matrix of base functions
   @param xi,eta,zeta - natural coordinates

   19.7.2001
*/
void linhex::bf_matrix (matrix &n,double xi,double eta,double zeta)
{
  long i,j,k,l;
  vector bf(nne);

  fillm (0.0,n);

  bf_lin_hex_3d (bf.a,xi,eta,zeta);
  
  j=0;  k=1;  l=2;
  for (i=0;i<nne;i++){
    n[0][j]=bf[i];  j+=3;
    n[1][k]=bf[i];  k+=3;
    n[2][l]=bf[i];  l+=3;
  }
}

/**
   function assembles geometric matrix

   @param gm - geometric matrix
   @param x,y,z - vectors containing element node coordinates
   @param xi,eta,zeta - naturalcoordinates
   @param jac - Jacobian

   19.7.2001
*/
void linhex::geom_matrix (matrix &gm,vector &x,vector &y,vector &z,
			  double xi,double eta,double zeta,double &jac)
{
  long i,j,k,l;
  vector dx(nne),dy(nne),dz(nne);

  dx_bf_lin_hex_3d (dx.a,eta,zeta);
  dy_bf_lin_hex_3d (dy.a,xi,zeta);
  dz_bf_lin_hex_3d (dz.a,xi,eta);

  derivatives_3d (dx,dy,dz,jac,x,y,z,xi,eta,zeta);

  fillm (0.0,gm);

  j=0;  k=1;  l=2;
  for (i=0;i<nne;i++){
    gm[0][j]=dx[i];
    gm[1][k]=dy[i];
    gm[2][l]=dz[i];
    
    gm[3][k]=dz[i];
    gm[3][l]=dy[i];
    
    gm[4][j]=dz[i];
    gm[4][l]=dx[i];
    
    gm[5][j]=dy[i];
    gm[5][k]=dx[i];
    
    j+=3;  k+=3;  l+=3;
  }
}

/**
   function assembles transformation matrix
   
   @param nodes - nodes of element
   @param tmat - transformation matrix
   
*/
void linhex::transf_matrix (ivector &nodes,matrix &tmat)
{
  long i,n,m;

  fillm (0.0,tmat);

  n=nodes.n;
  m=tmat.m;
  for (i=0;i<m;i++){
    tmat[i][i]=1.0;
  }
  
  for (i=0;i<n;i++){
    if (Mt->nodes[nodes[i]].transf>0){
      tmat[i*3+0][i*3]=Mt->nodes[nodes[i]].e1[0];
      tmat[i*3+1][i*3]=Mt->nodes[nodes[i]].e1[1];
      tmat[i*3+2][i*3]=Mt->nodes[nodes[i]].e1[2];

      tmat[i*3+0][i*3+1]=Mt->nodes[nodes[i]].e2[0];
      tmat[i*3+1][i*3+1]=Mt->nodes[nodes[i]].e2[1];
      tmat[i*3+2][i*3+1]=Mt->nodes[nodes[i]].e2[2];

      tmat[i*3+0][i*3+2]=Mt->nodes[nodes[i]].e3[0];
      tmat[i*3+1][i*3+2]=Mt->nodes[nodes[i]].e3[1];
      tmat[i*3+2][i*3+2]=Mt->nodes[nodes[i]].e3[2];
    }
  }
}

/**
   function computes stiffness matrix of one element

   @param eid - number of element
   @param ri,ci - row and column indices
   @param sm - stiffness matrix

   19.7.2001
*/
void linhex::stiffness_matrix (long eid,long ri,long ci,matrix &sm)
{
  long i,j,k,ii,jj,ipp,transf;
  double xi,eta,zeta,jac;
  vector x(nne),y(nne),z(nne),w,gp;
  matrix gm,d(tncomp,tncomp);
  
  Mt->give_node_coord3d (x,y,z,eid);
  
  fillm (0.0,sm);

  for (ii=0;ii<nb;ii++){
    allocm (ncomp[ii],ndofe,gm);
    for (jj=0;jj<nb;jj++){
      if (intordsm[ii][jj]==0)  continue;

      allocv (intordsm[ii][jj],w);
      allocv (intordsm[ii][jj],gp);
      
      gauss_points (gp.a,w.a,intordsm[ii][jj]);
      
      ipp=Mt->elements[eid].ipp[ri+ii][ci+jj];
      
      for (i=0;i<intordsm[ii][jj];i++){
	xi=gp[i];
	for (j=0;j<intordsm[ii][jj];j++){
	  eta=gp[j];
	  for (k=0;k<intordsm[ii][jj];k++){
	    zeta=gp[k];
	    
	    //  geometric matrices
	    geom_matrix (gm,x,y,z,xi,eta,zeta,jac);
	    
	    Mm->matstiff (d,ipp);  ipp++;
	    
	    jac=fabs(jac);
	    
	    jac*=w[i]*w[j]*w[k];
	    
	    //  contribution to the stiffness matrix of the element
	    bdbjac (sm,gm,d,gm,jac);
	    
	  }
	}
      }
      destrv (gp);  destrv (w);
    }
    destrm (gm);
  }
  
  //  transformation of stiffness matrix
  ivector nodes (nne);
  Mt->give_elemnodes (eid,nodes);
  transf = Mt->locsystems (nodes);
  if (transf>0){
    matrix tmat (ndofe,ndofe);
    transf_matrix (nodes,tmat);
    glmatrixtransf (sm,tmat);
  }
  
}

/**
   function assembles resulting stiffness matrix of the element
   
   @param eid - element id
   @param sm - stiffness matrix
   
   JK, 9.5.2002
*/
void linhex::res_stiffness_matrix (long eid,matrix &sm)
{
  stiffness_matrix (eid,0,0,sm);
}

/**
   function computes mass matrix

   @param eid - number of element
   @param mm - mass matrix

   19.7.2001
*/
void linhex::mass_matrix (long eid,matrix &mm)
{
  long i,j,k;
  double jac,xi,eta,zeta,rho;
  ivector nodes (nne);
  vector x(nne),y(nne),z(nne),w(intordmm),gp(intordmm),dens(nne);
  matrix n(napfun,ndofe);
  
  Mt->give_elemnodes (eid,nodes);
  Mc->give_density (eid,nodes,dens);

  Mt->give_node_coord3d (x,y,z,eid);
  gauss_points (gp.a,w.a,intordmm);
  
  fillm (0.0,mm);
  
  for (i=0;i<intordmm;i++){
    xi=gp[i];
    for (j=0;j<intordmm;j++){
      eta=gp[j];
      for (k=0;k<intordmm;k++){
	zeta=gp[k];
	
	jac_3d (jac,x,y,z,xi,eta,zeta);
	
	jac=fabs(jac);

	bf_matrix (n,xi,eta,zeta);

	rho = approx (xi,eta,zeta,dens);
	
	jac*=w[i]*w[j]*w[k]*rho;
	
	nnj (mm.a,n.a,jac,n.m,n.n);
      }
    }
  }
  
}

/**
   function computes load matrix

   @param eid - number of element
   @param lm - load matrix
   
   25.7.2001
*/
void linhex::load_matrix (long eid,matrix &lm)
{
  long i,j,k;
  double jac,xi,eta,zeta,w1,w2,w3;
  ivector nodes (nne);
  vector x(nne),y(nne),z(nne),w(intordmm),gp(intordmm);
  matrix n(napfun,ndofe);
  
  Mt->give_elemnodes (eid,nodes);
  Mt->give_node_coord3d (x,y,z,eid);
  gauss_points (gp.a,w.a,intordmm);
  
  fillm (0.0,lm);
  
  for (i=0;i<intordmm;i++){
    xi=gp[i];  w1=w[i];
    for (j=0;j<intordmm;j++){
      eta=gp[j];  w2=w[j];
      for (k=0;k<intordmm;k++){
	zeta=gp[k];  w3=w[k];
	
	jac_3d (jac,x,y,z,xi,eta,zeta);

	bf_matrix (n,xi,eta,zeta);

	jac*=w1*w2*w3;
	
	nnj (lm.a,n.a,jac,n.m,n.n);
      }
    }
  }
  
}







void linhex::res_mainip_strains (long lcid,long eid)
{
  mainip_strains (lcid,eid,0,0);
}


/**
   function computes strains in main integration points of element
   
   @param lcid - load case id
   @param eid - element id
   @param ri - row index
   @param ci - column index
   
   10.5.2002
*/
void linhex::mainip_strains (long lcid,long eid,long ri,long ci)
{
  long i,j,k,ii,ipp;
  double xi,eta,zeta,jac;
  vector x(nne),y(nne),z(nne),r(ndofe),gp,w,eps,aux;
  ivector nodes(nne),cn(ndofe);
  matrix gm,tmat;

  Mt->give_elemnodes (eid,nodes);
  Mt->give_node_coord3d (x,y,z,eid);
  Mt->give_code_numbers (eid,cn.a);
  eldispl (lcid,r.a,cn.a,ndofe);
  
  //  transformation of displacement vector
  long transf = Mt->locsystems (nodes);
  if (transf>0){
    allocv (ndofe,aux);
    allocm (ndofe,ndofe,tmat);
    transf_matrix (nodes,tmat);
    locglobtransf (aux,r,tmat);
    copyv (aux,r);
    destrv (aux);
    destrm (tmat);
  }
  
  
  for (ii=0;ii<nb;ii++){
    if (intordsm[ii][ii]==0)  continue;
    
    allocv (intordsm[ii][ii],gp);
    allocv (intordsm[ii][ii],w);
    allocv (ncomp[0],eps);
    allocm (ncomp[0],ndofe,gm);
    
    gauss_points (gp.a,w.a,intordsm[ii][ii]);
    
    ipp=Mt->elements[eid].ipp[ri+ii][ci+ii];
    for (i=0;i<intordsm[ii][ii];i++){
      xi=gp[i];
      for (j=0;j<intordsm[ii][ii];j++){
	eta=gp[j];
	for (k=0;k<intordsm[ii][ii];k++){
	  zeta=gp[k];
	  
	  geom_matrix (gm,x,y,z,xi,eta,zeta,jac);
	  
	  mxv (gm,r,eps);
	  
	  Mm->storestrain (lcid,ipp,cncomp[0],ncomp[0],eps);

	  ipp++;
	}
      }
    }
        
    destrm (gm);  destrv (w);  destrv (gp);
    destrv (eps);
  }
  
}

/**
   function computes strains in nodes of element
   
   @param lcid - load case id
   @param eid - element id
   
   10.5.2002
*/
void linhex::nod_strains (long lcid,long eid,long ri,long ci)
{
  long i,j,k,ii,ipp;
  double xi,eta,zeta,*lsm,*lhs,*rhs;
  vector nxi(nne),neta(nne),nzeta(nne),gp,w,eps,natcoord(3);
  ivector nodes(nne);

  lsm = new double [16];

  nodecoord (nxi,neta,nzeta);
  Mt->give_elemnodes (eid,nodes);
  
  for (ii=0;ii<nb;ii++){
    if (intordsm[ii][ii]==0)  continue;

    allocv (intordsm[ii][ii],gp);
    allocv (intordsm[ii][ii],w);
    allocv (ncomp[ii],eps);
    lhs = new double [ncomp[ii]*4];
    rhs = new double [ncomp[ii]*4];
    gauss_points (gp.a,w.a,intordsm[ii][ii]);
    
    nullv (lsm,16);
    nullv (rhs,ncomp[ii]*4);
    
    ipp=Mt->elements[eid].ipp[ri+ii][ci+ii];
    for (i=0;i<intordsm[ii][ii];i++){
      xi=gp[i];
      for (j=0;j<intordsm[ii][ii];j++){
	eta=gp[j];
	for (k=0;k<intordsm[ii][ii];k++){
	  zeta=gp[k];
	  
	  Mm->givestrain (lcid,ipp,cncomp[ii],ncomp[ii],eps);

	  natcoord[0]=xi;  natcoord[1]=eta;  natcoord[2]=zeta;
	  matassem_lsm (lsm,natcoord);
	  rhsassem_lsm (rhs,natcoord,eps);
	  
	  ipp++;
	}
      }
    }
    
    solve_lsm (lsm,lhs,rhs,Mp->zero,4,ncomp[ii]);
    Mt->strain_nodal_values (nodes,nxi,neta,nzeta,lhs,3,cncomp[ii],ncomp[ii],lcid);
    
    delete [] lhs;  delete [] rhs;
    destrv (eps);  destrv (w);  destrv (gp);
  }
  
  delete [] lsm;
}

/**
   function computes strains on element
   
   @param val - array containing strains on element
   @param lcid - load case id
   @param eid - element id
   
   19.9.2002
*/
void linhex::elem_strains (double **stra,long lcid,long eid,long ri,long ci)
{
  long i,j,k,ii,ipp;
  double xi,eta,zeta,*lsm,*lhs,*rhs;
  vector nxi(nne),neta(nne),nzeta(nne),gp,w,eps,natcoord(3);

  lsm = new double [16];

  nodecoord (nxi,neta,nzeta);
  
  for (ii=0;ii<nb;ii++){
    allocv (intordsm[ii][ii],gp);
    allocv (intordsm[ii][ii],w);
    allocv (ncomp[ii],eps);
    lhs = new double [ncomp[ii]*4];
    rhs = new double [ncomp[ii]*4];
    gauss_points (gp.a,w.a,intordsm[ii][ii]);
    
    nullv (lsm,16);
    nullv (rhs,ncomp[ii]*4);
    
    ipp=Mt->elements[eid].ipp[ri+ii][ci+ii];
    for (i=0;i<intordsm[ii][ii];i++){
      xi=gp[i];
      for (j=0;j<intordsm[ii][ii];j++){
	eta=gp[j];
	for (k=0;k<intordsm[ii][ii];k++){
	  zeta=gp[k];
	  
	  Mm->givestrain (lcid,ipp,cncomp[ii],ncomp[ii],eps);
	  
	  natcoord[0]=xi;  natcoord[1]=eta;  natcoord[2]=zeta;
	  matassem_lsm (lsm,natcoord);
	  rhsassem_lsm (rhs,natcoord,eps);
	  
	  ipp++;
	}
      }
    }
        
    solve_lsm (lsm,lhs,rhs,Mp->zero,4,ncomp[ii]);
    nodal_values (stra,nxi,neta,nzeta,lhs,3,cncomp[ii],ncomp[ii]);

    delete [] lhs;  delete [] rhs;
    destrv (eps);  destrv (w);  destrv (gp);
  }
  
  delete [] lsm;
}


/**
   function computes strains in arbitrary point on element
   
   @param lcid - load case id
   @param eid - element id
   @param xi, eta, zeta - natural coordinates of the point
   @param fi - first index
   @param ncomp - number of components
   @param eps - array containing strains
   
   11.5.2002
*/
void linhex::appstrain (long lcid,long eid,double xi,double eta,double zeta,long fi,long ncomp,vector &eps)
{
  long i,j,k;
  ivector nod(nne);
  vector nodval(nne);
  
  if (ncomp != eps.n){
    fprintf (stderr,"\n\n wrong interval of indices in function strain (%s, line %d).\n",__FILE__,__LINE__);
    abort ();
  }

  Mt->give_elemnodes (eid,nod);
  k=0;
  for (i=fi;i<fi+ncomp;i++){
    for (j=0;j<nne;j++){
      nodval[j]=Mt->nodes[nod[j]].strain[lcid*tncomp+i];
    }
    eps[k]=approx (xi,eta,zeta,nodval);
    k++;
  }
}

/**
   function computes strains in all integration points
   
   @param lcid - load case id
   @param eid - element id
   @param ri,ci - row and column indices
   
   10.5.2002
*/
void linhex::allip_strains (double **stra,long lcid,long eid,long ri,long ci)
{
  long i,j,k,ii,jj,ipp;
  double xi,eta,zeta;
  vector eps(tncomp),gp,w;
  
  for (ii=0;ii<nb;ii++){
    for (jj=0;jj<nb;jj++){
      if (intordsm[ii][jj]==0)  continue;
      
      allocv (intordsm[ii][jj],gp);
      allocv (intordsm[ii][jj],w);
      gauss_points (gp.a,w.a,intordsm[ii][jj]);
      ipp=Mt->elements[eid].ipp[ri+ii][ci+jj];
      
      for (i=0;i<intordsm[ii][jj];i++){
	xi=gp[i];
	for (j=0;j<intordsm[ii][jj];j++){
	  eta=gp[j];
	  for (k=0;k<intordsm[ii][jj];k++){
	    zeta=gp[k];
	    
	    if (Mp->strainaver==0)
	      appval (xi,eta,zeta,0,tncomp,eps,stra);
	    if (Mp->strainaver==1)
	      appstrain (lcid,eid,xi,eta,zeta,0,tncomp,eps);
	    
	    Mm->storestrain (lcid,ipp,eps);
	    ipp++;
	  }
	}
      }
      destrv (w);  destrv (gp);
    }
  }
}

void linhex::strains (long lcid,long eid,long ri,long ci)
{
  long i,naep,ncp,sid;
  double **stra;
  vector coord,eps;
  
  if (Mp->strainaver==0){
    stra = new double* [nne];
    for (i=0;i<nne;i++){
      stra[i] = new double [tncomp];
    }
    elem_strains (stra,lcid,eid,ri,ci);
  }
  
  switch (Mm->stra.tape[eid]){
  case nowhere:{
    break;
  }
  case intpts:{
    allip_strains (stra,lcid,eid,ri,ci);
    break;
  }
  case enodes:{
    break;
  }
  case userdefined:{
    //  number of auxiliary element points
    naep = Mm->stra.give_naep (eid);
    ncp = Mm->stra.give_ncomp (eid);
    sid = Mm->stra.give_sid (eid);
    allocv (ncp,eps);
    allocv (3,coord);
    for (i=0;i<naep;i++){
      Mm->stra.give_aepcoord (sid,i,coord);

      if (Mp->strainaver==0)
	appval (coord[0],coord[1],coord[2],0,ncp,eps,stra);
      if (Mp->strainaver==1)
	appstrain (lcid,eid,coord[0],coord[1],coord[2],0,ncp,eps);

      Mm->stra.storevalues(lcid,eid,i,eps);
    }
    destrv (eps);
    destrv (coord);
    break;
  }
  default:{
    fprintf (stderr,"\n\n unknown strain point is required in function planeelemlq::strains (%s, line %d).\n",__FILE__,__LINE__);
  }
  }
  
  if (Mp->strainaver==0){
    for (i=0;i<nne;i++){
      delete [] stra[i];
    }
    delete [] stra;
  }
}




/**
   function assembles natural coordinates of nodes of element
   
   @param xi - array containing natural coordinates xi
   @param eta - array containing natrual coordinates eta
   @param zeta - array containing natrual coordinates zeta
   
   10.5.2002
*/
void linhex::nodecoord (vector &xi,vector &eta,vector &zeta)
{
  xi[0] =  1.0;  eta[0] =  1.0;  zeta[0] =  1.0;
  xi[1] = -1.0;  eta[1] =  1.0;  zeta[1] =  1.0;
  xi[2] = -1.0;  eta[2] = -1.0;  zeta[2] =  1.0;
  xi[3] =  1.0;  eta[3] = -1.0;  zeta[3] =  1.0;
  xi[4] =  1.0;  eta[4] =  1.0;  zeta[4] = -1.0;
  xi[5] = -1.0;  eta[5] =  1.0;  zeta[5] = -1.0;
  xi[6] = -1.0;  eta[6] = -1.0;  zeta[6] = -1.0;
  xi[7] =  1.0;  eta[7] = -1.0;  zeta[7] = -1.0;
}

/**
   function computes strains in arbitrary point on element
   
   @param xi, eta - natural coordinates of the point
   @param eps - array containing strains
   @param val - array containing values on element
   
   11.5.2002
*/
void linhex::appval (double xi,double eta,double zeta,long fi,long nc,vector &eps,double **val)
{
  long i,j,k;
  vector nodval(nne);
  
  k=0;
  for (i=fi;i<fi+nc;i++){
    for (j=0;j<nne;j++){
      nodval[j]=val[j][i];
    }
    eps[k]=approx (xi,eta,zeta,nodval);
    k++;
  }
}


/**
   function computes stresses in main integration points of element
   
   @param lcid - load case id
   @param eid - element id
   @param ri - row index
   @param ci - column index
   
   10.5.2002
*/
void linhex::mainip_stresses (long lcid,long eid,long ri,long ci)
{
  long i,j,k,ii,jj,ipp;
  double xi,eta,zeta;
  vector gp,w,eps,epst,epstt,sig,auxsig;
  matrix d(tncomp,tncomp);

  for (ii=0;ii<nb;ii++){
    if (intordsm[ii][ii]==0)  continue;
    
    allocv (ncomp[ii],sig);
    allocv (ncomp[ii],auxsig);
    allocv (intordsm[ii][ii],gp);
    allocv (intordsm[ii][ii],w);
    
    gauss_points (gp.a,w.a,intordsm[ii][ii]);
    ipp=Mt->elements[eid].ipp[ri+ii][ci+ii];
    
    for (i=0;i<intordsm[ii][ii];i++){
      xi=gp[i];
      for (j=0;j<intordsm[ii][ii];j++){
	eta=gp[j];
	for (k=0;k<intordsm[ii][ii];k++){
	  zeta=gp[k];
	  
	  Mm->matstiff (d,ipp);
	  
	  fillv (0.0,sig);
	  for (jj=0;jj<nb;jj++){
	    allocv (ncomp[jj],eps);


	    if (Mp->strainaver==0)
	      Mm->givestrain (lcid,ipp,cncomp[jj],ncomp[jj],eps);
	    if (Mp->strainaver==1)
	      appstrain (lcid,eid,xi,eta,zeta,cncomp[jj],ncomp[jj],eps);
	    

	    mxv (d,eps,auxsig);
	    addv (auxsig,sig,sig);
	    destrv (eps);
	  }
	  
	  Mm->storestress (lcid,ipp,sig);
	  
	  ipp++;
	}
      }
    }
    
    destrv (w);  destrv (gp);  destrv (auxsig);  destrv (sig);
  }
}

/**
   function computes stresses in nodes
   
   @param lcid - load case id
   @param eid - element id
   @param ri,ci - row and column indices
   
   10.5.2002
*/
void linhex::nod_stresses (long lcid,long eid,long ri,long ci)
{
  long i,j,k,ii,jj,ipp;
  double xi,eta,zeta,*lsm,*lhs,*rhs;
  vector nxi(nne),neta(nne),nzeta(nne),r(ndofe),gp,w,eps,epst,epstt,sig,auxsig,natcoord(3);
  ivector nodes(nne);
  matrix d(tncomp,tncomp);

  lsm = new double [16];

  nodecoord (nxi,neta,nzeta);
  Mt->give_elemnodes (eid,nodes);

  for (ii=0;ii<nb;ii++){
    if (intordsm[ii][ii]==0)  continue;

    allocv (intordsm[ii][ii],gp);
    allocv (intordsm[ii][ii],w);
    allocv (ncomp[ii],sig);
    allocv (ncomp[ii],auxsig);
    lhs = new double [ncomp[ii]*4];
    rhs = new double [ncomp[ii]*4];
    gauss_points (gp.a,w.a,intordsm[ii][ii]);
    
    nullv (lsm,16);
    nullv (rhs,ncomp[ii]*4);
    
    ipp=Mt->elements[eid].ipp[ri+ii][ci+ii];
    
    for (i=0;i<intordsm[ii][ii];i++){
      xi=gp[i];
      for (j=0;j<intordsm[ii][ii];j++){
	eta=gp[j];
	for (k=0;k<intordsm[ii][ii];k++){
	  zeta=gp[k];
	  
	  Mm->matstiff (d,ipp);
	  
	  fillv (0.0,sig);
	  for (jj=0;jj<nb;jj++){
	    allocv (ncomp[jj],eps);
	    
	    if (Mp->strainaver==0)
	      Mm->givestrain (lcid,ipp,cncomp[jj],ncomp[jj],eps);
	    if (Mp->strainaver==1)
	      appstrain (lcid,eid,xi,eta,zeta,cncomp[jj],ncomp[jj],eps);
	    
	    mxv (d,eps,auxsig);
	    addv (auxsig,sig,sig);
	    destrv (eps);
	  }
	  
	  natcoord[0]=xi;  natcoord[1]=eta;  natcoord[2]=zeta;
	  matassem_lsm (lsm,natcoord);
	  rhsassem_lsm (rhs,natcoord,sig);

	  ipp++;
	}
      }
    }
    
    solve_lsm (lsm,lhs,rhs,Mp->zero,4,ncomp[ii]);
    Mt->stress_nodal_values (nodes,nxi,neta,nzeta,lhs,3,cncomp[ii],ncomp[ii],lcid);
    
    
    delete [] lhs;  delete [] rhs;
    destrv (auxsig);  destrv (sig);  destrv (eps);  destrv (w);  destrv (gp);
  }
  
  delete [] lsm;
}

void linhex::elem_stresses (double **stra,double **stre,long lcid,long eid,long ri,long ci)
{
  long i,j,k,ii,jj,ipp;
  double xi,eta,zeta,*lsm,*lhs,*rhs;
  vector nxi(nne),neta(nne),nzeta(nne),r,gp,w,eps,epst,epstt,sig,auxsig,natcoord(3);
  matrix d(tncomp,tncomp);

  lsm = new double [16];

  nodecoord (nxi,neta,nzeta);
  
  for (ii=0;ii<nb;ii++){
    allocv (intordsm[ii][ii],gp);
    allocv (intordsm[ii][ii],w);
    allocv (ncomp[ii],sig);
    allocv (ncomp[ii],auxsig);
    lhs = new double [ncomp[ii]*4];
    rhs = new double [ncomp[ii]*4];
    gauss_points (gp.a,w.a,intordsm[ii][ii]);
    
    nullv (lsm,16);
    nullv (rhs,ncomp[ii]*4);
    
    ipp=Mt->elements[eid].ipp[ri+ii][ci+ii];
    
    for (i=0;i<intordsm[ii][ii];i++){
      xi=gp[i];
      for (j=0;j<intordsm[ii][ii];j++){
	eta=gp[j];
	for (k=0;k<intordsm[ii][ii];k++){
	  zeta=gp[k];
	  
	  Mm->matstiff (d,ipp);
	  
	  fillv (0.0,sig);
	  for (jj=0;jj<nb;jj++){
	    allocv (ncomp[jj],eps);
	    
	    if (Mp->strainaver==0)
	      appval (xi,eta,zeta,cncomp[jj],ncomp[jj],eps,stra);
	    if (Mp->strainaver==1)
	      appstrain (lcid,eid,xi,eta,zeta,cncomp[jj],ncomp[jj],eps);
	    
	    mxv (d,eps,auxsig);
	    addv (auxsig,sig,sig);
	    destrv (eps);
	  }
	  
	  natcoord[0]=xi;  natcoord[1]=eta;  natcoord[2]=zeta;
	  matassem_lsm (lsm,natcoord);
	  rhsassem_lsm (rhs,natcoord,sig);
	  
	  ipp++;
	}
      }
    }
    
    solve_lsm (lsm,lhs,rhs,Mp->zero,4,ncomp[ii]);
    nodal_values (stre,nxi,neta,nzeta,lhs,3,cncomp[ii],ncomp[ii]);
    
    
    delete [] lhs;  delete [] rhs;
    destrv (auxsig);  destrv (sig);  destrv (eps);  destrv (w);  destrv (gp);
  }

  delete [] lsm;
}


/**
   function computes stresses in arbitrary point on element
   
   @param lcid - load case id
   @param eid - element id
   @param xi, eta, zeta - natural coordinates of the point
   @param fi,li - first and last indices
   @param sig - array containing stresses
   
   11.5.2002
*/
void linhex::appstress (long lcid,long eid,double xi,double eta,double zeta,long fi,long ncomp,vector &sig)
{
  long i,j,k;
  ivector nodes(nne);
  vector nodval(nne);
  
  if (ncomp != sig.n){
    fprintf (stderr,"\n\n wrong interval of indices in function stress (%s, line %d).\n",__FILE__,__LINE__);
    abort ();
  }

  Mt->give_elemnodes (eid,nodes);
  k=0;
  for (i=fi;i<fi+ncomp;i++){
    for (j=0;j<nne;j++){
      nodval[j]=Mt->nodes[nodes[j]].stress[lcid*tncomp+i];
    }
    sig[k]=approx (xi,eta,zeta,nodval);
    k++;
  }
}

/**
   function computes stresses in all integration points
   
   @param lcid - load case id
   @param eid - element id
   @param ri,ci - row and column indices
   
   10.5.2002
*/
void linhex::allip_stresses (double **stre,long lcid,long eid,long ri,long ci)
{
  long i,j,k,ii,jj,ipp;
  double xi,eta,zeta;
  vector sig(tncomp),gp,w;
  
  for (ii=0;ii<nb;ii++){
    for (jj=0;jj<nb;jj++){
      if (intordsm[ii][jj]==0)  continue;

      allocv (intordsm[ii][jj],gp);
      allocv (intordsm[ii][jj],w);
      gauss_points (gp.a,w.a,intordsm[ii][jj]);
      ipp=Mt->elements[eid].ipp[ri+ii][ci+jj];
      
      for (i=0;i<intordsm[ii][jj];i++){
	xi=gp[i];
	for (j=0;j<intordsm[ii][jj];j++){
	  eta=gp[j];
	  for (k=0;k<intordsm[ii][jj];k++){
	    zeta=gp[k];
	    
	    if (Mp->stressaver==0)
	      appval (xi,eta,zeta,0,tncomp,sig,stre);
	    if (Mp->stressaver==1)
	      appstress (lcid,eid,xi,eta,zeta,0,tncomp,sig);

	    Mm->storestress (lcid,ipp,sig);
	    ipp++;
	  }
	}
      }
      destrv (w);  destrv (gp);
    }
  }
}







void linhex::stresses (long lcid,long eid,long ri,long ci)
{
  long i,naep,ncp,sid;
  double **stra,**stre;
  vector coord,sig;
  
  if (Mp->stressaver==0){
    stra = new double* [nne];
    stre = new double* [nne];
    for (i=0;i<nne;i++){
      stra[i] = new double [tncomp];
      stre[i] = new double [tncomp];
    }
    elem_strains (stra,lcid,eid,ri,ci);
    elem_stresses (stra,stre,lcid,eid,ri,ci);
  }

  switch (Mm->stre.tape[eid]){
  case nowhere:{
    break;
  }
  case intpts:{
    allip_stresses (stre,lcid,eid,ri,ci);
    break;
  }
  case enodes:{
    break;
  }
  case userdefined:{
    //  number of auxiliary element points
    naep = Mm->stre.give_naep (eid);
    ncp = Mm->stre.give_ncomp (eid);
    sid = Mm->stre.give_sid (eid);
    allocv (ncp,sig);
    allocv (3,coord);
    for (i=0;i<naep;i++){
      Mm->stre.give_aepcoord (sid,i,coord);

      if (Mp->stressaver==0)
	appval (coord[0],coord[1],coord[2],0,ncp,sig,stre);
      if (Mp->stressaver==1)
	appstress (lcid,eid,coord[0],coord[1],coord[2],0,ncp,sig);
      
      Mm->stre.storevalues(lcid,eid,i,sig);
    }
    destrv (sig);
    destrv (coord);
    break;
  }
  default:{
    fprintf (stderr,"\n\n unknown stress point is required in function planeelemlq::stresses (%s, line %d).\n",__FILE__,__LINE__);
  }
  }

  if (Mp->stressaver==0){
    for (i=0;i<nne;i++){
      delete [] stra[i];
      delete [] stre[i];
    }
    delete [] stra;
    delete [] stre;
  }
}











void linhex::res_temp_forces (long lcid,long eid,vector &nfor)
{
  temp_forces (lcid,eid,0,0,nfor);
}

/**
   function computes nodal forces caused by temperature changes
   
   @param eid - element id
   @param ri,ci - row and column indices
   @param nfor - array containing nodal forces
   
   30.11.2002, JK
*/
void linhex::temp_forces (long lcid,long eid,long ri,long ci,vector &nfor)
{
  long i,j,k,l,ii,ipp;
  double xi,eta,zeta,jac;
  ivector nodes(nne);
  vector x(nne),y(nne),z(nne),eps,sig,contr(ndofe),t(nne),epst(tncomp),gp,w;
  matrix d(tncomp,tncomp),gm;
  
  Mt->give_elemnodes (eid,nodes);
  Mt->give_node_coord3d (x,y,z,eid);
  
  fillv (0.0,nfor);

  for (ii=0;ii<nb;ii++){
    
    allocv (intordsm[ii][ii],w);
    allocv (intordsm[ii][ii],gp);
    
    allocm (ncomp[ii],ndofe,gm);
    allocv (ncomp[ii],eps);
    allocv (ncomp[ii],sig);
    
    gauss_points (gp.a,w.a,intordsm[ii][ii]);
    
    ipp=Mt->elements[eid].ipp[ri+ii][ci+ii];
    for (i=0;i<intordsm[ii][ii];i++){
      xi=gp[i];
      for (j=0;j<intordsm[ii][ii];j++){
	eta=gp[j];
	for (k=0;k<intordsm[ii][ii];k++){
	  zeta=gp[k];
	  tempstrains (lcid,eid,ipp,xi,eta,zeta,epst);
	  extract (eps,epst,cncomp[ii],ncomp[ii]);
	  
	  Mm->matstiff (d,ipp);
	  ipp++;
	  
	  mxv (d,eps,sig);
	  geom_matrix (gm,x,y,z,xi,eta,zeta,jac);
	  mtxv (gm,sig,contr);
	  cmulv (jac*w[i]*w[j]*w[k],contr);
	  
	  for (l=0;l<contr.n;l++){
	    nfor[l]+=contr[l];
	  }
	}
      }
    }
    destrv (sig);  destrv (eps);  destrv (gp);  destrv (w);
    destrm (gm);
  }
}

/**
   function computes nodal forces caused by temperature changes
   
   @param eid - element id
   @param ri,ci - row and column indices
   @param nfor - array containing nodal forces
   
   30.11.2002, JK
*/
void linhex::temperaturestrains (long lcid,long eid,long ri,long ci)
{
  long i,j,k,ii,ipp;
  double xi,eta,zeta;
  vector epst(tncomp),gp,w;
  
  for (ii=0;ii<nb;ii++){
    
    allocv (intordsm[ii][ii],w);
    allocv (intordsm[ii][ii],gp);
    
    gauss_points (gp.a,w.a,intordsm[ii][ii]);
    
    ipp=Mt->elements[eid].ipp[ri+ii][ci+ii];
    for (i=0;i<intordsm[ii][ii];i++){
      xi=gp[i];
      for (j=0;j<intordsm[ii][ii];j++){
	eta=gp[j];
	for (k=0;k<intordsm[ii][ii];k++){
	  zeta=gp[k];

	  tempstrains (lcid,eid,ipp,xi,eta,zeta,epst);

	  ipp++;
	}
      }
    }
    destrv (gp);  destrv (w);
  }
}


/**
   function computes strains caused by temperature changes
   
   @param lcid - load case id
   @param eid - element id
   @param ipp - integration point pointer
   @param xi,eta,zeta - natural coordinates
   @param eps - array containing strains
   
   30.11.2002, JK
*/
void linhex::tempstrains (long lcid,long eid,long ipp,double xi,double eta,double zeta,vector &eps)
{
  double temp;
  ivector nodes(nne);
  vector dt(nne),tvect(tncomp);
  matrix d(tncomp,tncomp);

  Mt->give_elemnodes (eid,nodes);
  Mb->lc[lcid].tempchanges (dt.a,nodes);

  temp = approx (xi,eta,zeta,dt);
  fillv (temp,tvect);
  
  Mm->matdilat (d,ipp);
  mxv (d,tvect,eps);
  
  Mm->storeeigstrain (ipp,eps);
}

/**
   function computes internal forces

   @param lcid - number of load case
   @param eid - element id
   @param ri,ci - row and column indices
   @param ifor - vector of internal forces
   
   28.7.2001
*/
void linhex::internal_forces (long lcid,long eid,long ri,long ci,vector &ifor)
{
  long i,j,k,l,ii,ipp;
  double xi,eta,zeta,jac;
  ivector nodes(nne),cn(ndofe);
  vector x(nne),y(nne),z(nne),w,gp,r(ndofe),eps(tncomp),sig,contr(ndofe);
  matrix gm;

  Mt->give_node_coord3d (x,y,z,eid);
  Mt->give_code_numbers (eid,cn.a);
  eldispl (lcid,r.a,cn.a,ndofe);
  
  fillv (0.0,ifor);
  
  for (ii=0;ii<nb;ii++){
    if (intordsm[ii][ii]==0)  continue;
    
    allocv (intordsm[ii][ii],gp);
    allocv (intordsm[ii][ii],w);
    allocm (ncomp[ii],ndofe,gm);
    allocv (ncomp[ii],sig);
    
    gauss_points (gp.a,w.a,intordsm[ii][ii]);
    ipp=Mt->elements[eid].ipp[ri+ii][ci+ii];
    
    for (i=0;i<intordsm[ii][ii];i++){
      xi=gp[i];
      for (j=0;j<intordsm[ii][ii];j++){
	eta=gp[j];
	for (k=0;k<intordsm[ii][ii];k++){
	  zeta=gp[k];
	  
	  geom_matrix (gm,x,y,z,xi,eta,zeta,jac);
	  
	  mxv (gm,r,eps);
	  
	  Mm->storestrain (lcid,ipp,eps);
	  
	  Mm->computenlstresses (ipp);
	  
	  Mm->givestress (lcid,ipp,cncomp[ii],ncomp[ii],sig);
	  
	  geom_matrix (gm,x,y,z,xi,eta,zeta,jac);
	  mtxv (gm,sig,contr);
	  
	  cmulv (jac*w[i]*w[j]*w[k],contr);
	  
	  for (l=0;l<contr.n;l++){
	    ifor[l]+=contr[l];
	  }
	  
	  ipp++;
	}
      }
    }
    destrv (sig);  destrm (gm);  destrv (w);  destrv (gp);
  }

}

void linhex::res_internal_forces (long lcid,long eid,vector &ifor)
{
  internal_forces (lcid,eid,0,0,ifor);
}

/**
   function computes internal forces

   @param lcid - number of load case
   @param eid - element id
   @param ri,ci - row and column indices
   @param ifor - vector of internal forces
   
   28.7.2001
*/
void linhex::local_values (long lcid,long eid,long ri,long ci)
{
  long i,j,k,ii,ipp;
  double xi,eta,zeta;
  double **stra;
  vector w,gp,eps(tncomp);
  matrix gm;

  stra = new double* [nne];
  for (i=0;i<nne;i++){
    stra[i] = new double [tncomp];
  }
  elem_strains (stra,lcid,eid,ri,ci);

  for (ii=0;ii<nb;ii++){
    if (intordsm[ii][ii]==0)  continue;

    allocv (intordsm[ii][ii],gp);
    allocv (intordsm[ii][ii],w);
    
    gauss_points (gp.a,w.a,intordsm[ii][ii]);
    ipp=Mt->elements[eid].ipp[ri+ii][ci+ii];
    
    for (i=0;i<intordsm[ii][ii];i++){
      xi=gp[i];
      for (j=0;j<intordsm[ii][ii];j++){
	eta=gp[j];
	for (k=0;k<intordsm[ii][ii];k++){
	  zeta=gp[k];
	  
	  appval (xi,eta,zeta,0,tncomp,eps,stra);
	  
	  Mm->storestrain (lcid,ipp,eps);
	  
	  Mm->computenlstresses (ipp);
	  
	  ipp++;
	}
      }
    }
    destrv (w);  destrv (gp);
  }

  for (i=0;i<nne;i++){
    delete [] stra[i];
  }
  delete [] stra;
}


/**
   function computes internal forces

   @param lcid - number of load case
   @param eid - element id
   @param ri,ci - row and column indices
   @param ifor - vector of internal forces
   
   28.7.2001
*/
void linhex::nonloc_internal_forces (long lcid,long eid,long ri,long ci,vector &ifor)
{
  long i,j,k,l,ii,ipp;
  double xi,eta,zeta,jac;
  vector x(nne),y(nne),z(nne),w,gp,sig,contr(ndofe);
  matrix gm;


  Mt->give_node_coord3d (x,y,z,eid);
  
  fillv (0.0,ifor);

  for (ii=0;ii<nb;ii++){
    if (intordsm[ii][ii]==0)  continue;

    allocv (intordsm[ii][ii],gp);
    allocv (intordsm[ii][ii],w);
    allocm (ncomp[ii],ndofe,gm);
    allocv (ncomp[ii],sig);
    
    gauss_points (gp.a,w.a,intordsm[ii][ii]);
    ipp=Mt->elements[eid].ipp[ri+ii][ci+ii];
    
    for (i=0;i<intordsm[ii][ii];i++){
      xi=gp[i];
      for (j=0;j<intordsm[ii][ii];j++){
	eta=gp[j];
	for (k=0;k<intordsm[ii][ii];k++){
	  zeta=gp[k];
	  
	  Mm->compnonloc_nlstresses (ipp);
	  
	  Mm->givestress (lcid,ipp,cncomp[ii],ncomp[ii],sig);
	  
	  geom_matrix (gm,x,y,z,xi,eta,zeta,jac);
	  mtxv (gm,sig,contr);
	  
	  cmulv (jac*w[i]*w[j]*w[k],contr);
	  
	  for (l=0;l<contr.n;l++){
	    ifor[l]+=contr[l];
	  }
	  
	  ipp++;
	}
      }
    }
    destrv (sig);  destrm (gm);  destrv (w);  destrv (gp);
  }

}


/**
   function returns coordinates of integration points
   
   @param eid - element id
   @param ipp - integration point pointer
   @param coord - vector of coordinates
   
   10.1.2002
*/
void linhex::ipcoord (long eid,long ipp,long ri,long ci,vector &coord)
{
  long i,j,k,ii;
  double xi,eta,zeta;
  vector x(nne),y(nne),z(nne),w(intordsm[ri][ci]),gp(intordsm[ri][ci]);
  
  gauss_points (gp.a,w.a,intordsm[ri][ci]);
  Mt->give_node_coord3d (x,y,z,eid);
  ii=Mt->elements[eid].ipp[ri][ci];
  
  for (i=0;i<intordsm[ri][ci];i++){
    xi=gp[i];
    for (j=0;j<intordsm[ri][ci];j++){
      eta=gp[j];
      for (k=0;k<intordsm[ri][ci];k++){
	zeta=gp[k];
	if (ii==ipp){
	  coord[0]=approx (xi,eta,zeta,x);
	  coord[1]=approx (xi,eta,zeta,y);
	  coord[2]=approx (xi,eta,zeta,z);
	}
	ii++;
      }
    }
  }
}



void linhex::inicipval(long eid, long ri, long ci, matrix &nodval, inictype *ictn)
{
  long i, j, k, l, m, ipp;
  long ii, jj, nv = nodval.n;
  long nstra;
  double xi, eta, zeta, ipval;
  vector w, gp, anv(nne);

  nstra = 0;
  for (j = 0; j < nv; j++) // for all initial values
  {
    for(i = 0; i < nne; i++) // for all nodes on element
      anv[i] = nodval[i][j];
    for (ii = 0; ii < nb; ii++)
    {
      for (jj = 0; jj < nb; jj++)
      {
        ipp=Mt->elements[eid].ipp[ri+ii][ci+jj];
        if (intordsm[ii][jj] == 0)
          continue;
        allocv (intordsm[ii][jj],gp);
        allocv (intordsm[ii][jj],w);
        gauss_points (gp.a,w.a,intordsm[ii][jj]);
        for (k = 0; k < intordsm[ii][jj]; k++)
        {
          xi=gp[k];
          for (l = 0; l < intordsm[ii][jj]; l++)
          {
            eta=gp[l];
            for (m = 0; m < intordsm[ii][jj]; m++)
            {
              zeta=gp[m];
              //  value in integration point
              ipval = approx (xi,eta,zeta,anv);
              if ((ictn[i] & inistrain) && (j < Mm->ip[ipp].ncompstr))
              {
                Mm->ip[ipp].strain[j] += ipval;
                ipp++;
                continue;
              }
              if ((ictn[i] & inistress) && (j < nstra + Mm->ip[ipp].ncompstr))
              {
                Mm->ip[ipp].stress[j] += ipval;
                ipp++;
                continue;
              }
              if ((ictn[i] & iniother) && (j < nv))
              {
                Mm->ip[ipp].other[j] += ipval;
                ipp++;
                continue;
              }
              ipp++;
            }
          }
        }
        destrv (gp);  destrv (w);
      }
    }
    if (ictn[i] & inistrain) nstra++;
  }
}

/**
   function computes volume appropriate to integration point
   
   2.3.2004, JK
*/
void linhex::ipvolume (long eid,long ri,long ci)
{
  long i,j,k,ii,jj,ipp;
  double xi,eta,zeta,jac;
  vector x(nne),y(nne),z(nne),w,gp;
  
  Mt->give_node_coord3d (x,y,z,eid);
  
  for (ii=0;ii<nb;ii++){
    for (jj=0;jj<nb;jj++){
      if (intordsm[ii][jj]==0)  continue;

      allocv (intordsm[ii][jj],w);
      allocv (intordsm[ii][jj],gp);
      
      gauss_points (gp.a,w.a,intordsm[ii][jj]);
      
      ipp=Mt->elements[eid].ipp[ri+ii][ci+jj];
      
      for (i=0;i<intordsm[ii][jj];i++){
	xi=gp[i];
	for (j=0;j<intordsm[ii][jj];j++){
	  eta=gp[j];
	  for (k=0;k<intordsm[ii][jj];k++){
	    zeta=gp[k];
	    
	    jac_3d (jac,x,y,z,xi,eta,zeta);
	    jac=fabs(jac);
	    
	    jac*=w[i]*w[j]*w[k];
	    
	    Mm->storeipvol (ipp,jac);
	    ipp++;
	  }
	}
      }
      destrv (gp);  destrv (w);
    }
  }
  
}
