#include <string.h>
#include <stdlib.h>
#include "outdriverm.h"
#include "iotools.h"
#include "galias.h"
#include "global.h"
#include "probdesc.h"
#include "mechbclc.h"
#include "mechtop.h"
#include "mechprint.h"
#include "node.h"
#include "element.h"
#include "globmat.h"
#include "intpoints.h"
#include "vecttens.h"
#include <math.h>


static int prdisp = 15;
static int prstra = 15;
static int prstre = 15;
static int proth  = 15;
static int prreac = 15;

/**
  Constructor initializes data members to zero values.
*/
outdriverm::outdriverm()
{
  memset(outfn, 0, sizeof(*outfn)*FNAMELEN);
  memset(outdiagfn, 0, sizeof(*outdiagfn)*FNAMELEN);
  memset(outgrfn, 0, sizeof(*outgrfn)*FNAMELEN);
  outf = outgr = NULL;
  outdiagf = NULL;
  textout = off;
  nlcs = 0;
  nclcs = NULL;
  bvlcs = NULL;
  gf = grfmt_no;
  ncut = 0;
  idn1 = 1;
  ide1 = 1;
  ndiag = 0;
  odiag = NULL;
  vtk_num = 0;
}



/**
   Destructor deallocates used memory.
*/
outdriverm::~outdriverm()
{
  long i;

  delete [] outdiagf;
  delete [] nclcs;
  for (i=0; i<nlcs; i++)
    delete [] bvlcs[i];
  delete [] odiag;
}



/**
  Function reads description of required output from the opened text file in.

  @param in - pointer to opened text file

  @retval 0 - on success
  @retval 1 - error reading output text filename
  @retval 2 - error reading nodal output description
  @retval 3 - error reading element output description
  @retval 4 - error reading user defined point output description
  @retval 5 - error reading output graphics format
  @retval 6 - error reading nodal graphics output description
  @retval 7 - error reading element graphics output description
  @retval 8 - error reading output diagram description
*/
long outdriverm::read(XFILE *in)
{
  long i;
  const char *str;

  xfscanf(in, "%k%m", "textout", &flagsw_kwdset, &textout);
  if (textout)
  {
    if (xfscanf(in," %a", outfn) != 1)
    {
      fprintf(stderr, "\n\nError reading filename for output\n");
      fprintf(stderr, " function outdriverm::read (file %s, line %d)\n", __FILE__, __LINE__);
      return 1;
    }
    fprintf(stdout, "\n Output file name (mechanics): %s", outfn);

    if (no.read(in))
      return 2;
    if (no.dstep.st == sel_no)
      str = " not";
    else
      str = "";
    fprintf(stdout, "\n Output of mechanical nodal values will%s be performed", str);

    if (eo.read(in))
      return 3;
    if (eo.dstep.st == sel_no)
      str = " not";
    else
      str = "";
    fprintf(stdout, "\n Output of mechanical element values will%s be performed", str);
    if (po.read(in))
      return 4;
    if (po.dstep.st == sel_no)
      str = " not";
    else
      str = "";
    fprintf(stdout, "\n Output of mechanical UDP values will%s be performed", str);
  }
  if (xfscanf(in, "%k%m", "outgr_format", &graphfmt_kwdset, (int *)&gf) != 2)
  {
    fprintf(stderr, "\n\nError reading type of grahics output\n");
    fprintf(stderr, " function outdriverm::read (file %s, line %d)\n", __FILE__, __LINE__);
    return 5;
  }
  if (gf == grfmt_open_dx)
  {
    Mp->straincomp = 1;
    Mp->strainaver = 1;
    Mp->stresscomp = 1;
    Mp->stressaver = 1;
    Mp->othercomp = 1;
    Mp->otheraver = 1;
    Mp->reactcomp = 1;
  }
  if (gf != grfmt_no)
  {
    if (xfscanf(in, " %a", outgrfn) != 1)
    {
      fprintf(stderr, "\n\nError reading filename for graphics output\n");
      fprintf(stderr, " function outdriverm::read (file %s, line %d)\n", __FILE__, __LINE__);
      return 5;
    }
  }
  switch (gf)
  {
    case grfmt_no:
      fprintf(stdout, "\n Graphic output will not be performed");
      break;
    case grfmt_open_dx:
      fprintf(stdout, "\n Graphic output will be in OpenDX format");
      fprintf(stdout, "\n Filename of graphic output : %s", outgrfn);
      break;
    case grfmt_femcad:
      fprintf(stdout, "\n Graphic output will be in FemCAD format");
      fprintf(stdout, "\n Filename of graphic output : %s", outgrfn);
      break;
    case grfmt_gid:
      fprintf(stdout, "\n Graphic output will be in GiD format");
      fprintf(stdout, "\n Filename of graphic output : %s", outgrfn);
      break;
    case grfmt_gid_sep:
      fprintf(stdout, "\n Graphic output will be in GiD format separated to quantities");
      fprintf(stdout, "\n Filename of graphic output : %s", outgrfn);
      break;
   case grfmt_vtk:
      fprintf(stdout, "\n Graphic output will be in VTK format");
      fprintf(stdout, "\n Filename of graphic output : %s", outgrfn);
      break;
  }

  if (gf != grfmt_no)
  {
    if (nog.read(in))
      return 6;
    if (nog.dstep.st == sel_no)
      str = " not";
    else
      str = "";
    fprintf(stdout, "\n Graphical output of mechanical nodal values will%s be performed", str);
    
    if (eog.read(in))
      return 7;
    if (eog.dstep.st == sel_no)
      str = " not";
    else
      str = "";
    fprintf(stdout, "\n Graphical output of mechanical element values will%s be performed", str);
  }
  switch (Mp->tprob)
  {
  case lin_floating_subdomain:
    case mat_nonlinear_statics:
    case forced_dynamics:
    case mech_timedependent_prob:
    case nonlin_floating_subdomain:
    case growing_mech_structure:
    case earth_pressure:
      if (xfscanf(in, "%k%ld", "numdiag", &ndiag) != 2)
      {
        print_err("cannot read number of diagrams", __FILE__, __LINE__, __func__);
        return 8;
      }
      if (ndiag == 0)
      {
        fprintf (stdout, "\n Number of diagrams : %ld", ndiag);
        return 0;
      }
      if (xfscanf(in, " %a", outdiagfn) != 1)
      {
        fprintf(stderr, "\n\nError reading filename for diagrams\n");
        fprintf(stderr, " function outdriverm::read (file %s, line %d)\n", __FILE__, __LINE__);
        return 8;
      }
      fprintf (stdout, "\n Number of diagrams : %ld", ndiag);
      fprintf (stdout, "\n Diagram filename : %s",outdiagfn);
      odiag = new outdiagm[ndiag];
      outdiagf = new FILE* [ndiag];
      memset(outdiagf, 0, sizeof(*outdiagf)*ndiag);
      for (i=0; i<ndiag; i++)
      {
        if (odiag[i].read(in))
          return 8;
      }      
      break;
    default:
      break;
  } 
  return 0;
}


/**
  Function prints data with output description to the text file given by out.
  
  @param out - pointer to opened text file for output    
*/
void outdriverm::print(FILE *out)
{
  long i;
  fprintf(out, "\n#\n# outdriver section\n#\n");
  fprintf(out, "%d\n", (int)textout);
  if (textout == on)
  {
    fprintf(out, "%s\n", outfn);
    no.print(out);
    eo.print(out);
    po.print(out);
  }
  fprintf(out, "%d\n", (int)gf);
  if (gf != grfmt_no)
  {
    fprintf(out, "%s\n", outgrfn);
    nog.print(out);
    eog.print(out);
  }
  switch (Mp->tprob)
    {
    case lin_floating_subdomain:
    case mat_nonlinear_statics:
    case forced_dynamics:
    case mech_timedependent_prob:
    case growing_mech_structure:
    case nonlin_floating_subdomain:
    case earth_pressure:
      fprintf(out, "%ld\n", ndiag);
      if (ndiag == 0)
        break;
      fprintf(out, "%s\n", outdiagfn);
      for (i=0; i<ndiag; i++)
        odiag[i].print(out);
      break;
    default:
      break;
  } 
}



/**
  Function prints header to the output text file.
  
  @param out - pointer to the opened text output file
*/
void outdriverm::print_header(FILE *out)
{
  fprintf(out, "%15s ****  *  ****  ****  *\n", " ");
  fprintf(out, "%15s *     *  *     *     *\n", " ");
  fprintf(out, "%15s  *    *  ***   ***   *\n", " ");
  fprintf(out, "%15s   *   *  *     *     *\n", " ");
  fprintf(out, "%15s****   *  *     ****  ****  MEFEL OUTPUT\n", " ");

/*
  long i;
  for (i=0; i<72; i++)
    fprintf(out, "*");
  fprintf(out, "\n\n*%15s SIFEL - MEFEL OUTPUT %15s*\n\n", " ", " ");
  for (i=0; i<72; i++)
    fprintf(out, "*");*/
  fprintf(out, "\n%s\n", Mp->name);
  fprintf(out, "\n\n\n\n\n");
}



/**
  Function prints step number to the output text file.
  
  @param out  - pointer to the opened text output file
  @param step - integer step id
  @param time - time or load step
*/
void outdriverm::print_newstep(FILE *out, long step, double time)
{
  long i;
  double lambda;

  if (textout)
  {
    if(Mp->tpr == seconds)
      lambda = time;
    if(Mp->tpr == minutes)
      lambda = time/60.0; 
    if(Mp->tpr == hours)
      lambda = time/3600.0;
    if(Mp->tpr == days)
      lambda = time/86400.0;
    
    for (i=0; i<53; i++)
      fprintf(out, "*");
    fprintf(out, "\n%10sStep number=% ld, time/load step=% g\n", " ", step, lambda);
    for (i=0; i<53; i++)
      fprintf(out, "*");
    fprintf(out, "\n\n\n\n");
  }
}


/**
  Function prints required output values to the output text file.
  
  @param out - pointer to the opened text file 
  @param lcid - load case id
  @param istep - step id
  
*/
void outdriverm::print_out(FILE *out, long lcid, long istep, double time)
{
  if (textout)
  {
    fprintf(out, "* LOAD CASE NUMBER %ld\n\n\n", lcid+1);
  
    if (no.sellc.presence_id(lcid) && (no.dstep.presence_id(istep, time)))
      no.print_out(out, lcid);
    if (eo.sellc.presence_id(lcid) && (eo.dstep.presence_id(istep, time)))
      eo.print_out(out, lcid);
    if (po.sellc.presence_id(lcid) && (po.dstep.presence_id(istep, time))) 
      po.print_out(out, lcid);
  }
}



/**
  Function prints required output values to the output text file without respect to
  to selected step/time (forced printing).
  
  @param out - pointer to the opened text file 
  @param lcid - load case id
  @param istep - step id
  
*/
void outdriverm::print_out_forced(FILE *out, long lcid, long istep, double time)
{
  if (textout)
  {
    fprintf(out, "* LOAD CASE NUMBER %ld\n\n\n", lcid+1);
  
    if (no.sellc.presence_id(lcid))
      no.print_out(out, lcid);
    if (eo.sellc.presence_id(lcid))
      eo.print_out(out, lcid);
    if (po.sellc.presence_id(lcid)) 
      po.print_out(out, lcid);
  }
}



/**
  Function prints diagrams.
  
  @param lcid - required load case id
  @param lambda - actual load parameter/actual time
  @param istep - actual step id
  @param fi - array with values of load vector
  
*/
void outdriverm::print_diags(long lcid, double lambda, long istep, double *fi)
{
  long i;
  for(i=0; i<ndiag; i++)
    odiag[i].printval(outdiagf[i], lcid, lambda, istep, fi);
}



/**
  Function prints diagrams without respect to
  to selected step/time (forced printing).
  
  @param lcid - required load case id
  @param lambda - actual load parameter/actual time
  @param istep - actual step id
  @param fi - array with values of load vector
  
*/
void outdriverm::print_diags_forced(long lcid, double lambda, long istep, double *fi)
{
  long i;
  for(i=0; i<ndiag; i++)
    odiag[i].printval_forced(outdiagf[i], lcid, lambda, istep, fi);
}



/**
  Function prints required value to the graphics files .
  
  @param out - pointer to the opened text file 
  @param lcid - required load case id
  @param lambda - actual load parameter/actual time
  @param istep - actual step id
  @param fi - array with values of load vector
  
*/
void outdriverm::print_graphics(FILE *out, long lcid, double lambda, long istep, double *fi)
{
  long i;
  char dlcid[50];

  switch (gf)
  {
    case grfmt_no:
      break;;
    case grfmt_femcad:
      break;
    case grfmt_open_dx:
    {
      for (i=0; i<Mb->nlc; i++)
      {
        if (gf == grfmt_open_dx)
          print_default_dx (Gtm, Mp, Mt, Mm, i, outgrfn);
      }
      break;
    }
    case grfmt_gid:
    {
      if (nog.sellc.presence_id(lcid) && (nog.dstep.presence_id(istep, lambda, Mp->timecon)))
      {
  switch (Mp->tprob){
        case mech_timedependent_prob:
  case growing_mech_structure:
    if(Mp->tpr == seconds)
      sprintf(dlcid, "%e", lambda);
    if(Mp->tpr == minutes)
      sprintf(dlcid, "%e", lambda/60.0); 
    if(Mp->tpr == hours)
      sprintf(dlcid, "%e", lambda/3600.0);
    if(Mp->tpr == days)
      sprintf(dlcid, "%e", lambda/86400.0);
          nog.print_graphics(out, lcid, dlcid, gf, fi);
    break;
  default:{
    sprintf(dlcid, "%ld", istep);
    nog.print_graphics(out, lcid, dlcid, gf, fi);
  }
  }
      }
      if (eog.sellc.presence_id(lcid) && (eog.dstep.presence_id(istep, lambda, Mp->timecon)))
  {
    switch (Mp->tprob){
    case mech_timedependent_prob:
    case growing_mech_structure:
      if(Mp->tpr == seconds)
        sprintf(dlcid, "%e", lambda);
      if(Mp->tpr == minutes)
        sprintf(dlcid, "%e", lambda/60.0); 
      if(Mp->tpr == hours)
        sprintf(dlcid, "%e", lambda/3600.0);
      if(Mp->tpr == days)
        sprintf(dlcid, "%e", lambda/86400.0);
      eog.print_graphics(out, lcid, dlcid, gf, ide1);
      break;
    default:{
      sprintf(dlcid, "%ld", istep);
      eog.print_graphics(out, lcid, dlcid, gf, ide1);
    }
    }
      }
      break;
    }
    case grfmt_gid_sep:
    {      
      if (nog.sellc.presence_id(lcid) && (nog.dstep.presence_id(istep, lambda, Mp->timecon)))
      {
        switch (Mp->tprob){
        case mech_timedependent_prob:
  case growing_mech_structure:
    if(Mp->tpr == seconds)
      sprintf(dlcid, "%e", lambda);
    if(Mp->tpr == minutes)
      sprintf(dlcid, "%e", lambda/60.0); 
    if(Mp->tpr == hours)
      sprintf(dlcid, "%e", lambda/3600.0);
    if(Mp->tpr == days)
      sprintf(dlcid, "%e", lambda/86400.0);
          nog.print_graphics(outgrfngs, "at", lcid, dlcid, gf, fi);
    break;
  default:{
    sprintf(dlcid, "%ld", istep);
    nog.print_graphics(outgrfngs, "at", lcid, dlcid, gf, fi);
  }
  }
      }
      if (eog.sellc.presence_id(lcid) && (eog.dstep.presence_id(istep, lambda, Mp->timecon)))
      {
        switch (Mp->tprob){
        case mech_timedependent_prob:
  case growing_mech_structure:
    if(Mp->tpr == seconds)
      sprintf(dlcid, "%e", lambda);
    if(Mp->tpr == minutes)
      sprintf(dlcid, "%e", lambda/60.0); 
    if(Mp->tpr == hours)
      sprintf(dlcid, "%e", lambda/3600.0);
    if(Mp->tpr == days)
      sprintf(dlcid, "%e", lambda/86400.0);
          eog.print_graphics(outgrfngs, "at", lcid, dlcid, gf, ide1);
    break;
  default:{
    sprintf(dlcid, "%ld", istep);
    eog.print_graphics(outgrfngs, "at", lcid, dlcid, gf, ide1);
  }
  }
     }
      break;
    }
    case grfmt_vtk://not implemented for growing structure
      {
        if (nog.sellc.presence_id(lcid) && (nog.dstep.presence_id(istep, lambda, Mp->timecon)))//need to include eog separately
        {
          char fname[FNAMELEN+70];
          FILE *vtk_file;
          char *path;
          char *name;
          char *suffix;

          Mp->filename_decomposition (outgrfn,path,name,suffix);
          sprintf(fname,"%s%s%04ld%s", path, name, istep, suffix);//increase index in filename
          delete [] path; delete [] name; delete [] suffix;

          if((vtk_file = fopen(fname, "w"))==NULL){
            fprintf(stderr, "\n\nUnable to open graphics file '%s' in function print_graphics()\n", fname);
            fprintf(stderr, " (file %s, line %d)\n", __FILE__, __LINE__);
            abort();
          }
          switch (Mp->tprob){
            case mech_timedependent_prob:
              if(Mp->tpr == seconds)
                sprintf(dlcid, "%e", lambda);
              if(Mp->tpr == minutes)
                sprintf(dlcid, "%e", lambda/60.0);
              if(Mp->tpr == hours)
                sprintf(dlcid, "%e", lambda/3600.0);
              if(Mp->tpr == days)
                sprintf(dlcid, "%e", lambda/86400.0);
              print_default_vtk(vtk_file, lcid, istep, (double) atof(dlcid));//print header,points,nodes,elements
              write_vtk_unkn(vtk_file, lcid, dlcid);//print datapoints in sections
              fclose(vtk_file);
              vtk_num++;
            break;
            default:
            {
            //not implemented yet
            }
          }
        }
      break;
      }
   default:
    {
      fprintf(stderr, "\nUnknown type of graphics format is required in function\n");
      fprintf(stderr, "outdriverm::print_graphics (file %s, line %d)\n", __FILE__, __LINE__);
    }
  }
}



/**
  Function prints required value to the graphics files without respect to
  to selected step/time (forced printing).
  
  @param out - pointer to the opened text file 
  @param lcid - required load case id
  @param lambda - actual load parameter/actual time
  @param istep - actual step id
  @param fi - array with values of load vector
  
*/
void outdriverm::print_graphics_forced(FILE *out, long lcid, double lambda, long istep, double *fi)
{
  long i;
  char dlcid[50];

  switch (gf)
  {
    case grfmt_no:
      break;;
    case grfmt_femcad:
      break;
    case grfmt_open_dx:
    {
      for (i=0; i<Mb->nlc; i++)
      {
        if (gf == grfmt_open_dx)
          print_default_dx (Gtm, Mp, Mt, Mm, i, outgrfn);
      }
      break;
    }
    case grfmt_gid:
    {
      if (nog.sellc.presence_id(lcid))
      {
  switch (Mp->tprob){
        case mech_timedependent_prob:
  case growing_mech_structure:
    if(Mp->tpr == seconds)
      sprintf(dlcid, "%e", lambda);
    if(Mp->tpr == minutes)
      sprintf(dlcid, "%e", lambda/60.0); 
    if(Mp->tpr == hours)
      sprintf(dlcid, "%e", lambda/3600.0);
    if(Mp->tpr == days)
      sprintf(dlcid, "%e", lambda/86400.0);
          nog.print_graphics(out, lcid, dlcid, gf, fi);
    break;
  default:{
    sprintf(dlcid, "%ld", istep);
    nog.print_graphics(out, lcid, dlcid, gf, fi);
  }
  }
      }
      if (eog.sellc.presence_id(lcid))
  {
    switch (Mp->tprob){
    case mech_timedependent_prob:
    case growing_mech_structure:
      if(Mp->tpr == seconds)
        sprintf(dlcid, "%e", lambda);
      if(Mp->tpr == minutes)
        sprintf(dlcid, "%e", lambda/60.0); 
      if(Mp->tpr == hours)
        sprintf(dlcid, "%e", lambda/3600.0);
      if(Mp->tpr == days)
        sprintf(dlcid, "%e", lambda/86400.0);
      eog.print_graphics(out, lcid, dlcid, gf, ide1);
      break;
    default:{
      sprintf(dlcid, "%ld", istep);
      eog.print_graphics(out, lcid, dlcid, gf, ide1);
    }
    }
      }
      break;
    }
    case grfmt_gid_sep:
    {      
      if (nog.sellc.presence_id(lcid))
      {
        switch (Mp->tprob){
        case mech_timedependent_prob:
  case growing_mech_structure:
    if(Mp->tpr == seconds)
      sprintf(dlcid, "%e", lambda);
    if(Mp->tpr == minutes)
      sprintf(dlcid, "%e", lambda/60.0); 
    if(Mp->tpr == hours)
      sprintf(dlcid, "%e", lambda/3600.0);
    if(Mp->tpr == days)
      sprintf(dlcid, "%e", lambda/86400.0);
          nog.print_graphics(outgrfngs, "at", lcid, dlcid, gf, fi);
    break;
  default:{
    sprintf(dlcid, "%ld", istep);
    nog.print_graphics(outgrfngs, "at", lcid, dlcid, gf, fi);
  }
  }
      }
      if (eog.sellc.presence_id(lcid))
      {
        switch (Mp->tprob){
        case mech_timedependent_prob:
  case growing_mech_structure:
    if(Mp->tpr == seconds)
      sprintf(dlcid, "%e", lambda);
    if(Mp->tpr == minutes)
      sprintf(dlcid, "%e", lambda/60.0); 
    if(Mp->tpr == hours)
      sprintf(dlcid, "%e", lambda/3600.0);
    if(Mp->tpr == days)
      sprintf(dlcid, "%e", lambda/86400.0);
          eog.print_graphics(outgrfngs, "at", lcid, dlcid, gf, ide1);
    break;
  default:{
    sprintf(dlcid, "%ld", istep);
    eog.print_graphics(outgrfngs, "at", lcid, dlcid, gf, ide1);
  }
  }
     }
      break;
    }
    default:
    {
      fprintf(stderr, "\nUnknown type of graphics format is required in function\n");
      fprintf(stderr, "outdriverm::print_graphics (file %s, line %d)\n", __FILE__, __LINE__);
    }
  }
}



/**
  Function creates output graphics files and their headers for GiD separated format (grfmt_gidsp).
  For this format (grfmt_gidsp) each required quantity is printed to the sparated file named 
  in following way : filename.{elem|nodal}_{quanitity}{indexOfQuantity}.res
  quatnitity can be : eps, sig, pesp, psig, other.
  indexOfQuantity can be : 
    * index number for scalars
    * _v{firstIndex}-{numCompVec}_s{selId} for vectors, where 
      - firstIndex is the index of the first vector component in the given array (eps, sig, other)
      - numCompVec is the number of vector component
      - selId is the index of selection for nodes/elements which the selected vector belongs to
    * _m[{firstIndex}-{numCompMtx}]_s{selId} for tensors/matrices, where
      - firstIndex is the index of the first tensor/matrix component in the other array 
      - numCompMtx is the number of tensor/matrix component in the other array
      - selId is the index of selection for nodes/elements which the selected tensor/matrix belongs to

  @param mode - opening mode for opened text files

*/
void outdriverm::create_files_gidsp(const char *mode)
{
  char dlcid[50];

  if (gf != grfmt_gid_sep)
  {
    fprintf(stderr, "\nWrong graphics format is required in function outdriverm::create_files_gidsp\n");
    fprintf(stderr, "(file %s, line %d)\n", __FILE__, __LINE__);
    abort();
  }
  if (nog.dstep.st != sel_no)
    nog.print_graphics(outgrfngs, mode, 0, dlcid, gf, NULL);
  if (eog.dstep.st != sel_no)
    eog.print_graphics(outgrfngs, mode, 0, dlcid, gf, ide1);
}







/**
  Constructor initializes data to zero values
*/
nodeoutm::nodeoutm()
{
  react = 0;
  seldisp = selstra = selstre = seloth = NULL;
  transtra = transtre = NULL; 
}



/**
  Destructor deallocates used memory
*/
nodeoutm::~nodeoutm()
{
  delete [] seldisp;
  delete [] selstra;
  delete [] selstre; 
  delete [] seloth;
  delete [] transtra;
  delete [] transtre;
}



/**
   Function reads data with description for output of nodal values for the text output file.

   @param in - pointer to the opened text file

   @retval 0 - on success
   @retval 1 - error reading step or load case selection
   @retval 2 - error reading displacement selection
   @retval 3 - error reading strain selection
   @retval 4 - error reading stress selection
   @retval 5 - error reading other values selection
   @retval 6 - error reading reaction output flag
*/
long nodeoutm::read(XFILE *in)
{
  long i;
  // step and loadcases
  xfscanf(in, "%k", "sel_nodstep");
  dstep.read(in);
  if (dstep.st == sel_no)
    return 0;
  xfscanf(in, "%k", "sel_nodlc");
  sellc.read(in);

  // displacements
  xfscanf(in, "%k", "displ_nodes");
  selndisp.read(in);
  switch (selndisp.st)
  {
    case sel_no:
      break;
    case sel_all:
    case sel_range:
    case sel_list:
      xfscanf(in, "%k", "displ_comp");
      seldisp = new sel[selndisp.n];
      for (i=0; i<selndisp.n; i++)
        seldisp[i].read(in);
      break;
    default:
      fprintf(stderr, "\n\nUnknown type of selection is required in function nodeoutm::read\n");
      fprintf(stderr, " (file %s, line %d)", __FILE__, __LINE__);
      return 2;
  }

  // strains
  xfscanf(in, "%k", "strain_nodes");
  selnstra.read(in);
  switch (selnstra.st)
  {
    case sel_no:
      break;
    case sel_all:
    case sel_range:
    case sel_list:
      xfscanf(in, "%k", "nodstrain_comp");
      Mp->straincomp = 1;
      Mp->strainaver = 1;
      transtra = new long[selnstra.n];
      memset(transtra, 0, sizeof(*transtra)*selnstra.n);
      selstra = new sel[selnstra.n];
      for(i=0; i<selnstra.n; i++)
        selstra[i].read(in);
      xfscanf(in, "%k", "nodstra_transfid");
      for(i=0; i<selnstra.n; i++)
      {
        if (xfscanf(in, "%ld", transtra+i) != 1)
        {
          fprintf(stderr, "\n\nError reading strain selection\n");
          fprintf(stderr, " in function nodeoutm::read, (file %s, line %d)\n", __FILE__, __LINE__);
          return 3;
        }
      }
      break;
    default:
      fprintf(stderr, "\n\nUnknown type of selection is required in function nodeoutm::read\n");
      fprintf(stderr, " (file %s, line %d)", __FILE__, __LINE__);
      return 3;
  }


  // stresses
  xfscanf(in, "%k", "stress_nodes");
  selnstre.read(in);
  switch (selnstre.st)
  {
    case sel_no:
      break;
    case sel_all:
    case sel_range:
    case sel_list:
      xfscanf(in, "%k", "nodstress_comp");
      Mp->straincomp = 1;
      Mp->strainaver = 1;
      Mp->stresscomp = 1;
      Mp->stressaver = 1;
      transtre = new long[selnstre.n];
      memset(transtre, 0, sizeof(*transtre)*selnstre.n);
      selstre = new sel[selnstre.n];
      for(i=0; i<selnstre.n; i++)
        selstre[i].read(in);
      xfscanf(in, "%k", "nodstre_transfid");
      for(i=0; i<selnstre.n; i++)
      {
        if (xfscanf(in, "%ld", transtre+i) != 1)
        {
          fprintf(stderr, "\n\nError reading stress selection\n");
          fprintf(stderr, " in function nodeoutm::read, (file %s, line %d)\n", __FILE__, __LINE__);
          return 3;
        }
      }
      break;
    default:
      fprintf(stderr, "\n\nUnknown type of selection is required in function nodeoutm::read\n");
      fprintf(stderr, " (file %s, line %d)", __FILE__, __LINE__);
      return 4;
  }

  // other values
  xfscanf(in, "%k", "other_nodes");
  selnoth.read(in);
  switch (selnoth.st)
  {
    case sel_no:
      break;
    case sel_all:
    case sel_range:
    case sel_list:
      xfscanf(in, "%k", "nodother_comp");
      if(Mp->othercomp == 0)
  Mp->othercomp = 1;
      if(Mp->otheraver == 0)
  Mp->otheraver = 1;
      seloth = new sel[selnoth.n];
      for (i=0; i < selnoth.n; i++)
        seloth[i].read(in);
      break; 
    default:
      fprintf(stderr, "\n\nUnknown type of selection is required in function nodeoutm::read\n");
      fprintf(stderr, " (file %s, line %d)", __FILE__, __LINE__);
      return 5;
  }

  // reactions
  if (xfscanf(in, "%k%ld", "reactions", &react) != 2)
  {
    fprintf(stderr, "\n\nError reading reactions selection\n");
    fprintf(stderr, " in function nodeoutm::read, (file %s, line %d)\n", __FILE__, __LINE__);
    return 6;
  }
  if (react)
    Mp->reactcomp = 1;
  return 0;
}



/**
  Function prints data with description for output of nodal values to the text file.

  @param out - pointer to the opened text file
*/
void nodeoutm::print(FILE *out)
{
  long i;

  // step and loadcases
  dstep.print(out);
  if (dstep.st == sel_no)
    return;
  sellc.print(out);
  fprintf(out, "\n");

  // displacements
  selndisp.print(out);
  switch(selndisp.st)
  {
    case sel_no:
      break;
    case sel_all:
    case sel_range:
    case sel_list:
      for(i=0; i<selndisp.n; i++)
        seldisp[i].print(out);
      break;
    default:
      fprintf(stderr, "\n\nUnknown type of selection is required in function nodeoutm::print\n");
      fprintf(stderr, " (file %s, line %d)", __FILE__, __LINE__);
      return;
  }


  // strains
  selnstra.print(out);
  switch (selnstra.st)
  {
    case sel_no:
      break;
    case sel_all:
    case sel_range:
    case sel_list:
      for(i=0; i<selnstra.n; i++)
        selstra[i].print(out);
      for(i=0; i<selnstra.n; i++)
        fprintf(out, "%ld ", transtra[i]);
      fprintf(out, "\n");
      break;
    default:
      fprintf(stderr, "\n\nUnknown type of selection is required in function nodeoutm::print\n");
      fprintf(stderr, " (file %s, line %d)", __FILE__, __LINE__);
      return;
  }

  // stresses
  selnstre.print(out);
  switch (selnstre.st)
  {
    case sel_no:
      break;
    case sel_all:
    case sel_range:
    case sel_list:
      for(i=0; i<selnstre.n; i++)
        selstre[i].print(out);
      for(i=0; i<selnstre.n; i++)
        fprintf(out, "%ld ", transtre[i]);
      fprintf(out, "\n");
      break;
    default:
      fprintf(stderr, "\n\nUnknown type of selection is required in function nodeoutm::print\n");
      fprintf(stderr, " (file %s, line %d)", __FILE__, __LINE__);
      return;
  }

  // other values
  selnoth.print(out);
  switch (selnoth.st)
  {
    case sel_no:
      break;
    case sel_all:
    case sel_range:
    case sel_list:
      for(i=0; i<selnoth.n; i++)
        seloth[i].print(out);
      break;
    default:
      fprintf(stderr, "\n\nUnknown type of selection is required in function nodeoutm::print\n");
      fprintf(stderr, " (file %s, line %d)", __FILE__, __LINE__);
      return;
  }

  // reactions
  fprintf(out, "%ld\n", react);
}


/**
  Function prints required output values for selected nodes and for given load case
  to the output text file.
  
  @param out  - pointer to the opened text file
  @param lcid - load case id

 */
void nodeoutm::print_out(FILE *out, long lcid)
{
  if (selndisp.st != sel_no)
  {
    fprintf(out, "** Nodal displacements :\n\n");
    print_disp(out, lcid);
  }
  if (selnstra.st != sel_no)
  {
    fprintf(out, "** Nodal averaged strains :\n\n");
    print_stra(out, lcid);
  }
  if (selnstre.st != sel_no)
  {
    fprintf(out, "** Nodal averaged stresses :\n\n");
    print_stre(out, lcid);
  }
  if (selnoth.st != sel_no)
  {
    fprintf(out, "** Nodal averaged other values :\n\n");
    print_other(out);
  }
  if (react)
  {
    fprintf(out, "** Reactions :\n\n");
    print_react(out, lcid);
  }
}

/**
  Function prints required displacements for selected nodes and for given load case
  to the output text file.
  
  @param out  - pointer to the opened text file
  @param lcid - load case id

 */
void nodeoutm::print_disp(FILE *out, long lcid)
{
  long i, j, ndofn;
  double *r;

  for (i=0; i<Mt->nn; i++)
  {
    if (selndisp.presence_id(i))
    {
      fprintf(out, " Node %7ld", i+1);
      ndofn = Mt->give_ndofn(i);
      r = new double[ndofn];
      noddispl(lcid, r, i);
      for (j=0; j<ndofn; j++)
      {
        if (selndisp.presence_id(seldisp,i,j))
          fprintf(out, "   r_%ld=% .*e", j+1, prdisp, r[j]);
      }
      delete [] r;
      fprintf(out, "\n");
    } 
  }
  fprintf(out, "\n");
}



/**
  Function prints required strains for selected nodes and for given load case
  to the output text file.
  
  @param out  - pointer to the opened text file
  @param lcid - load case id

 */
void nodeoutm::print_stra(FILE *out, long lcid)
{
  long i, j, ncomp, id, ir;

  for (i=0; i<Mt->nn; i++)
  {
    if (selnstra.presence_id(i, ir))
    {
      fprintf(out, " Node %7ld", i+1);
      ncomp = Mt->nodes[i].ncompstr;
      id = lcid*ncomp;
      for (j=0; j<ncomp; j++)
      {
        if (selnstra.presence_id(selstra,i,j))
          fprintf(out, "   eps_%ld=% .*e", j+1, prstra, Mt->nodes[i].strain[id+j]);
      }
      if (transtra[ir] < 0)
      {
        vector eps(4), peps(3); 
        matrix epst(3,3), pvect(3,3);
        fillv(0.0, eps);
        for (j=0; j<ncomp; j++)
          eps[j] = Mt->nodes[i].strain[j];
        if (ncomp == 4)
          vector_tensor (eps, epst, planestrain, strain);
        if (ncomp == 6)
          vector_tensor (eps, epst, spacestress, strain);
        princ_val (epst, peps, pvect, 20, 1.0e-4, Mp->zero, 3,1);
        if (Mt->nodes[i].pstra == NULL)
          Mt->nodes[i].pstra = new double [9];
        for (j=0; j<3; j++)
        {
          Mt->nodes[i].pstra[j] = peps[j];
          fprintf(out, "   peps_%ld=% .*e", j+1, prstra, peps[j]);
        }
        for (j=0; j<3; j++)
        {
          Mt->nodes[i].pstra[3+j] = pvect[0][j];
          fprintf(out, "   a_1%ld=% .*e", j+1, prstra, pvect[0][j]);
        }
        for (j=0; j<3; j++)
        {
          Mt->nodes[i].pstra[6+j] = pvect[2][j];
          fprintf(out, "   a_3%ld=% .*e", j+1, prstra, pvect[2][j]);
        }  
      }
      fprintf(out, "\n");
    } 
  }
  fprintf(out, "\n");
}



/**
  Function prints required stresses for selected nodes and for given load case
  to the output text file.
  
  @param out  - pointer to the opened text file
  @param lcid - load case id

 */
void nodeoutm::print_stre(FILE *out, long lcid)
{
  long i, ir, j, ncomp, id;

  for (i=0; i<Mt->nn; i++)
  {
    if (selnstre.presence_id(i, ir))
    {
      fprintf(out, " Node %7ld", i+1);
      ncomp = Mt->nodes[i].ncompstr;
      id = lcid*ncomp;
      for (j=0; j<ncomp; j++)
      {
        if (selnstre.presence_id(selstre, i, j))
          fprintf(out, "   sig_%ld=% .*e", j+1, prstre, Mt->nodes[i].stress[id+j]);
      }
      if (transtre[ir] < 0)
      {
        vector sig(ncomp), psig(3); 
        matrix sigt(3,3), pvect(3,3);
        fillv(0.0, sig);
        for (j=0; j<ncomp; j++)
          sig[j] = Mt->nodes[i].stress[j];
        if (ncomp == 4)
          vector_tensor (sig, sigt, planestrain, stress);
        if (ncomp == 6)
          vector_tensor (sig, sigt, spacestress, stress);
        
        princ_val (sigt, psig, pvect, 20, 1.0e-4, Mp->zero, 3,1);
        if (Mt->nodes[i].pstre == NULL)
          Mt->nodes[i].pstre = new double [9];
        for (j=0; j<3; j++)
        {
          Mt->nodes[i].pstre[j] = psig[j];
          fprintf(out, "   psig_%ld=% .*e", j+1, prstre, psig[j]);
        }
        fprintf(out, "   tau_max=% .*e", prstre, (psig[2]-psig[0])/2);
        for (j=0; j<3; j++)
        {
          Mt->nodes[i].pstre[3+j] = pvect[0][j];
          fprintf(out, "   a_1%ld=% .*e", j+1, prstre, pvect[0][j]);
        }
        for (j=0; j<3; j++)
        {
          Mt->nodes[i].pstre[6+j] = pvect[2][j];
          fprintf(out, "   a_3%ld=% .*e", j+1, prstre, pvect[2][j]);
        }  
      }
      fprintf(out, "\n");
   }
 }
  fprintf(out, "\n");
}



/**
  Function prints required other values for selected nodes and for given load case
  to the output text file.
  
  @param out  - pointer to the opened text file
  @param lcid - load case id

 */
void nodeoutm::print_other(FILE *out)
{
  long i, j, ncomp;

  for (i=0; i<Mt->nn; i++)
  {
    ncomp = Mt->nodes[i].ncompother;
    if (selnoth.presence_id(i))
    {
      fprintf(out, " Node %7ld", i+1);
      ncomp = Mt->nodes[i].ncompother;
      for (j=0; j<ncomp; j++)
      {
        if (selnoth.presence_id(seloth, i, j))
          fprintf(out, "   other_%ld=% .*e", j+1, proth, Mt->nodes[i].other[j]);
      }
      fprintf(out, "\n");
    } 
  }
  fprintf(out, "\n");
}



/**
  Function prints all reactions for for given load case
  to the output text file.
  
  @param out  - pointer to the opened text file
  @param lcid - load case id

 */
void nodeoutm::print_react(FILE *out, long lcid)
{
  long i, j, ndofn;

  for (i=0; i<Mt->nn; i++)
  {
    if (Mt->nodes[i].react)
    {
      fprintf(out, " Node %7ld", i+1);
      ndofn = Mt->give_ndofn(i);
      for (j=0; j<ndofn; j++)
        fprintf(out, "   R_%ld=% .*e", j+1, prreac, Mt->nodes[i].r[lcid*ndofn+j]);
      fprintf(out, "\n");
    } 
  }
  fprintf(out, "\n");
}








/**
  Constructor initializes data to zero values
*/
nodeoutgm::nodeoutgm()
{
  seldisp = selstra = selstre = seloth = NULL;
  transtra = transtre = NULL; 
}



/**
  Destructor deallocates used memory
*/
nodeoutgm::~nodeoutgm()
{
  delete [] seldisp;
  delete [] selstra;
  delete [] selstre; 
  delete [] seloth;
  delete [] transtra;
  delete [] transtre;
}



long nodeoutgm::read(XFILE *in)
{
  long i;
  // step and loadcases
  xfscanf(in, "%k", "sel_nodstep");
  dstep.read(in);
  if (dstep.st == sel_no)
    return 0;
  xfscanf(in, "%k", "sel_nodlc");
  sellc.read(in);

  // displacements
  xfscanf(in, "%k", "displ_nodes");
  selndisp.read(in);
  switch (selndisp.st)
  {
    case sel_no:
      break;
    case sel_all:
    case sel_range:
    case sel_list:
      xfscanf(in, "%k", "displ_comp");
      seldisp = new sel[selndisp.n];
      for (i=0; i<selndisp.n; i++)
        seldisp[i].read(in);
      break;
    default:
      fprintf(stderr, "\n\nUnknown type of selection is required in function nodeoutm::read\n");
      fprintf(stderr, " (file %s, line %d)", __FILE__, __LINE__);
      return 2;
  }

  // strains
  xfscanf(in, "%k", "strain_nodes");
  selnstra.read(in);
  switch (selnstra.st)
  {
    case sel_no:
      break;
    case sel_all:
    case sel_range:
    case sel_list:
      xfscanf(in, "%k", "nodstrain_comp");
      Mp->straincomp = 1;
      Mp->strainaver = 1;
      transtra = new long[selnstra.n];
      memset(transtra, 0, sizeof(*transtra)*selnstra.n);
      selstra = new sel[selnstra.n];
      for(i=0; i<selnstra.n; i++)
        selstra[i].read(in);
      xfscanf(in, "%k", "nodstra_transfid");
      for(i=0; i<selnstra.n; i++)
      {
        if (xfscanf(in, "%ld", transtra+i) != 1)
        {
          fprintf(stderr, "\n\nError reading strain selection\n");
          fprintf(stderr, " in function nodeoutm::read, (file %s, line %d)\n", __FILE__, __LINE__);
          return 3;
        }
      }
      break;
    default:
      fprintf(stderr, "\n\nUnknown type of selection is required in function nodeoutm::read\n");
      fprintf(stderr, " (file %s, line %d)", __FILE__, __LINE__);
      return 3;
  }


  // stresses
  xfscanf(in, "%k", "stress_nodes");
  selnstre.read(in);
  switch (selnstre.st)
  {
    case sel_no:
      break;
    case sel_all:
    case sel_range:
    case sel_list:
      xfscanf(in, "%k", "nodstress_comp");
      Mp->straincomp = 1;
      Mp->strainaver = 1;
      Mp->stresscomp = 1;
      Mp->stressaver = 1;
      transtre = new long[selnstre.n];
      memset(transtre, 0, sizeof(*transtre)*selnstre.n);
      selstre = new sel[selnstre.n];
      for(i=0; i<selnstre.n; i++)
        selstre[i].read(in);
      xfscanf(in, "%k", "nodstre_transfid");
      for(i=0; i<selnstre.n; i++)
      {
        if (xfscanf(in, "%ld", transtre+i) != 1)
        {
          fprintf(stderr, "\n\nError reading stress selection\n");
          fprintf(stderr, " in function nodeoutm::read, (file %s, line %d)\n", __FILE__, __LINE__);
          return 3;
        }
      }
      break;
    default:
      fprintf(stderr, "\n\nUnknown type of selection is required in function nodeoutm::read\n");
      fprintf(stderr, " (file %s, line %d)", __FILE__, __LINE__);
      return 4;
  }

  // other values
  xfscanf(in, "%k", "other_nodes");
  selnoth.read(in);
  switch (selnoth.st)
  {
    case sel_no:
      break;
    case sel_all:
    case sel_range:
    case sel_list:
      xfscanf(in, "%k", "nodother_comp");
      if(Mp->othercomp == 0)
  Mp->othercomp = 1;
      if(Mp->otheraver == 0)
  Mp->otheraver = 1;
      seloth = new sel[selnoth.n];
      for (i=0; i < selnoth.n; i++)
        seloth[i].read(in);
      break; 
    default:
      fprintf(stderr, "\n\nUnknown type of selection is required in function nodeoutgm::read\n");
      fprintf(stderr, " (file %s, line %d)", __FILE__, __LINE__);
      return 5;
  }

  // 
  xfscanf(in, "%k", "force_nodes");
  selnforce.read(in);
  switch (selnforce.st)
  {
    case sel_no:
      break;
    case sel_all:
    case sel_range:
    case sel_list:
      xfscanf(in, "%k", "force_comp");
      selforce = new sel[selnforce.n];
      for (i=0; i<selndisp.n; i++)
        selforce[i].read(in);
      break;
    default:
      fprintf(stderr, "\n\nUnknown type of selection is required in function nodeoutm::read\n");
      fprintf(stderr, " (file %s, line %d)", __FILE__, __LINE__);
      return 2;
  }

  return 0;
}

/**
  Function prints data with description for output of nodal values to the text file.

  @param out - pointer to the opened text file
*/
void nodeoutgm::print(FILE *out)
{
  long i;

  // step and loadcases
  dstep.print(out);
  if (dstep.st == sel_no)
    return;
  sellc.print(out);
  fprintf(out, "\n");

  // displacements
  selndisp.print(out);
  switch(selndisp.st)
  {
    case sel_no:
      break;
    case sel_all:
    case sel_range:
    case sel_list:
      for(i=0; i<selndisp.n; i++)
        seldisp[i].print(out);
      break;
    default:
      fprintf(stderr, "\n\nUnknown type of selection is required in function nodeoutm::print\n");
      fprintf(stderr, " (file %s, line %d)", __FILE__, __LINE__);
      return;
  }


  // strains
  selnstra.print(out);
  switch (selnstra.st)
  {
    case sel_no:
      break;
    case sel_all:
    case sel_range:
    case sel_list:
      for(i=0; i<selnstra.n; i++)
        selstra[i].print(out);
      for(i=0; i<selnstra.n; i++)
        fprintf(out, "%ld ", transtra[i]);
      fprintf(out, "\n");
      break;
    default:
      fprintf(stderr, "\n\nUnknown type of selection is required in function nodeoutm::print\n");
      fprintf(stderr, " (file %s, line %d)", __FILE__, __LINE__);
      return;
  }

  // stresses
  selnstre.print(out);
  switch (selnstre.st)
  {
    case sel_no:
      break;
    case sel_all:
    case sel_range:
    case sel_list:
      for(i=0; i<selnstre.n; i++)
        selstre[i].print(out);
      for(i=0; i<selnstre.n; i++)
        fprintf(out, "%ld ", transtre[i]);
      fprintf(out, "\n");
      break;
    default:
      fprintf(stderr, "\n\nUnknown type of selection is required in function nodeoutm::print\n");
      fprintf(stderr, " (file %s, line %d)", __FILE__, __LINE__);
      return;
  }

  // other values
  selnoth.print(out);
  switch (selnoth.st)
  {
    case sel_no:
      break;
    case sel_all:
    case sel_range:
    case sel_list:
      for(i=0; i<selnoth.n; i++)
        seloth[i].print(out);
      break;
    default:
      fprintf(stderr, "\n\nUnknown type of selection is required in function nodeoutm::print\n");
      fprintf(stderr, " (file %s, line %d)", __FILE__, __LINE__);
      return;
  }
  selnforce.print(out);
  switch (selnforce.st)
  {
    case sel_no:
      break;
    case sel_all:
    case sel_range:
    case sel_list:
      for (i=0; i<selndisp.n; i++)
        selforce[i].print(out);
      break;
    default:
      fprintf(stderr, "\n\nUnknown type of selection is required in function nodeoutm::print\n");
      fprintf(stderr, " (file %s, line %d)", __FILE__, __LINE__);
      return;
  }
}

/**
  Function prints required output values for selected nodes and for given load case and step
  to the output grahics file.
  
  @param out  - pointer to the opened grahics file
  @param lcid - load case id
  @param desclcid - load case description
  @param gf - graphics format
  @param ifor - vector of nodal forces

 */
void nodeoutgm::print_graphics(FILE *out, long lcid, const char *desclcid, graphfmt gf, double *ifor)
{
  if (gf == grfmt_open_dx)
    return;

  if (selndisp.st != sel_no)
  {
    if (gf == grfmt_gid)    
      write_gid_displ(out, lcid, desclcid);
    else
      write_displ(out, lcid, desclcid);
  }
  print_gr_stra_scal(out, lcid, desclcid, gf);
  print_gr_stre_scal(out, lcid, desclcid, gf);
  print_gr_oth_scal(out, lcid, desclcid, gf);

  print_gr_stra_vec(out, lcid, desclcid, gf);
  print_gr_stre_vec(out, lcid, desclcid, gf);
  print_gr_oth_vec(out, lcid, desclcid, gf);

  print_gr_stra_mtx(out, lcid, desclcid, gf);
  print_gr_stre_mtx(out, lcid, desclcid, gf);
  print_gr_oth_mtx(out, lcid, desclcid, gf);

  if (selnforce.st != sel_no)
  {
    if (gf == grfmt_gid)    
      write_gid_nforces(out, lcid, desclcid, ifor);
    else
      write_nforces(out, lcid, desclcid, ifor);
  }
}

/**
  Function prints required output values for selected nodes and for given load case and step
  to the several output grahics files named by printed quantity component.
  
  @param outfn  - string with file name part
  @param mode - opening mode for graphics files
  @param lcid - load case id
  @param desclcid - load case description
  @param gf - graphics format
  @param ifor - vector of nodal forces

 */
void nodeoutgm::print_graphics(const char *outfn, const char *mode, long lcid, const char *desclcid, graphfmt gf, double *ifor)
{
  char fname[FNAMELEN+70];
  FILE *out;

  if (gf != grfmt_gid_sep)
  {
    fprintf(stderr, "\nInvalid graphics format is required in function nodeoutgm::print_graphics\n");
    fprintf(stderr, "(file %s, line %d)", __FILE__, __LINE__);
    return;
  }

  if (selndisp.st != sel_no)
  {
    sprintf(fname, "%s.displ.res", outfn);
    out = fopen(fname, mode);
    if (out == NULL)
    {
      fprintf(stderr, "\n\nUnable to open graphics file '%s' in function nodeoutgm::print_graphics\n", fname);
      fprintf(stderr, " (file %s, line %d)\n", __FILE__, __LINE__);
      abort();
    }
    fseek(out, 0, SEEK_END); // MS Visual C++ requires that
    if (ftell(out) == 0)
      fprintf(out, "GiD Post Results File 1.0\n");
    else
      write_gid_displ(out, lcid, desclcid);
    fclose(out);
  }

  print_gr_stra_scal(outfn, mode, lcid, desclcid);
  print_gr_stre_scal(outfn, mode, lcid, desclcid);
  print_gr_oth_scal(outfn, mode, lcid, desclcid);

  print_gr_stra_vec(outfn, mode, lcid, desclcid);
  print_gr_stre_vec(outfn, mode, lcid, desclcid);
  print_gr_oth_vec(outfn, mode, lcid, desclcid);

  print_gr_stra_mtx(outfn, mode, lcid, desclcid);
  print_gr_stre_mtx(outfn, mode, lcid, desclcid);
  print_gr_oth_mtx(outfn, mode, lcid, desclcid);

  if (selnforce.st != sel_no)
  {
    sprintf(fname, "%s.force.res", outfn);
    out = fopen(fname, mode);
    if (out == NULL)
    {
      fprintf(stderr, "\n\nUnable to open graphics file '%s' in function nodeoutgm::print_graphics\n", fname);
      fprintf(stderr, " (file %s, line %d)\n", __FILE__, __LINE__);
      abort();
    }
    fseek(out, 0, SEEK_END); // MS Visual C++ requires that
    if (ftell(out) == 0)
      fprintf(out, "GiD Post Results File 1.0\n");
    else
      write_gid_nforces(out, lcid, desclcid, ifor);
    fclose(out);
  }
}



/** Function prints values of selected strains as scalars for selected load case to graphics file.

    @param out  - pointer to the opened grahics file
    @param lcid - load case id
    @param desclcid - load case description
    @param gf - graphics format
*/
void nodeoutgm::print_gr_stra_scal(FILE *out, long lcid, const char *desclcid, graphfmt gf)
{
  long i, j, k;

  if (selnstra.st != sel_no)
  {
    for(i=0; i<Mm->max_ncompstrn; i++)
    {
      for (j=0; j<selnstra.n; j++)
      {
        if ((selstra[j].st == sel_mtx) || (selstra[j].st == sel_range_mtx) ||
            (selstra[j].st == sel_range_vec))
          continue;
        if (selstra[j].presence_id(i))
        {
          if (gf == grfmt_gid)    
            write_gid_nodscalar(out, strain, lcid, i, desclcid);
          else
            write_nodscalar(out, strain, lcid, i, desclcid);
          break;
        }
      }
    }
    for (j=0; j<selnstra.n; j++)
    {
      if ((transtra[j] < 0) && (selstra[j].st != sel_mtx) && 
          (selstra[j].st != sel_range_mtx) && (selstra[j].st != sel_range_vec))
      {
        for(k=0; k<3; k++)
  {
          if (gf == grfmt_gid)    
            write_gid_nodscalar(out, pstrain, lcid, k, desclcid);
          else
            write_nodscalar(out, -k-1, desclcid);
  }
        break;
      }
    }
  }
  return;
}



/** Function prints values of selected stresses as scalars for selected load case to graphics file.

    @param out  - pointer to the opened grahics file
    @param lcid - load case id
    @param desclcid - load case description
    @param gf - graphics format
*/
void nodeoutgm::print_gr_stre_scal(FILE *out, long lcid, const char *desclcid, graphfmt gf)
{
  long i, j, k;

  if (selnstre.st != sel_no)
  {
    for(i=0; i<Mm->max_ncompstrn; i++)
    {
      for (j=0; j<selnstre.n; j++)
      {
        if ((selstre[j].st == sel_mtx) || (selstre[j].st == sel_range_mtx) ||
            (selstre[j].st == sel_range_vec))
          continue;
        if (selstre[j].presence_id(i))
        {
          if (gf == grfmt_gid)    
            write_gid_nodscalar(out, stress, lcid, i, desclcid);
          else
            write_nodscalar(out, stress, lcid, i, desclcid);
          break;
        }
      }
    }
    for (j=0; j<selnstre.n; j++)
    {
      if ((transtre[j] < 0) && (selstre[j].st != sel_mtx) && (selstre[j].st != sel_range_mtx) &&
          (selstre[j].st != sel_range_vec))
      {
        for(k=0; k<4; k++)
  {
          if (gf == grfmt_gid)    
            write_gid_nodscalar(out, pstress, lcid, k, desclcid);
          else 
            write_nodscalar(out, k, desclcid);
  }
        break;
      }
    }
  }
  return;
}



/** Function prints values of selected other items as scalars for selected load case to graphics file.

    @param out  - pointer to the opened grahics file
    @param lcid - load case id
    @param desclcid - load case description
    @param gf - graphics format
*/
void nodeoutgm::print_gr_oth_scal(FILE *out, long lcid, const char *desclcid, graphfmt gf)
{
  long i, j;

  if (selnoth.st != sel_no)
  {
    for(i=0; i<Mm->max_nncompo; i++)
    {
      for (j=0; j<selnoth.n; j++)
      {
        if ((seloth[j].st == sel_mtx) || (seloth[j].st == sel_range_mtx) ||
            (seloth[j].st == sel_range_vec))
          continue;
        if (seloth[j].presence_id(i))
        {
          if (gf == grfmt_gid)    
            write_gid_nodscalar(out, other, lcid, i, desclcid);
          else
            write_nodscalar(out, other, lcid, i, desclcid);
          break;
        }
      }
    }
  }
  return;
}



/** Function prints values of selected strains as vectors for selected load case to graphics file.

    @param out  - pointer to the opened grahics file
    @param lcid - load case id
    @param desclcid - load case description
    @param gf - graphics format
*/
void nodeoutgm::print_gr_stra_vec(FILE *out, long lcid, const char *desclcid, graphfmt gf)
{
  long j, id1, n;
  if (selnstra.st != sel_no)
  {
    for (j=0; j<selnstra.n; j++)
    {
      if (selstra[j].st != sel_range_vec)
        continue;
      id1 = selstra[j].id1[0];
      n  = selstra[j].ncomp[0];
      if (id1+n-1 < Mm->max_ncompstrn)
      {
        if (gf == grfmt_gid) 
          write_gid_nodvector(out, strain, lcid, j, desclcid);
        else
  {
          fprintf(stderr, "\n\nUnsupported export format of a strain vector is required\n");
          fprintf(stderr, " in function nodeoutgm::print_gr_stra_vec(), (file %s, line %d)\n", __FILE__, __LINE__);
          abort();
  }
      }
      else
      {
        fprintf(stderr, "\n\nError - required index of strain vector exceeds limits\n");
        fprintf(stderr, " selected range is id1=%ld, n=%ld, max ncompst=%ld\n",id1, n, Mm->max_ncompstrn);
        fprintf(stderr, " Function nodeoutgm::print_gr_stra_vec(), (file %s, line %d)\n", __FILE__, __LINE__);
        abort();
      }
    }
  }
  return;
}



/** Function prints values of selected stresses as vectors for selected load case to graphics file.

    @param out  - pointer to the opened grahics file
    @param lcid - load case id
    @param desclcid - load case description
    @param gf - graphics format
*/
void nodeoutgm::print_gr_stre_vec(FILE *out, long lcid, const char *desclcid, graphfmt gf)
{
  long j, id1, n;
  if (selnstre.st != sel_no)
  {
    for (j=0; j<selnstre.n; j++)
    {
      if (selstre[j].st != sel_range_vec)
        continue;
      id1 = selstre[j].id1[0];
      n  = selstre[j].ncomp[0];
      if (id1+n-1 < Mm->max_ncompstrn)
      {
        if (gf == grfmt_gid)    
          write_gid_nodvector(out, stress, lcid, j, desclcid);
        else
  {
          fprintf(stderr, "\n\nUnsupported export format of a stress vector is required\n");
          fprintf(stderr, " in function nodeoutgm::print_gr_stre_vec(), (file %s, line %d)\n", __FILE__, __LINE__);
          abort();
  }
      }
      else
      {
        fprintf(stderr, "\n\nError - required index of stress vector exceeds limits\n");
        fprintf(stderr, " selected range is id1=%ld, n=%ld, max ncompst=%ld\n",id1, n, Mm->max_ncompstrn);
        fprintf(stderr, " Function nodeoutgm::print_gr_stre_vec(), (file %s, line %d)\n", __FILE__, __LINE__);
        abort();
      }
    }
  }
  return;
}



/** Function prints values of selected other items as vectors for selected load case to graphics file.

    @param out  - pointer to the opened grahics file
    @param lcid - load case id
    @param desclcid - load case description
    @param gf - graphics format
*/
void nodeoutgm::print_gr_oth_vec(FILE *out, long lcid, const char *desclcid, graphfmt gf)
{
  long j, id1, n;
  if (selnoth.st != sel_no)
  {
    for (j=0; j<selnoth.n; j++)
    {
      if (seloth[j].st != sel_range_vec)
        continue;
      id1 = seloth[j].id1[0];
      n  = seloth[j].ncomp[0];
      if (id1+n-1 < Mm->max_nncompo)
      {
        if (gf == grfmt_gid)    
          write_gid_nodvector(out, other, lcid, j, desclcid);
        else
  {
          fprintf(stderr, "\n\nUnsupported export format of a other values vector is required\n");
          fprintf(stderr, " in function nodeoutgm::print_gr_oth_vec(), (file %s, line %d)\n", __FILE__, __LINE__);
          abort();
  }
        break;
      }
      else
      {
        fprintf(stderr, "\n\nError - required index of other values vector exceeds limits\n");
        fprintf(stderr, " selected range is id1=%ld, n=%ld, max nncompo=%ld\n",id1, n, Mm->max_nncompo);
        fprintf(stderr, " Function nodeoutgm::print_gr_oth_vec(), (file %s, line %d)\n", __FILE__, __LINE__);
        abort();
      }
    }
  }
  return;
}



/** Function prints values of selected strains as a tensor for selected load case to graphics file.

    @param out  - pointer to the opened grahics file
    @param lcid - load case id
    @param desclcid - load case description
    @param gf - graphics format
*/
void nodeoutgm::print_gr_stra_mtx(FILE *out, long lcid, const char *desclcid, graphfmt gf)
{
  long j;
  if (selnstra.st != sel_no)
  {
    for (j=0; j<selnstra.n; j++)
    {
      if ((selstra[j].st != sel_mtx) && (selstra[j].st != sel_range_mtx))
        continue;
      if (selstra[j].st == sel_range_mtx)
      {
        fprintf(stderr, "\n\nUnsupported export of a strain tensor is required, use sel_mtx(%d)\n", sel_mtx);
        fprintf(stderr, " in function nodeoutgm::print_gr_stra_mtx(), (file %s, line %d)\n", __FILE__, __LINE__);
        abort();
      }
      if (gf == grfmt_gid)    
        write_gid_nodtensor(out, strain, lcid, j, desclcid);
      else
      {
        fprintf(stderr, "\n\nUnsupported export format of a strain tensor is required\n");
        fprintf(stderr, " in function nodeoutgm::print_gr_stra_mtx(), (file %s, line %d)\n", __FILE__, __LINE__);
        abort();
      }
    }
  }
  return;
}



/** Function prints values of selected stresses as a tensor for selected load case to graphics file.

    @param out  - pointer to the opened grahics file
    @param lcid - load case id
    @param desclcid - load case description
    @param gf - graphics format
*/
void nodeoutgm::print_gr_stre_mtx(FILE *out, long lcid, const char *desclcid, graphfmt gf)
{
  long j;
  if (selnstre.st != sel_no)
  {
    for (j=0; j<selnstre.n; j++)
    {
      if ((selstre[j].st != sel_mtx) && (selstre[j].st != sel_range_mtx))
        continue;
      if (selstre[j].st == sel_range_mtx)
      {
        fprintf(stderr, "\n\nUnsupported export of a stress tensor is required, use sel_mtx(%d)\n", sel_mtx);
        fprintf(stderr, " in function nodeoutgm::print_gr_stre_mtx(), (file %s, line %d)\n", __FILE__, __LINE__);
        abort();
      }
      if (gf == grfmt_gid)    
        write_gid_nodtensor(out, stress, lcid, j, desclcid);
      else
      {
        fprintf(stderr, "\n\nUnsupported export format of a stress tensor is required\n");
        fprintf(stderr, " in function nodeoutgm::print_gr_stre_mtx(), (file %s, line %d)\n", __FILE__, __LINE__);
        abort();
      }
    }
  }
  return;
}



/** Function prints values of selected other values as a tensor for selected load case to graphics file.

    @param out  - pointer to the opened grahics file
    @param lcid - load case id
    @param desclcid - load case description
    @param gf - graphics format
*/
void nodeoutgm::print_gr_oth_mtx(FILE *out, long lcid, const char *desclcid, graphfmt gf)
{
  long j;
  if (selnoth.st != sel_no)
  {
    for (j=0; j<selnoth.n; j++)
    {
      if ((seloth[j].st != sel_mtx) && (seloth[j].st != sel_range_mtx))
        continue;
      if (seloth[j].st == sel_mtx)
      {
        fprintf(stderr, "\n\nUnsupported export of a other values tensor is required, use sel_range_mtx(%d)\n", sel_range_mtx);
        fprintf(stderr, " in function nodeoutgm::print_gr_oth_mtx(), (file %s, line %d)\n", __FILE__, __LINE__);
        abort();
      }
      if (gf == grfmt_gid)    
        write_gid_nodtensor(out, other, lcid, j, desclcid);
      else
      {
        fprintf(stderr, "\n\nUnsupported export format of a other values tensor is required\n");
        fprintf(stderr, " in function nodeoutgm::print_gr_oth_mtx(), (file %s, line %d)\n", __FILE__, __LINE__);
        abort();
      }
    }
  }
  return;
}



//-------------------------------------------------------------------------------------------------------------------

// Functions for output to separated graphic files for nodes

//-------------------------------------------------------------------------------------------------------------------



/** Function prints values of selected strains as scalars for selected load case to the separated 
    graphics files named by printed quantity component.

  @param outfn  - string with file name part
  @param mode - opening mode for graphics files
  @param lcid - load case id
  @param desclcid - load case description

*/
void nodeoutgm::print_gr_stra_scal(const char *outfn, const char *mode, long lcid, const char *desclcid)
{
  long i, j, k;
  char fname[FNAMELEN+70];
  FILE *out;

  if (selnstra.st != sel_no)
  {
    for(i=0; i<Mm->max_ncompstrn; i++)
    {
      for (j=0; j<selnstra.n; j++)
      {
        if ((selstra[j].st == sel_mtx) || (selstra[j].st == sel_range_mtx) ||
            (selstra[j].st == sel_range_vec))
          continue;
        if (selstra[j].presence_id(i))
        {
          sprintf(fname, "%s.nodal_eps%ld.res", outfn, i+1);
          out = fopen(fname, mode);
          if (out == NULL)
          {
            fprintf(stderr, "\n\nUnable to open graphics file '%s' in function nodeoutgm::print_gr_stra_scal\n", fname);
            fprintf(stderr, " (file %s, line %d)\n", __FILE__, __LINE__);
            abort();
          }
          fseek(out, 0, SEEK_END); // MS Visual C++ requires that
          if (ftell(out) == 0)  // new file -> only print header
            fprintf(out, "GiD Post Results File 1.0\n");
          else
            write_gid_nodscalar(out, strain, lcid, i, desclcid);
          fclose(out);
          break;
        }
      }
    }
    for (j=0; j<selnstra.n; j++)
    {
      if ((transtra[j] < 0) && (selstra[j].st != sel_mtx) && 
          (selstra[j].st != sel_range_mtx) && (selstra[j].st != sel_range_vec))
      {
        for(k=0; k<3; k++)
  {
          sprintf(fname, "%s.nodal_peps%ld.res", outfn, k+1);
          out = fopen(fname, mode);
          if (out == NULL)
          {
            fprintf(stderr, "\n\nUnable to open graphics file '%s' in function nodeoutgm::print_gr_stra_scal\n", fname);
            fprintf(stderr, " (file %s, line %d)\n", __FILE__, __LINE__);
            abort();
          }
          fseek(out, 0, SEEK_END); // MS Visual C++ requires that
          if (ftell(out) == 0)  // new -> only print header
            fprintf(out, "GiD Post Results File 1.0\n");
          else
            write_gid_nodscalar(out, pstrain, lcid, k, desclcid);
          fclose(out);
  }
        break;
      }
    }
  }
  return;
}



/** Function prints values of selected stresses as scalars for selected load case to the
    separated graphics files named by printed quantity component.

  @param outfn  - string with file name part
  @param mode - opening mode for graphics files
  @param lcid - load case id
  @param desclcid - load case description
*/
void nodeoutgm::print_gr_stre_scal(const char *outfn, const char *mode, long lcid, const char *desclcid)
{
  long i, j, k;
  char fname[FNAMELEN+70];
  FILE *out;

  if (selnstre.st != sel_no)
  {
    for(i=0; i<Mm->max_ncompstrn; i++)
    {
      for (j=0; j<selnstre.n; j++)
      {
        if ((selstre[j].st == sel_mtx) || (selstre[j].st == sel_range_mtx) ||
            (selstre[j].st == sel_range_vec))
          continue;
        if (selstre[j].presence_id(i))
        {
          sprintf(fname, "%s.nodal_sig%ld.res", outfn, i+1);
          out = fopen(fname, mode);
          if (out == NULL)
          {
            fprintf(stderr, "\n\nUnable to open graphics file '%s' in function nodeoutgm::print_gr_stre_scal\n", fname);
            fprintf(stderr, " (file %s, line %d)\n", __FILE__, __LINE__);
            abort();
          }
          fseek(out, 0, SEEK_END); // MS Visual C++ requires that
          if (ftell(out) == 0)  // new file -> only print header
            fprintf(out, "GiD Post Results File 1.0\n");
          else
            write_gid_nodscalar(out, stress, lcid, i, desclcid);
          fclose(out);
          break;
        }
      }
    }
    for (j=0; j<selnstre.n; j++)
    {
      if ((transtre[j] < 0) && (selstre[j].st != sel_mtx) && (selstre[j].st != sel_range_mtx) &&
          (selstre[j].st != sel_range_vec))
      {
        for(k=0; k<4; k++)
  {
          sprintf(fname, "%s.nodal_psig%ld.res", outfn, k+1);
          out = fopen(fname, mode);
          if (out == NULL)
          {
            fprintf(stderr, "\n\nUnable to open graphics file '%s' in function nodeoutgm::print_gr_stre_scal\n", fname);
            fprintf(stderr, " (file %s, line %d)\n", __FILE__, __LINE__);
            abort();
          }
          fseek(out, 0, SEEK_END); // MS Visual C++ requires that
          if (ftell(out) == 0)  // new file -> only print header
            fprintf(out, "GiD Post Results File 1.0\n");
          else
            write_gid_nodscalar(out, pstress, lcid, k, desclcid);
          fclose(out);
  }
        break;
      }
    }
  }
  return;
}



/** Function prints values of selected other items as scalars for selected load case to separated 
    graphics files named by printed quantity component.

  @param outfn  - string with file name part
  @param mode - opening mode for graphics files
  @param lcid - load case id
  @param desclcid - load case description
*/
void nodeoutgm::print_gr_oth_scal(const char *outfn, const char *mode, long lcid, const char *desclcid)
{
  long i, j;
  char fname[FNAMELEN+70];
  FILE *out;

  if (selnoth.st != sel_no)
  {
    for(i=0; i<Mm->max_nncompo; i++)
    {
      for (j=0; j<selnoth.n; j++)
      {
        if ((seloth[j].st == sel_mtx) || (seloth[j].st == sel_range_mtx) ||
            (seloth[j].st == sel_range_vec))
          continue;
        if (seloth[j].presence_id(i))
        {
          sprintf(fname, "%s.nodal_other%ld.res", outfn, i+1);
          out = fopen(fname, mode);
          if (out == NULL)
          {
            fprintf(stderr, "\n\nUnable to open graphics file '%s' in function nodeoutgm::print_gr_oth_scal\n", fname);
            fprintf(stderr, " (file %s, line %d)\n", __FILE__, __LINE__);
            abort();
          }
          fseek(out, 0, SEEK_END); // MS Visual C++ requires that
          if (ftell(out) == 0)
            fprintf(out, "GiD Post Results File 1.0\n");
          else
            write_gid_nodscalar(out, other, lcid, i, desclcid);
          fclose(out);
          break;
        }
      }
    }
  }
  return;
}



/** Function prints values of selected strains as vectors for selected load case to separated graphics files
    named by printed quantity component.

  @param outfn  - string with file name part
  @param mode - opening mode for graphics files
  @param lcid - load case id
  @param desclcid - load case description
*/
void nodeoutgm::print_gr_stra_vec(const char *outfn, const char *mode, long lcid, const char *desclcid)
{
  long j, id1, n;
  char fname[FNAMELEN+70];
  FILE *out;

  if (selnstra.st != sel_no)
  {
    for (j=0; j<selnstra.n; j++)
    {
      if (selstra[j].st != sel_range_vec)
        continue;
      id1 = selstra[j].id1[0];
      n  = selstra[j].ncomp[0];
      if (id1+n-1 < Mm->max_ncompstrn)
      {
        sprintf(fname, "%s.nodal_eps_v%ld-%ld_s%ld.res", outfn, id1+1, n, j+1);
        out = fopen(fname, mode);
        if (out == NULL)
        {
          fprintf(stderr, "\n\nUnable to open graphics file '%s'\n", fname);
          fprintf(stderr, "  in function nodeoutgm::print_gr_stra_vec (file %s, line %d)\n", __FILE__, __LINE__);
          abort();
        }
        fseek(out, 0, SEEK_END); // MS Visual C++ requires that
        if (ftell(out) == 0)
          fprintf(out, "GiD Post Results File 1.0\n");
        else
          write_gid_nodvector(out, strain, lcid, j, desclcid);
        fclose(out);
      }
      else
      {
        fprintf(stderr, "\n\nError - required index of strain vector exceeds limits\n");
        fprintf(stderr, " selected range is id1=%ld, n=%ld, max ncompst=%ld\n",id1, n, Mm->max_ncompstrn);
        fprintf(stderr, " Function nodeoutgm::print_gr_stra_vec(), (file %s, line %d)\n", __FILE__, __LINE__);
        abort();
      }
    }
  }
  return;
}



/** Function prints values of selected stresses as vectors for selected load case to separated 
    graphics files named by printed quantity component.

  @param outfn  - string with file name part
  @param mode - opening mode for graphics files
  @param lcid - load case id
  @param desclcid - load case description
*/
void nodeoutgm::print_gr_stre_vec(const char *outfn, const char *mode, long lcid, const char *desclcid)
{
  long j, id1, n;
  char fname[FNAMELEN+70];
  FILE *out;

  if (selnstre.st != sel_no)
  {
    for (j=0; j<selnstre.n; j++)
    {
      if (selstre[j].st != sel_range_vec)
        continue;
      id1 = selstre[j].id1[0];
      n  = selstre[j].ncomp[0];
      if (id1+n-1 < Mm->max_ncompstrn)
      {
        sprintf(fname, "%s.nodal_sig_v%ld-%ld_s%ld.res", outfn, id1+1, n, j+1);
        out = fopen(fname, mode);
        if (out == NULL)
        {
          fprintf(stderr, "\n\nUnable to open graphics file '%s'\n", fname);
          fprintf(stderr, "  in function nodeoutgm::print_gr_stre_vec (file %s, line %d)\n", __FILE__, __LINE__);
          abort();
        }
        fseek(out, 0, SEEK_END); // MS Visual C++ requires that
        if (ftell(out) == 0)
          fprintf(out, "GiD Post Results File 1.0\n");
        else
          write_gid_nodvector(out, stress, lcid, j, desclcid);
        fclose(out);
      }
      else
      {
        fprintf(stderr, "\n\nError - required index of stress vector exceeds limits\n");
        fprintf(stderr, " selected range is id1=%ld, n=%ld, max ncompst=%ld\n",id1, n, Mm->max_ncompstrn);
        fprintf(stderr, " Function nodeoutgm::print_gr_stre_vec(), (file %s, line %d)\n", __FILE__, __LINE__);
        abort();
      }
    }
  }
  return;
}



/** Function prints values of selected other items as vectors for selected load case to 
    separated graphics file named by printed quantity component.

  @param outfn  - string with file name part
  @param mode - opening mode for graphics files
  @param lcid - load case id
  @param desclcid - load case description
*/
void nodeoutgm::print_gr_oth_vec(const char *outfn, const char *mode, long lcid, const char *desclcid)
{
  long j, id1, n;
  char fname[FNAMELEN+70];
  FILE *out;

  if (selnoth.st != sel_no)
  {
    for (j=0; j<selnoth.n; j++)
    {
      if (seloth[j].st != sel_range_vec)
        continue;
      id1 = seloth[j].id1[0];
      n  = seloth[j].ncomp[0];
      if (id1+n-1 < Mm->max_nncompo)
      {
        sprintf(fname, "%s.nodal_other_v%ld-%ld_s%ld.res", outfn, id1+1, n, j+1);
        out = fopen(fname, mode);
        if (out == NULL)
        {
          fprintf(stderr, "\n\nUnable to open graphics file '%s'\n", fname);
          fprintf(stderr, "  in function nodeoutgm::print_gr_oth_vec (file %s, line %d)\n", __FILE__, __LINE__);
          abort();
        }
        fseek(out, 0, SEEK_END); // MS Visual C++ requires that
        if (ftell(out) == 0)
          fprintf(out, "GiD Post Results File 1.0\n");
        else
          write_gid_nodvector(out, other, lcid, j, desclcid);
        fclose(out);
      }
      else
      {
        fprintf(stderr, "\n\nError - required index of other values vector exceeds limits\n");
        fprintf(stderr, " selected range is id1=%ld, n=%ld, max nncompo=%ld\n",id1, n, Mm->max_nncompo);
        fprintf(stderr, " Function nodeoutgm::print_gr_oth_vec(), (file %s, line %d)\n", __FILE__, __LINE__);
        abort();
      }
    }
  }
  return;
}



/** Function prints values of selected strains as a tensor for selected load case to graphics file.

  @param outfn  - string with file name part
  @param mode - opening mode for graphics files
  @param lcid - load case id
  @param desclcid - load case description
*/
void nodeoutgm::print_gr_stra_mtx(const char *outfn, const char *mode, long lcid, const char *desclcid)
{
  long j;
  char fname[FNAMELEN+70];
  FILE *out;

  if (selnstra.st != sel_no)
  {
    for (j=0; j<selnstra.n; j++)
    {
      if ((selstra[j].st != sel_mtx) && (selstra[j].st != sel_range_mtx))
        continue;
      if (selstra[j].st == sel_range_mtx)
      {
        fprintf(stderr, "\n\nUnsupported export of a strain tensor is required, use sel_mtx(%d)\n", sel_mtx);
        fprintf(stderr, " in function nodeoutgm::print_gr_stra_mtx(), (file %s, line %d)\n", __FILE__, __LINE__);
        abort();
      }
      sprintf(fname, "%s.nodal_eps_m_s%ld.res", outfn, j+1);
      out = fopen(fname, mode);
      if (out == NULL)
      {
        fprintf(stderr, "\n\nUnable to open graphics file '%s'\n", fname);
        fprintf(stderr, "  in function nodeoutgm::print_gr_stra_mtx (file %s, line %d)\n", __FILE__, __LINE__);
        abort();
      }
      fseek(out, 0, SEEK_END); // MS Visual C++ requires that
      if (ftell(out) == 0)
        fprintf(out, "GiD Post Results File 1.0\n");
      else
        write_gid_nodtensor(out, strain, lcid, j, desclcid);
      fclose(out);
    }
  }
  return;
}



/** Function prints values of selected stresses as a tensor for selected load case to graphics file.

  @param outfn  - string with file name part
  @param mode - opening mode for graphics files
  @param lcid - load case id
  @param desclcid - load case description
*/
void nodeoutgm::print_gr_stre_mtx(const char *outfn, const char *mode, long lcid, const char *desclcid)
{
  long j;
  char fname[FNAMELEN+70];
  FILE *out;

  if (selnstre.st != sel_no)
  {
    for (j=0; j<selnstre.n; j++)
    {
      if ((selstre[j].st != sel_mtx) && (selstre[j].st != sel_range_mtx))
        continue;
      if (selstre[j].st == sel_range_mtx)
      {
        fprintf(stderr, "\n\nUnsupported export of a stress tensor is required, use sel_mtx(%d)\n", sel_mtx);
        fprintf(stderr, " in function nodeoutgm::print_gr_stre_mtx(), (file %s, line %d)\n", __FILE__, __LINE__);
        abort();
      }
      sprintf(fname, "%s.nodal_sig_m_s%ld.res", outfn, j+1);
      out = fopen(fname, mode);
      if (out == NULL)
      {
        fprintf(stderr, "\n\nUnable to open graphics file '%s'\n", fname);
        fprintf(stderr, "  in function nodeoutgm::print_gr_stre_mtx (file %s, line %d)\n", __FILE__, __LINE__);
        abort();
      }
      fseek(out, 0, SEEK_END); // MS Visual C++ requires that
      if (ftell(out) == 0)
        fprintf(out, "GiD Post Results File 1.0\n");
      else
        write_gid_nodtensor(out, stress, lcid, j, desclcid);
      fclose(out);
    }
  }
  return;
}



/** Function prints values of selected other values as a tensor for selected load case to graphics file.

  @param outfn  - string with file name part
  @param mode - opening mode for graphics files
  @param lcid - load case id
  @param desclcid - load case description
*/
void nodeoutgm::print_gr_oth_mtx(const char *outfn, const char *mode, long lcid, const char *desclcid)
{
  long j;
  char fname[FNAMELEN+70];
  FILE *out;

  if (selnoth.st != sel_no)
  {
    for (j=0; j<selnoth.n; j++)
    {
      if ((seloth[j].st != sel_mtx) && (seloth[j].st != sel_range_mtx))
        continue;
      if (seloth[j].st == sel_mtx)
      {
        fprintf(stderr, "\n\nUnsupported export of an other values tensor is required, use sel_range_mtx(%d)\n", sel_range_mtx);
        fprintf(stderr, " in function nodeoutgm::print_gr_oth_mtx(), (file %s, line %d)\n", __FILE__, __LINE__);
        abort();
      }
      sprintf(fname, "%s.nodal_other_m%ld-%ld_s%ld.res", outfn, seloth[j].id1[0]+1, seloth[j].ncomp[0], j+1);
      out = fopen(fname, mode);
      if (out == NULL)
      {
        fprintf(stderr, "\n\nUnable to open graphics file '%s'\n", fname);
        fprintf(stderr, "  in function nodeoutgm::print_gr_oth_mtx (file %s, line %d)\n", __FILE__, __LINE__);
        abort();
      }
      fseek(out, 0, SEEK_END); // MS Visual C++ requires that
      if (ftell(out) == 0)
        fprintf(out, "GiD Post Results File 1.0\n");
      else
        write_gid_nodtensor(out, other, lcid,j, desclcid);
      fclose(out);
    }
  }
  return;
}



/**
  Constructor initializes data to zero values
*/
elemoutm::elemoutm()
{
  selstra = selstre = seloth = NULL;
  transtra = transtre = NULL; 
}



/**
  Destructor deallocates used memory
*/
elemoutm::~elemoutm()
{
  delete [] selstra;
  delete [] selstre;
  delete [] seloth;
  delete [] transtra;
  delete [] transtre;
}



/**
  Function reads data with description for output of element values from the text file.

  @param in - pointer to the opened text file

  @retval 0 - on success
  @retval 1 - error reading step 
  @retval 2 - error reading strain selection
  @retval 3 - error reading stress selection
  @retval 4 - error reading other values selection
 */
long elemoutm::read(XFILE *in)
{
  long i;
  // step and loadcases
  xfscanf(in, "%k", "sel_elemstep");
  dstep.read(in);
  if (dstep.st == sel_no)
    return 0;
  xfscanf(in, "%k", "sel_elemlc");
  sellc.read(in);
  if (sellc.st == sel_no)
    return 0;


  // strains
  xfscanf(in, "%k", "strain_elems");
  selestra.read(in);
  switch (selestra.st)
  {
    case sel_no:
      break;
    case sel_all:
    case sel_range:
    case sel_list:
      xfscanf(in, "%k", "elemstrain_comp");
      Mp->straincomp = 1;
      transtra = new long[selestra.n];
      memset(transtra, 0, sizeof(*transtra)*selestra.n);
      selstra = new sel[selestra.n];
      for(i=0; i<selestra.n; i++)
        selstra[i].read(in);
      xfscanf(in, "%k", "elemstra_transfid");
      for(i=0; i<selestra.n; i++)
      {
        if (xfscanf(in, "%ld", transtra+i) != 1)
        {
          fprintf(stderr, "\n\nError reading strain selection\n");
          fprintf(stderr, " in function elemoutm::read, (file %s, line %d)\n", __FILE__, __LINE__);
          return 2;
        }
      }
      break;
    default:
      fprintf(stderr, "\n\nUnknown type of selection is required in function elemoutm::read\n");
      fprintf(stderr, " (file %s, line %d)", __FILE__, __LINE__);
      return 2;
  }


  // stresses
  xfscanf(in, "%k", "stress_elems");
  selestre.read(in);
  switch (selestre.st)
  {
    case sel_no:
      break;
    case sel_all:
    case sel_range:
    case sel_list:
      xfscanf(in, "%k", "elemstress_comp");
      Mp->straincomp = 1;
      Mp->stresscomp = 1;
      transtre = new long[selestre.n];
      memset(transtre, 0, sizeof(*transtre)*selestre.n);
      selstre = new sel[selestre.n];
      for(i=0; i<selestre.n; i++)
        selstre[i].read(in);
      xfscanf(in, "%k", "elemstre_transfid");
      for(i=0; i<selestre.n; i++)
      {
        if (xfscanf(in, "%ld", transtre+i) != 1)
        {
          fprintf(stderr, "\n\nError reading stress selection\n");
          fprintf(stderr, " in function elemoutm::read, (file %s, line %d)\n", __FILE__, __LINE__);
          return 3;
        }
      }
      break;
    default:
      fprintf(stderr, "\n\nUnknown type of selection is required in function elemoutm::read\n");
      fprintf(stderr, " (file %s, line %d)", __FILE__, __LINE__);
      return 3;
  }

  // other values
  xfscanf(in, "%k", "other_elems");
  seleoth.read(in);
  switch (seleoth.st)
  {
    case sel_no:
      break;
    case sel_all:
    case sel_range:
    case sel_list:
      xfscanf(in, "%k", "elemother_comp");
      //Mp->othercomp = 1;
      seloth = new sel[seleoth.n]; 
      for (i=0; i<seleoth.n; i++)
        seloth[i].read(in);
      break;
    default:
      fprintf(stderr, "\n\nUnknown type of selection is required in function elemoutm::read\n");
      fprintf(stderr, " (file %s, line %d)", __FILE__, __LINE__);
      return 4;
  }
  return 0;
}



/**
  Function prints data with description for output of element values to the text file.

  @param out - pointer to the opened text file
*/
void elemoutm::print(FILE *out)
{
  long i;

  // step and loadcases
  dstep.print(out);
  if (dstep.st == sel_no)
    return;
  sellc.print(out);
  if (sellc.st == sel_no)
    return;

  // strains
  selestra.print(out);
  switch (selestra.st)
  {
    case sel_no:
      break;
    case sel_all:
    case sel_range:
    case sel_list:
      for(i=0; i<selestra.n; i++)
        selstra[i].print(out);
      for(i=0; i<selestra.n; i++)
        fprintf(out, "%ld ", transtra[i]);
      fprintf(out, "\n");
      break;
    default:
      fprintf(stderr, "\n\nUnknown type of selection is required in function elemoutm::print\n");
      fprintf(stderr, " (file %s, line %d)", __FILE__, __LINE__);
      return;
  }

  // stresses
  selestre.print(out);
  switch (selestre.st)
  {
    case sel_no:
      break;
    case sel_all:
    case sel_range:
    case sel_list:
      for(i=0; i<selestre.n; i++)
        selstre[i].print(out);
      for(i=0; i<selestre.n; i++)
        fprintf(out, "%ld ", transtre[i]);
      fprintf(out, "\n");
      break;
    default:
      fprintf(stderr, "\n\nUnknown type of selection is required in function elemoutm::print\n");
      fprintf(stderr, " (file %s, line %d)", __FILE__, __LINE__);
      return;
  }

  // other values
  seleoth.print(out);
  switch (seleoth.st)
  {
    case sel_no:
      break;
    case sel_all:
    case sel_range:
    case sel_list:
      for(i=0; i<seleoth.n; i++)
        seloth[i].print(out);
      break;
    default:
      fprintf(stderr, "\n\nUnknown type of selection is required in function elemoutm::print\n");
      fprintf(stderr, " (file %s, line %d)", __FILE__, __LINE__);
      return;
  }
}


/**
  Function prints required output values for selected elements and for given load case
  to the output text file.
  
  @param out  - pointer to the opened text file
  @param lcid - load case id

 */
void elemoutm::print_out(FILE *out, long lcid)
{
  if (selestra.st != sel_no)
  {
    fprintf(out, "** Element strains :\n\n");
    print_stra(out, lcid);
  }
  if (selestre.st != sel_no)
  {
    fprintf(out, "** Element stresses :\n\n");
    print_stre(out, lcid);
  }
  if (seleoth.st != sel_no)
  {
    fprintf(out, "** Element other values :\n\n");
    print_other(out);
  }
}

/**
  Function prints required strains for selected elements and for given load case
  to the output text file.
  
  @param out  - pointer to the opened text file
  @param lcid - load case id

 */
void elemoutm::print_stra(FILE *out, long lcid)
{
  long i, j, k, ncomp, id, ipp, tnipe;

  for (i=0; i<Mt->ne; i++)
  {
    if (selestra.presence_id(i))
    {
      ipp = Mt->elements[i].ipp[0][0];
      tnipe = Mt->give_totnip(i);
      fprintf(out, "   Element %7ld, integration points %ld - %ld\n  ", i+1, ipp+1, ipp+tnipe);
      ncomp = Mm->ip[ipp].ncompstr;
      id = lcid*ncomp;
      for (j=0; j<tnipe; j++)
      {
        for (k=0; k<ncomp; k++)
        {
           if (selestra.presence_id(selstra, i, k))
             fprintf(out, "   eps_%ld=% .*e", k+1, prstra, Mm->ip[ipp+j].strain[id+k]);
        }
        if (j < tnipe-1)
          fprintf(out, "\n  ");
        else
          fprintf(out, "\n");
      }
    } 
  }
  fprintf(out, "\n");
}



/**
  Function prints required stresses for selected elements and for given load case
  to the output text file.
  
  @param out  - pointer to the opened text file
  @param lcid - load case id

 */
void elemoutm::print_stre(FILE *out, long lcid)
{
  long i, j, k, ncomp, id, ipp, tnipe;

  for (i=0; i<Mt->ne; i++)
  {
    if (selestre.presence_id(i))
    {
      ipp = Mt->elements[i].ipp[0][0];
      tnipe = Mt->give_totnip(i);
      fprintf(out, "   Element %7ld, integration points %ld - %ld\n  ", i+1, ipp+1, ipp+tnipe);
      ncomp = Mm->ip[ipp].ncompstr;
      id = lcid*ncomp;
      for (j=0; j<tnipe; j++)
      {
        for (k=0; k<ncomp; k++)
        {
          if (selestre.presence_id(selstre, i, k))
            fprintf(out, "   sig_%ld=% .*e", k+1, prstre, Mm->ip[ipp+j].stress[id+k]);
        }
        if (j < tnipe-1)
          fprintf(out, "\n  ");
        else
          fprintf(out, "\n");
      }
    } 
  }
  fprintf(out, "\n");
}



/**
  Function prints required other values for selected elements and for given load case
  to the output text file.
  
  @param out  - pointer to the opened text file

 */
void elemoutm::print_other(FILE *out)
{
  long i, j, k, ncomp, ipp, tnipe;

  for (i=0; i<Mt->ne; i++)
  {
    if (seleoth.presence_id(i))
    {
      ipp = Mt->elements[i].ipp[0][0];
      tnipe = Mt->give_totnip(i);
      fprintf(out, "   Element %7ld, integration points %ld - %ld\n  ", i+1, ipp+1, ipp+tnipe);
      ncomp = Mm->ip[ipp].ncompeqother;
      for (j=0; j<tnipe; j++)
      {
        for (k=0; k<ncomp; k++)
        {
          if (seleoth.presence_id(seloth, i, k))
            fprintf(out, "   other_%ld=% .*e", k+1, proth, Mm->ip[ipp+j].eqother[k]);
        }
        if (j < tnipe-1)
          fprintf(out, "\n  ");
        else
          fprintf(out, "\n");
      }
    } 
  }
  fprintf(out, "\n");
}





/**
  Constructor initializes data to zero values
*/
elemoutgm::elemoutgm()
{
  selstra = selstre = seloth = NULL;
  transtra = transtre = NULL; 
  ide1 = 1;
}



/**
  Destructor deallocates used memory
*/
elemoutgm::~elemoutgm()
{
  delete [] selstra;
  delete [] selstre;
  delete [] seloth;
  delete [] transtra;
  delete [] transtre;
}



/**
  Function reads data with description for output of element values from the text file.

  @param in - pointer to the opened text file

  @retval 0 - on success
  @retval 1 - error reading step 
  @retval 2 - error reading strain selection
  @retval 3 - error reading stress selection
  @retval 4 - error reading other values selection
 */
long elemoutgm::read(XFILE *in)
{
  long i;
  // step and loadcases
  xfscanf(in, "%k", "sel_elemstep");
  dstep.read(in);
  if (dstep.st == sel_no)
    return 0;
  xfscanf(in, "%k", "sel_elemlc");
  sellc.read(in);
  if (sellc.st == sel_no)
    return 0;


  // strains
  xfscanf(in, "%k", "strain_elems");
  selestra.read(in);
  switch (selestra.st)
  {
    case sel_no:
      break;
    case sel_all:
    case sel_range:
    case sel_list:
      xfscanf(in, "%k", "elemstrain_comp");
      Mp->straincomp = 1;
      transtra = new long[selestra.n];
      memset(transtra, 0, sizeof(*transtra)*selestra.n);
      selstra = new sel[selestra.n];
      for(i=0; i<selestra.n; i++)
        selstra[i].read(in);
      xfscanf(in, "%k", "elemstra_transfid");
      for(i=0; i<selestra.n; i++)
      {
        if (xfscanf(in, "%ld", transtra+i) != 1)
        {
          fprintf(stderr, "\n\nError reading strain selection\n");
          fprintf(stderr, " in function elemoutgm::read, (file %s, line %d)\n", __FILE__, __LINE__);
          return 2;
        }
      }
      break;
    default:
      fprintf(stderr, "\n\nUnknown type of selection is required in function elemoutgm::read\n");
      fprintf(stderr, " (file %s, line %d)", __FILE__, __LINE__);
      return 2;
  }


  // stresses
  xfscanf(in, "%k", "stress_elems");
  selestre.read(in);
  switch (selestre.st)
  {
    case sel_no:
      break;
    case sel_all:
    case sel_range:
    case sel_list:
      xfscanf(in, "%k", "elemstress_comp");
      Mp->straincomp = 1;
      Mp->stresscomp = 1;
      transtre = new long[selestre.n];
      memset(transtre, 0, sizeof(*transtre)*selestre.n);
      selstre = new sel[selestre.n];
      for(i=0; i<selestre.n; i++)
        selstre[i].read(in);
      xfscanf(in, "%k", "elemstre_transfid");
      for(i=0; i<selestre.n; i++)
      {
        if (xfscanf(in, "%ld", transtre+i) != 1)
        {
          fprintf(stderr, "\n\nError reading stress selection\n");
          fprintf(stderr, " in function elemoutgm::read, (file %s, line %d)\n", __FILE__, __LINE__);
          return 3;
        }
      }
      break;
    default:
      fprintf(stderr, "\n\nUnknown type of selection is required in function elemoutgm::read\n");
      fprintf(stderr, " (file %s, line %d)", __FILE__, __LINE__);
      return 3;
  }

  // other values
  xfscanf(in, "%k", "other_elems");
  seleoth.read(in);
  switch (seleoth.st)
  {
    case sel_no:
      break;
    case sel_all:
    case sel_range:
    case sel_list:
      xfscanf(in, "%k", "elemother_comp");
      //Mp->othercomp = 1;
      seloth = new sel[seleoth.n]; 
      for (i=0; i<seleoth.n; i++)
        seloth[i].read(in);
      break;
    default:
      fprintf(stderr, "\n\nUnknown type of selection is required in function elemoutgm::read\n");
      fprintf(stderr, " (file %s, line %d)", __FILE__, __LINE__);
      return 4;
  }
  return 0;
}



/**
  Function prints data with description for output of element values to the text file.

  @param out - pointer to the opened text file
*/
void elemoutgm::print(FILE *out)
{
  long i;

  // step and loadcases
  dstep.print(out);
  if (dstep.st == sel_no)
    return;
  sellc.print(out);
  if (sellc.st == sel_no)
    return;

  // strains
  selestra.print(out);
  switch (selestra.st)
  {
    case sel_no:
      break;
    case sel_all:
    case sel_range:
    case sel_list:
      for(i=0; i<selestra.n; i++)
        selstra[i].print(out);
      for(i=0; i<selestra.n; i++)
        fprintf(out, "%ld ", transtra[i]);
      fprintf(out, "\n");
      break;
    default:
      fprintf(stderr, "\n\nUnknown type of selection is required in function elemoutm::print\n");
      fprintf(stderr, " (file %s, line %d)", __FILE__, __LINE__);
      return;
  }

  // stresses
  selestre.print(out);
  switch (selestre.st)
  {
    case sel_no:
      break;
    case sel_all:
    case sel_range:
    case sel_list:
      for(i=0; i<selestre.n; i++)
        selstre[i].print(out);
      for(i=0; i<selestre.n; i++)
        fprintf(out, "%ld ", transtre[i]);
      fprintf(out, "\n");
      break;
    default:
      fprintf(stderr, "\n\nUnknown type of selection is required in function elemoutm::print\n");
      fprintf(stderr, " (file %s, line %d)", __FILE__, __LINE__);
      return;
  }

  // other values
  seleoth.print(out);
  switch (seleoth.st)
  {
    case sel_no:
      break;
    case sel_all:
    case sel_range:
    case sel_list:
      for(i=0; i<seleoth.n; i++)
        seloth[i].print(out);
      break;
    default:
      fprintf(stderr, "\n\nUnknown type of selection is required in function elemoutm::print\n");
      fprintf(stderr, " (file %s, line %d)", __FILE__, __LINE__);
      return;
  }
}

/**  Function prints required output values for selected elements and for given load case and step
     to the output grahics file.
  
     @param out  - pointer to the opened grahics file
     @param lcid - load case id
     @param desclcid - load case description
     @param gf - graphics format
     @param idelem1 - number of the first element for GiD output (normally should be 1), 
                   set by print_graphics function

 */
void elemoutgm::print_graphics(FILE *out, long lcid, const char *desclcid, graphfmt gf, long idelem1)
{
  ide1 = idelem1;
  if (gf == grfmt_open_dx)
    return;
    
  print_gr_stra_scal(out, lcid, desclcid, gf);
  print_gr_stre_scal(out, lcid, desclcid, gf);
  print_gr_oth_scal(out, lcid, desclcid, gf);

  print_gr_stra_vec(out, lcid, desclcid, gf);
  print_gr_stre_vec(out, lcid, desclcid, gf);
  print_gr_oth_vec(out, lcid, desclcid, gf);

  print_gr_stra_mtx(out, lcid, desclcid, gf);
  print_gr_stre_mtx(out, lcid, desclcid, gf);
  print_gr_oth_mtx(out, lcid, desclcid, gf);

}



/** Function prints values of selected strains as scalars for selected load case to graphics file.

    @param out  - pointer to the opened grahics file
    @param lcid - load case id
    @param desclcid - load case description
    @param gf - graphics format
*/
void elemoutgm::print_gr_stra_scal(FILE *out, long lcid, const char *desclcid, graphfmt gf)
{
  long i, j;

  if (selestra.st != sel_no)
  {
    for(i=0; i<Mm->max_ncompstre; i++)
    {
      for (j=0; j<selestra.n; j++)
      {
        if ((selstra[j].st == sel_mtx) || (selstra[j].st == sel_range_mtx) ||
            (selstra[j].st == sel_range_vec))
          continue;
        if (selstra[j].presence_id(i))
        {
          if (gf == grfmt_gid)
            write_gid_elemscalar(out, strain, lcid, i, desclcid);
          if (gf == grfmt_femcad)
            write_elemscalar(out, strain, lcid, i, desclcid);
          break;
        }
      }
    }
  }
}



/** Function prints values of selected stresses as scalars for selected load case to graphics file.

    @param out  - pointer to the opened grahics file
    @param lcid - load case id
    @param desclcid - load case description
    @param gf - graphics format
*/
void elemoutgm::print_gr_stre_scal(FILE *out, long lcid, const char *desclcid, graphfmt gf)
{
  long i, j;

  if (selestre.st != sel_no)
  {
    for(i=0; i<Mm->max_ncompstre; i++)
    {
      for (j=0; j<selestre.n; j++)
      {
        if ((selstre[j].st == sel_mtx) || (selstre[j].st == sel_range_mtx) ||
            (selstre[j].st == sel_range_vec))
          continue;
        if (selstre[j].presence_id(i))
        {
          if (gf == grfmt_gid)
            write_gid_elemscalar(out, stress, lcid, i, desclcid);
          if (gf == grfmt_femcad)
            write_elemscalar(out, stress, lcid, i, desclcid);
          break;
        }
      }
    }
  }
}



/** Function prints values of selected other items as scalars for selected load case to graphics file.

    @param out  - pointer to the opened grahics file
    @param lcid - load case id
    @param desclcid - load case description
    @param gf - graphics format
*/
void elemoutgm::print_gr_oth_scal(FILE *out, long lcid, const char *desclcid, graphfmt gf)
{
  long i, j;

  if (seleoth.st != sel_no)
  {
    for(i=0; i<Mm->max_encompo; i++)
    {
      for (j=0; j<seleoth.n; j++)
      {
        if ((seloth[j].st == sel_mtx) || (seloth[j].st == sel_range_mtx) ||
            (seloth[j].st == sel_range_vec))
          continue;
        if (seloth[j].presence_id(i))
        {
          if (gf == grfmt_gid)
            write_gid_elemscalar(out, other, lcid, i, desclcid);
          if (gf == grfmt_femcad)
            write_elemscalar(out, other, lcid, i, desclcid);
          break;
        }
      }
    }
  }
}



/** Function prints values of selected strains as vectors for selected load case to graphics file.

    @param out  - pointer to the opened grahics file
    @param lcid - load case id
    @param desclcid - load case description
    @param gf - graphics format
*/
void elemoutgm::print_gr_stra_vec(FILE *out, long lcid, const char *desclcid, graphfmt gf)
{
  long j, id1, n;

  if (selestra.st != sel_no)
  {
    for (j=0; j<selestra.n; j++)
    {
      if (selstra[j].st != sel_range_vec)
        continue;
      id1 = selstra[j].id1[0];
      n  = selstra[j].ncomp[0];
      if (id1+n-1 < Mm->max_ncompstre)
      {
        if (gf == grfmt_gid)
          write_gid_elemvector(out, strain, lcid, j, desclcid);
        else
  {
          fprintf(stderr, "\n\nUnsupported export format of a strain vector is required\n");
          fprintf(stderr, " in function elemoutgm::print_gr_stra_vec(), (file %s, line %d)\n", __FILE__, __LINE__);
          abort();
  }
      }
      else
      {
        fprintf(stderr, "\n\nError - required index of strain vector exceeds limits\n");
        fprintf(stderr, " selected range is id1=%ld, n=%ld, max ncompst=%ld\n",id1, n, Mm->max_ncompstrn);
        fprintf(stderr, " Function elemoutgm::print_gr_stra_vec(), (file %s, line %d)\n", __FILE__, __LINE__);
        abort();
      }
    }
  }
}



/** Function prints values of selected stresses as vectors for selected load case to graphics file.

    @param out  - pointer to the opened grahics file
    @param lcid - load case id
    @param desclcid - load case description
    @param gf - graphics format
*/
void elemoutgm::print_gr_stre_vec(FILE *out, long lcid, const char *desclcid, graphfmt gf)
{
  long j, id1, n;

  if (selestre.st != sel_no)
  {
    for (j=0; j<selestre.n; j++)
    {
      if (selstre[j].st != sel_range_vec)
        continue;
      id1 = selstre[j].id1[0];
      n  = selstre[j].ncomp[0];
      if (id1+n-1 < Mm->max_ncompstre)
      {
        if (gf == grfmt_gid)
          write_gid_elemvector(out, stress, lcid, j, desclcid);
        else
  {
          fprintf(stderr, "\n\nUnsupported export format of a stress vector is required\n");
          fprintf(stderr, " in function elemoutgm::print_gr_stre_vec(), (file %s, line %d)\n", __FILE__, __LINE__);
          abort();
  }
      }
      else
      {
        fprintf(stderr, "\n\nError - required index of stress vector exceeds limits\n");
        fprintf(stderr, " selected range is id1=%ld, n=%ld, max ncompst=%ld\n",id1, n, Mm->max_ncompstrn);
        fprintf(stderr, " Function elemoutgm::print_gr_stre_vec(), (file %s, line %d)\n", __FILE__, __LINE__);
        abort();
      }
    }
  }
}



/** Function prints values of selected other items as vectors for selected load case to graphics file.

    @param out  - pointer to the opened grahics file
    @param lcid - load case id
    @param desclcid - load case description
    @param gf - graphics format
*/
void elemoutgm::print_gr_oth_vec(FILE *out, long lcid, const char *desclcid, graphfmt gf)
{
  long j;
  long id1, n;

  if (seleoth.st != sel_no)
  {
    for (j=0; j<seleoth.n; j++)
    {
      if (seloth[j].st != sel_range_vec)
        continue;
      id1 = seloth[j].id1[0];
      n  = seloth[j].ncomp[0];
      if (id1+n-1 < Mm->max_encompo)
      {
        if (gf == grfmt_gid)
          write_gid_elemvector(out, other, lcid, j, desclcid);
        else
  {
          fprintf(stderr, "\n\nUnsupported export format of an other values vector is required\n");
          fprintf(stderr, " in function elemoutgm::print_gr_oth_vec(), (file %s, line %d)\n", __FILE__, __LINE__);
          abort();
  }
      }
      else
      {
        fprintf(stderr, "\n\nError - required index of other values vector exceeds limits\n");
        fprintf(stderr, " selected range is id1=%ld, n=%ld, max encompo=%ld\n",id1, n, Mm->max_encompo);
        fprintf(stderr, " Function elemoutgm::print_gr_oth_vec(), (file %s, line %d)\n", __FILE__, __LINE__);
        abort();
      }
    }
  }
}



/** Function prints values of selected strains as tensors for selected load case to graphics file.

    @param out  - pointer to the opened grahics file
    @param lcid - load case id
    @param desclcid - load case description
    @param gf - graphics format
*/
void elemoutgm::print_gr_stra_mtx(FILE *out, long lcid, const char *desclcid, graphfmt gf)
{
  long j;

  if (selestra.st != sel_no)
  {
    for (j=0; j<selestra.n; j++)
    {
      if ((selstra[j].st != sel_mtx) && (selstra[j].st != sel_range_mtx))
        continue;
      if (selstra[j].st == sel_range_mtx)
      {
        fprintf(stderr, "\n\nUnsupported export of a strain tensor is required, use sel_mtx(%d)\n", sel_mtx);
        fprintf(stderr, " in function elemoutgm::print_gr_stra_mtx(), (file %s, line %d)\n", __FILE__, __LINE__);
        abort();
      }
      if (gf == grfmt_gid)
        write_gid_elemtensor(out, strain, lcid, j, desclcid);
      else
      {
        fprintf(stderr, "\n\nUnsupported export format of a strain tensor is required\n");
        fprintf(stderr, " in function elemoutgm::print_gr_stra_mtx(), (file %s, line %d)\n", __FILE__, __LINE__);
        abort();
      }
    }
  }
}



/** Function prints values of selected stresses as tensors for selected load case to graphics file.

    @param out  - pointer to the opened grahics file
    @param lcid - load case id
    @param desclcid - load case description
    @param gf - graphics format
*/
void elemoutgm::print_gr_stre_mtx(FILE *out, long lcid, const char *desclcid, graphfmt gf)
{
  long j;

  if (selestre.st != sel_no)
  {
    for (j=0; j<selestre.n; j++)
    {
      if ((selstre[j].st != sel_mtx) && (selstre[j].st != sel_range_mtx))
        continue;
      if (selstre[j].st == sel_range_mtx)
      {
        fprintf(stderr, "\n\nUnsupported export of a stress tensor is required, use sel_mtx(%d)\n", sel_mtx);
        fprintf(stderr, " in function elemoutgm::print_gr_stre_mtx(), (file %s, line %d)\n", __FILE__, __LINE__);
        abort();
      }
      if (gf == grfmt_gid)
        write_gid_elemtensor(out, stress, lcid, j, desclcid);
      else
      {
        fprintf(stderr, "\n\nUnsupported export format of a stress tensor is required\n");
        fprintf(stderr, " in function elemoutgm::print_gr_stre_mtx(), (file %s, line %d)\n", __FILE__, __LINE__);
        abort();
      }
    }
  }
}



/** Function prints values of selected other items as tensors for selected load case to graphics file.

    @param out  - pointer to the opened grahics file
    @param lcid - load case id
    @param desclcid - load case description
    @param gf - graphics format
*/
void elemoutgm::print_gr_oth_mtx(FILE *out, long lcid, const char *desclcid, graphfmt gf)
{
  long j, id1, n;

  if (seleoth.st != sel_no)
  {
    for (j=0; j<seleoth.n; j++)
    {
      if ((seloth[j].st != sel_mtx) && (seloth[j].st != sel_range_mtx))
        continue;
      if (seloth[j].st == sel_mtx)
      {
        fprintf(stderr, "\n\nUnsupported export of an other values tensor is required, use sel_range_mtx(%d)\n", sel_range_mtx);
        fprintf(stderr, " in function elemoutgm::print_gr_oth_mtx(), (file %s, line %d)\n", __FILE__, __LINE__);
        abort();
      }
      id1 = seloth[j].id1[0];
      n  = seloth[j].ncomp[0];
      if (id1+n-1 < Mm->max_encompo)
      {
        if (gf == grfmt_gid)
          write_gid_elemtensor(out, other, lcid, j, desclcid);
        else
  {
          fprintf(stderr, "\n\nUnsupported export format of an other values tensor is required\n");
          fprintf(stderr, " in function elemoutgm::print_gr_oth_mtx(), (file %s, line %d)\n", __FILE__, __LINE__);
          abort();
  }
      }
      else
      {
        fprintf(stderr, "\n\nError - required index of other values vector exceeds limits\n");
        fprintf(stderr, " selected range is id1=%ld, n=%ld, max encompo=%ld\n",id1, n, Mm->max_encompo);
        fprintf(stderr, " Function elemoutgm::print_gr_oth_vec(), (file %s, line %d)\n", __FILE__, __LINE__);
        abort();
      }
    }
  }
}



/**  Function prints required output values for selected elements and for given load case and step
     to the several output grahics files named by printed quantity component. If the opening mode is 
    "wt", than the output graphics files are only opened and header is printed, no data are written.
  
  @param outfn  - string with file name part
  @param mode - opening mode for graphics files
  @param lcid - load case id
  @param desclcid - load case description
  @param gf - graphics format
  @param idelem1 - number of the first element for GiD output (normally should be 1), 
                   set by print_graphics function

 */
void elemoutgm::print_graphics(const char *outfn, const char *mode, long lcid, const char *desclcid, graphfmt gf, long idelem1)
{
  ide1 = idelem1;
  if (gf != grfmt_gid_sep)
  {
    fprintf(stderr, "\nInvalid graphics format is required in function nodeoutm::print_graphics\n");
    fprintf(stderr, "(file %s, line %d)", __FILE__, __LINE__);
    return;
  }
  print_gr_stra_scal(outfn, mode, lcid, desclcid);
  print_gr_stre_scal(outfn, mode, lcid, desclcid);
  print_gr_oth_scal(outfn, mode, lcid, desclcid);

  print_gr_stra_vec(outfn, mode, lcid, desclcid);
  print_gr_stre_vec(outfn, mode, lcid, desclcid);
  print_gr_oth_vec(outfn, mode, lcid, desclcid);

  print_gr_stra_mtx(outfn, mode, lcid, desclcid);
  print_gr_stre_mtx(outfn, mode, lcid, desclcid);
  print_gr_oth_mtx(outfn, mode, lcid, desclcid);
  return;
}




/**  Function prints values of selected strains as scalars for selected load case to separated
     graphics files named by printed quantity component.

     @param outfn  - string with file name part
     @param mode - opening mode for graphics files
     @param lcid - load case id
     @param desclcid - load case description
*/
void elemoutgm::print_gr_stra_scal(const char *outfn, const char *mode, long lcid, const char *desclcid)
{
  long i, j;
  char fname[FNAMELEN+70];
  FILE *out;

  if (selestra.st != sel_no)
  {
    for(i=0; i<Mm->max_ncompstre; i++)
    {
      for (j=0; j<selestra.n; j++)
      {
        if ((selstra[j].st == sel_mtx) || (selstra[j].st == sel_range_mtx) ||
            (selstra[j].st == sel_range_vec))
          continue;
        if (selstra[j].presence_id(i))
        {
          sprintf(fname, "%s.elem_eps%ld.res", outfn, i+1);
          out = fopen(fname, mode);
          if (out == NULL)
          {
            fprintf(stderr, "\n\nUnable to open graphics file '%s' in function elemoutgm::print_gr_stra_scal\n", fname);
            fprintf(stderr, " (file %s, line %d)\n", __FILE__, __LINE__);
            abort();
          }
          fseek(out, 0, SEEK_END); // MS Visual C++ requires that
          if (ftell(out) == 0)
    {
            fprintf(out, "GiD Post Results File 1.0\n");
            export_gid_gauss_pt(out, ide1);
    }
          else
            write_gid_elemscalar(out, strain, lcid, i, desclcid);
          fclose(out);
          break;
        }
      }
    }
  }
}



/**  Function prints values of selected stresses as scalars for selected load case to separated
     graphics files named by printed quantity component.

     @param outfn  - string with file name part
     @param mode - opening mode for graphics files
     @param lcid - load case id
     @param desclcid - load case description

*/
void elemoutgm::print_gr_stre_scal(const char *outfn, const char *mode, long lcid, const char *desclcid)
{
  long i, j;
  char fname[FNAMELEN+70];
  FILE *out;

  if (selestre.st != sel_no)
  {
    for(i=0; i<Mm->max_ncompstre; i++)
    {
      for (j=0; j<selestre.n; j++)
      {
        if ((selstre[j].st == sel_mtx) || (selstre[j].st == sel_range_mtx) ||
            (selstre[j].st == sel_range_vec))
          continue;
        if (selstre[j].presence_id(i))
        {
          sprintf(fname, "%s.elem_sig%ld.res", outfn, i+1);
          out = fopen(fname, mode);
          if (out == NULL)
          {
            fprintf(stderr, "\n\nUnable to open graphics file '%s' in function elemeoutgm::print_gr_stre_scal\n", fname);
            fprintf(stderr, " (file %s, line %d)\n", __FILE__, __LINE__);
            abort();
          }
          fseek(out, 0, SEEK_END); // MS Visual C++ requires that
          if (ftell(out) == 0)
    {
            fprintf(out, "GiD Post Results File 1.0\n");
            export_gid_gauss_pt(out, ide1);
    }
          else
            write_gid_elemscalar(out, stress, lcid, i, desclcid);
          fclose(out);
          break;
        }
      }
    }
  }
}



/**  Function prints values of selected other items as scalars for selected load case to separated
     graphics files named by printed quantity component.

     @param outfn  - string with file name part
     @param mode - opening mode for graphics files
     @param lcid - load case id
     @param desclcid - load case description

*/
void elemoutgm::print_gr_oth_scal(const char *outfn, const char *mode, long lcid, const char *desclcid)
{
  long i, j;
  char fname[FNAMELEN+70];
  FILE *out;

  if (seleoth.st != sel_no)
  {
    for(i=0; i<Mm->max_encompo; i++)
    {
      for (j=0; j<seleoth.n; j++)
      {
        if ((seloth[j].st == sel_mtx) || (seloth[j].st == sel_range_mtx) ||
            (seloth[j].st == sel_range_vec))
          continue;
        if (seloth[j].presence_id(i))
        {
          sprintf(fname, "%s.elem_other%ld.res", outfn, i+1);
          out = fopen(fname, mode);
          if (out == NULL)
          {
            fprintf(stderr, "\n\nUnable to open graphics file '%s' in function elemoutgm::print_gr_oth_scal\n", fname);
            fprintf(stderr, " (file %s, line %d)\n", __FILE__, __LINE__);
            abort();
          }
          fseek(out, 0, SEEK_END); // MS Visual C++ requires that
          if (ftell(out) == 0)
    {
            fprintf(out, "GiD Post Results File 1.0\n");
            export_gid_gauss_pt(out, ide1);
    }
          else
            write_gid_elemscalar(out, other, lcid, i, desclcid);
          fclose(out);
          break;
        }
      }
    }
  }
}



/**  Function prints values of selected strains as vectors for selected load case to separated
     graphics files named by printed quantity component.

     @param outfn  - string with file name part
     @param mode - opening mode for graphics files
     @param lcid - load case id
     @param desclcid - load case description

*/
void elemoutgm::print_gr_stra_vec(const char *outfn, const char *mode, long lcid, const char *desclcid)
{
  long j, id1, n;
  char fname[FNAMELEN+70];
  FILE *out;

  if (selestra.st != sel_no)
  {
    for (j=0; j<selestra.n; j++)
    {
      if (selstra[j].st != sel_range_vec)
        continue;
      id1 = selstra[j].id1[0];
      n  = selstra[j].ncomp[0];
      if (id1+n-1 < Mm->max_ncompstre)
      {
        sprintf(fname, "%s.elem_eps_v%ld-%ld_s%ld.res", outfn, id1+1, n, j+1);
        out = fopen(fname, mode);
        if (out == NULL)
        {
          fprintf(stderr, "\n\nUnable to open graphics file '%s' in function elemoutgm::print_gr_stra_vec\n", fname);
          fprintf(stderr, " (file %s, line %d)\n", __FILE__, __LINE__);
          abort();
        }
        fseek(out, 0, SEEK_END); // MS Visual C++ requires that
        if (ftell(out) == 0)
        {
          fprintf(out, "GiD Post Results File 1.0\n");
          export_gid_gauss_pt(out, ide1);
        }
        else
          write_gid_elemvector(out, strain, lcid, j, desclcid);
        fclose(out);
      }
      else
      {
        fprintf(stderr, "\n\nError - required index of strain vector exceeds limits\n");
        fprintf(stderr, " selected range is id1=%ld, n=%ld, max ncompst=%ld\n",id1, n, Mm->max_ncompstre);
        fprintf(stderr, " Function elemoutgm::print_gr_stra_vec(), (file %s, line %d)\n", __FILE__, __LINE__);
        abort();
      }
    }
  }
}



/**  Function prints values of selected stresses as vectors for selected load case to separated
     graphics files named by printed quantity component.

     @param outfn  - string with file name part
     @param mode - opening mode for graphics files
     @param lcid - load case id
     @param desclcid - load case description

*/
void elemoutgm::print_gr_stre_vec(const char *outfn, const char *mode, long lcid, const char *desclcid)
{
  long j, id1, n;
  char fname[FNAMELEN+70];
  FILE *out;

  if (selestre.st != sel_no)
  {
    for (j=0; j<selestre.n; j++)
    {
      if (selstre[j].st != sel_range_vec)
        continue;
      id1 = selstre[j].id1[0];
      n  = selstre[j].ncomp[0];
      if (id1+n-1 < Mm->max_ncompstre)
      {
        sprintf(fname, "%s.elem_sig_v%ld-%ld_s%ld.res", outfn, id1+1, n, j+1);
        out = fopen(fname, mode);
        if (out == NULL)
        {
          fprintf(stderr, "\n\nUnable to open graphics file '%s' in function elemoutgm::print_gr_stre_vec\n", fname);
          fprintf(stderr, " (file %s, line %d)\n", __FILE__, __LINE__);
          abort();
        }
        fseek(out, 0, SEEK_END); // MS Visual C++ requires that
        if (ftell(out) == 0)
        {
          fprintf(out, "GiD Post Results File 1.0\n");
          export_gid_gauss_pt(out, ide1);
        }
        else
          write_gid_elemvector(out, stress, lcid, j, desclcid);
        fclose(out);
      }
      else
      {
        fprintf(stderr, "\n\nError - required index of stress vector exceeds limits\n");
        fprintf(stderr, " selected range is id1=%ld, n=%ld, max ncompst=%ld\n",id1, n, Mm->max_ncompstre);
        fprintf(stderr, " Function elemoutgm::print_gr_stre_vec(), (file %s, line %d)\n", __FILE__, __LINE__);
        abort();
      }
    }
  }
}



/**  Function prints values of selected other items as vectors for selected load case to separated
     graphics files named by printed quantity component.

     @param outfn  - string with file name part
     @param mode - opening mode for graphics files
     @param lcid - load case id
     @param desclcid - load case description

*/
void elemoutgm::print_gr_oth_vec(const char *outfn, const char *mode, long lcid, const char *desclcid)
{
  long id1, n, j;
  char fname[FNAMELEN+70];
  FILE *out;

  if (seleoth.st != sel_no)
  {
    for (j=0; j<seleoth.n; j++)
    {
      if (seloth[j].st != sel_range_vec)
        continue;
      id1 = seloth[j].id1[0];
      n  = seloth[j].ncomp[0];
      if (id1+n-1 < Mm->max_encompo)
      {
        sprintf(fname, "%s.elem_other_v%ld-%ld_s%ld.res", outfn, id1+1, n, j+1);
        out = fopen(fname, mode);
        if (out == NULL)
        {
          fprintf(stderr, "\n\nUnable to open graphics file '%s' in function elemoutgm::print_gr_oth_vec\n", fname);
          fprintf(stderr, " (file %s, line %d)\n", __FILE__, __LINE__);
          abort();
        }
        fseek(out, 0, SEEK_END); // MS Visual C++ requires that
        if (ftell(out) == 0)
        {
          fprintf(out, "GiD Post Results File 1.0\n");
          export_gid_gauss_pt(out, ide1);
        }
        else
          write_gid_elemvector(out, other, lcid, j, desclcid);
        fclose(out);
      }
      else
      {
        fprintf(stderr, "\n\nError - required index of other values vector exceeds limits\n");
        fprintf(stderr, " selected range is id1=%ld, n=%ld, max encompo=%ld\n",id1, n, Mm->max_encompo);
        fprintf(stderr, " Function elemoutgm::print_gr_oth_vec(), (file %s, line %d)\n", __FILE__, __LINE__);
        abort();
      }
    }
  }
}



/**  Function prints values of selected strains as tensors for selected load case to separated
     graphics files named by printed quantity component.

     @param outfn  - string with file name part
     @param mode - opening mode for graphics files
     @param lcid - load case id
     @param desclcid - load case description

*/
void elemoutgm::print_gr_stra_mtx(const char *outfn, const char *mode, long lcid, const char *desclcid)
{
  long j;
  char fname[FNAMELEN+70];
  FILE *out;

  if (selestra.st != sel_no)
  {
    for (j=0; j<selestra.n; j++)
    {
      if ((selstra[j].st != sel_mtx) && (selstra[j].st != sel_range_mtx))
        continue;
      if (selstra[j].st == sel_range_mtx)
      {
        fprintf(stderr, "\n\nUnsupported export of a strain tensor is required, use sel_mtx(%d)\n", sel_mtx);
        fprintf(stderr, " in function elemoutgm::print_gr_stra_mtx(), (file %s, line %d)\n", __FILE__, __LINE__);
        abort();
      }
      sprintf(fname, "%s.elem_eps_m_s%ld.res", outfn, j+1);
      out = fopen(fname, mode);
      if (out == NULL)
      {
        fprintf(stderr, "\n\nUnable to open graphics file '%s' in function elemoutgm::print_gr_stra_vec\n", fname);
        fprintf(stderr, " (file %s, line %d)\n", __FILE__, __LINE__);
        abort();
      }
      fseek(out, 0, SEEK_END); // MS Visual C++ requires that
      if (ftell(out) == 0)
      {
        fprintf(out, "GiD Post Results File 1.0\n");
        export_gid_gauss_pt(out, ide1);
      }
      else
        write_gid_elemtensor(out, strain, lcid, j, desclcid);
      fclose(out);
    }
  }
}



/**  Function prints values of selected stresses as tensors for selected load case to separated
     graphics files named by printed quantity component.

     @param outfn  - string with file name part
     @param mode - opening mode for graphics files
     @param lcid - load case id
     @param desclcid - load case description
*/
void elemoutgm::print_gr_stre_mtx(const char *outfn, const char *mode, long lcid, const char *desclcid)
{
  long j;
  char fname[FNAMELEN+70];
  FILE *out;

  if (selestre.st != sel_no)
  {
    for (j=0; j<selestre.n; j++)
    {
      if ((selstre[j].st != sel_mtx) && (selstre[j].st != sel_range_mtx))
        continue;
      if (selstre[j].st == sel_range_mtx)
      {
        fprintf(stderr, "\n\nUnsupported export of a stress tensor is required, use sel_mtx(%d)\n", sel_mtx);
        fprintf(stderr, " in function elemoutgm::print_gr_stre_mtx(), (file %s, line %d)\n", __FILE__, __LINE__);
        abort();
      }
      sprintf(fname, "%s.elem_sig_m_s%ld.res", outfn, j+1);
      out = fopen(fname, mode);
      if (out == NULL)
      {
        fprintf(stderr, "\n\nUnable to open graphics file '%s' in function elemoutgm::print_gr_stre_mtx\n", fname);
        fprintf(stderr, " (file %s, line %d)\n", __FILE__, __LINE__);
        abort();
      }
      fseek(out, 0, SEEK_END); // MS Visual C++ requires that
      if (ftell(out) == 0)
      {
        fprintf(out, "GiD Post Results File 1.0\n");
        export_gid_gauss_pt(out, ide1);
      }
      else
        write_gid_elemtensor(out, stress, lcid, j, desclcid);
      fclose(out);
    }
  }
}



/**  Function prints values of selected other items as tensors for selected load case to separated
     graphics files named by printed quantity component.

     @param outfn  - string with file name part
     @param mode - opening mode for graphics files
     @param lcid - load case id
     @param desclcid - load case description

*/
void elemoutgm::print_gr_oth_mtx(const char *outfn, const char *mode, long lcid, const char *desclcid)
{
  long j, id1, n;
  char fname[FNAMELEN+70];
  FILE *out;

  if (seleoth.st != sel_no)
  {
    for (j=0; j<seleoth.n; j++)
    {
      if ((seloth[j].st != sel_mtx) && (seloth[j].st != sel_range_mtx))
        continue;
      if (seloth[j].st == sel_mtx)
      {
        fprintf(stderr, "\n\nUnsupported export of an other values tensor is required, use sel_range_mtx(%d)\n", sel_range_mtx);
        fprintf(stderr, " in function nodeoutgm::print_gr_oth_mtx(), (file %s, line %d)\n", __FILE__, __LINE__);
        abort();
      }
      id1 = seloth[j].id1[0];
      n  = seloth[j].ncomp[0];
      if (id1+n-1 < Mm->max_encompo)
      {
        sprintf(fname, "%s.elem_other_m%ld-%ld_s%ld.res", outfn, id1+1, n, j+1);
        out = fopen(fname, mode);
        if (out == NULL)
        {
          fprintf(stderr, "\n\nUnable to open graphics file '%s' in function elemoutgm::print_gr_oth_mtx\n", fname);
          fprintf(stderr, " (file %s, line %d)\n", __FILE__, __LINE__);
          abort();
        }
        fseek(out, 0, SEEK_END); // MS Visual C++ requires that
        if (ftell(out) == 0)
        {
          fprintf(out, "GiD Post Results File 1.0\n");
          export_gid_gauss_pt(out, ide1);
        }
        else
          write_gid_elemtensor(out, other, lcid, j, desclcid);
        fclose(out);
      }
      else
      {
        fprintf(stderr, "\n\nError - required index of other values tensor exceeds limits\n");
        fprintf(stderr, " selected range is id1=%ld, n=%ld, max encompo=%ld\n",id1, n, Mm->max_encompo);
        fprintf(stderr, " Function elemoutgm::print_gr_oth_mtx(), (file %s, line %d)\n", __FILE__, __LINE__);
        abort();
      }
    }
  }
}




/**
  Constructor initializes data to zero values
*/
pointoutm::pointoutm()
{
  npnt = 0;
  ksi = eta = zeta = NULL;
  selpnt = selstra = selstre = seloth = NULL;
  transtra = transtre = NULL; 
}



/**
  Destructor deallocates used memory
*/
pointoutm::~pointoutm()
{
  delete [] ksi;
  delete [] eta;
  delete [] zeta;
  delete [] selpnt;
  delete [] selstra;
  delete [] selstre;
  delete [] seloth;
  delete [] transtra;
  delete [] transtre;
}



/**
  Function reads data with description for output of user defined point values from the text file.

  @param in - pointer to the opened text file

  @retval 0 - on success
  @retval 1 - error reading step or number of points
  @retval 2 - error reading natural coordinates
  @retval 3 - error reading sets of points on elements
  @retval 4 - error reading transformation id
 */
long pointoutm::read(XFILE *in)
{
  long i;
  // step and number of points
  xfscanf(in, "%k", "sel_pointstep");
  dstep.read(in);
  if (dstep.st == sel_no)
    return 0;
  if (xfscanf(in, "%k%ld", "numpoints", &npnt) != 2)
  {
    fprintf(stderr, "\n\nError reading number of points\n");
    fprintf(stderr, " in function pointoutm::read, (file %s, line %d)\n", __FILE__, __LINE__);
    return 1;
  }
  if (npnt == 0)
    return 0;

  ksi  = new double [npnt];
  eta  = new double [npnt];
  zeta = new double [npnt];
  memset(ksi, 0, sizeof(*ksi)*npnt);
  memset(eta, 0, sizeof(*eta)*npnt);
  memset(zeta, 0, sizeof(*zeta)*npnt);
  for (i=0; i<npnt; i++)
  {
    if (xfscanf(in, "%k %le %k %le %k %le", "ksi", ksi+i, "eta", eta+i, "zeta", zeta+i) != 6)
    {
      fprintf(stderr, "\n\nError reading coordinates of UDP\n");
      fprintf(stderr, " in function pointoutm::read, (file %s, line %d)\n", __FILE__, __LINE__);
      return 2;
    }
  }

  selelem.read(in);
  switch(selelem.st)
  {
    case sel_no:
      return 0;
    case sel_all:
    case sel_range:
    case sel_list:
      selpnt = new sel[selelem.n];
      for (i=0; i<selelem.n; i++)
        selpnt[i].read(in);
      break;
    default:
      fprintf(stderr, "\n\nUnknown type of selection is required in function pointoutm::read\n");
      fprintf(stderr, " (file %s, line %d)", __FILE__, __LINE__);
      return 3;
  }

  sellc.read(in);  
  if (sellc.st == sel_no)
    return 0;

  selstra = new sel[npnt];
  selstre = new sel[npnt];
  seloth  = new sel[npnt];

  for (i=0; i<npnt; i++)
  {
    // strains
    selstra[i].read(in);
    if (xfscanf(in, "%k%ld", "transfid", transtra+i) != 2)
    {
      fprintf(stderr, "\n\nError reading strain transformation id\n");
      fprintf(stderr, " in function pointoutm::read, (file %s, line %d)\n", __FILE__, __LINE__);
      return 4;
    }
    // stresses
    selstre[i].read(in);
    if (xfscanf(in, "%k%ld", "transfid", transtre+i) != 2)
    {
      fprintf(stderr, "\n\nError reading stress transformation id\n");
      fprintf(stderr, " in function pointoutm::read, (file %s, line %d)\n", __FILE__, __LINE__);
      return 4;
    }
    // other values
    seloth[i].read(in);

    // setup of flags
    if (selstra[i].st != sel_no)
      Mp->straincomp = 1;
    if (selstre[i].st != sel_no)
      Mp->stresscomp = 1;
    //if (seloth[i].st != sel_no)
    //Mp->othercomp = 1;
  }
  return 0;
}



/**
  Function prints data with description for output of element values to the text file.

  @param out - pointer to the opened tetx file
*/
void pointoutm::print(FILE *out)
{
  long i;
  dstep.print(out);
  if ((dstep.st == sel_no) || (npnt == 0))
  {
    fprintf(out, "\n");
    return;
  }
  fprintf(out, " %ld\n", npnt);
  if (npnt == 0)
    return;
  
  for (i=0; i<npnt; i++)
    fprintf(out, "%e %e %e\n", ksi[i], eta[i], zeta[i]);


  selelem.print(out);
  switch(selelem.st)
  {
    case sel_no:
      return;
    case sel_all:
    case sel_range:
    case sel_list:
      for (i=0; i<selelem.n; i++)
        selpnt[i].print(out);
      break;
    default:
      fprintf(stderr, "\n\nUnknown type of selection is required in function pointoutm::read\n");
      fprintf(stderr, " (file %s, line %d)", __FILE__, __LINE__);
      return;
  }

  sellc.print(out);  
  if (sellc.st == sel_no)
    return;

  for (i=0; i<npnt; i++)
  {
    // strains
    selstra[i].print(out);
    fprintf(out, "%ld\n", transtra[i]);
    // stresses
    selstre[i].print(out);
    fprintf(out, "%ld\n", transtre[i]);
    // other values
    seloth[i].print(out);
  }
}



/**
  Function prints required output values for user defined point and for given load case
  to the output text file.
  
  @param out  - pointer to the opened text file
  @param lcid - load case id  

 */
void pointoutm::print_out(FILE *out, long lcid)
{
  long i;

  for (i=0; i<npnt; i++)
  {
    if (selstra[i].st != sel_no)
      fprintf(out, "** UDP strains :\n\n");
  }
  print_stra(out, lcid);

  for (i=0; i<npnt; i++)
  {
    if (selstre[i].st != sel_no)
      fprintf(out, "** UDP stresses :\n\n");
  }
  print_stre(out, lcid);

  for (i=0; i<npnt; i++)
  {
    if (seloth[i].st != sel_no)
      fprintf(out, "** UDP other values :\n\n");
    print_other(out, i);
  }
}



/**
  Function prints required strains for user defined point and for given load case
  to the output text file.
  
  @param out  - pointer to the opened text file
  @param lcid - load case id

 */
void pointoutm::print_stra(FILE *out, long lcid)
{
  long i,j,k,l,n, ncomp, id;

  for (i=0; i<Mt->ne; i++)
  {
    if (selelem.presence_id(i))
    {
      fprintf(out, "   Element %7ld:\n",i+1);
      ncomp = Mt->give_tncomp(i);
      id = lcid*ncomp;
      switch(selelem.st)
      {
        case sel_no:
          n=0;
          break;
        case sel_all:
        case sel_range:
        case sel_list:
          n=selelem.n;
          break;
        default:
          fprintf(stderr, "\n\nUnknown type of selection is required in function pointoutm::print_stra\n");
          fprintf(stderr, " (file %s, line %d)", __FILE__, __LINE__);
          return;
      }
      for (j=0; j<n; j++)
      {
        for (k=0; k<npnt; k++)
        {
          if (selpnt[j].presence_id(k))
          {
            fprintf(out, "    point %ld:", k+1);
            for (l=0; l<ncomp; l++)
            {
              if(selstra[k].presence_id(l))
                fprintf(out, "   eps_%ld=% *e", l+1, prstra, Mm->stra.ev[i][k][id+l]);
            }
            fprintf(out, "\n"); 
          }
        }
      }
      fprintf(out, "\n");
    }
  } 
}



/**
  Function prints required stresses for user defined point and for given load case
  to the output text file.
  
  @param out  - pointer to the opened text file
  @param lcid - load case id
  @param pid  - user defined point id

 */
void pointoutm::print_stre(FILE *out, long lcid)
{
  long i,j,k,l,n, ncomp, id;

  for (i=0; i<Mt->ne; i++)
  {
    if (selelem.presence_id(i))
    {
      fprintf(out, "   Element %7ld:\n", i+1);
      ncomp = Mt->give_tncomp(i);
      id = lcid*ncomp;
      switch(selelem.st)
      {
        case sel_no:
          n=0;
          break;
        case sel_all:
        case sel_range:
        case sel_list:
          n=selelem.n;
          break;
        default:
          fprintf(stderr, "\n\nUnknown type of selection is required in function pointoutm::print_stre\n");
          fprintf(stderr, " (file %s, line %d)", __FILE__, __LINE__);
          return;
      }
      for (j=0; j<n; j++)
      {
        for (k=0; k<npnt; k++)
        {
          if (selpnt[j].presence_id(k))
          {
            fprintf(out, "    point %ld:", k+1);
            for (l=0; l<ncomp; l++)
            {
              if(selstra[k].presence_id(l))
                fprintf(out, "   sig_%ld=% *e", l+1, prstre, Mm->stre.ev[i][k][id+l]);
            }
            fprintf(out, "\n"); 
          }
        }
      }
      fprintf(out, "\n");
    }
  } 
}





/**
  Function prints required other values for user defined point and for given load case
  to the output text file.
  
  @param out  - pointer to the opened text file
  @param pid  - user defined point id

 */
void pointoutm::print_other(FILE *out, long pid)
{
  long i,j,k,l,n, ncomp;

  for (i=0; i<Mt->ne; i++)
  {
    if (selelem.presence_id(i))
    {
//      fprintf(out, "   Element %7ld:\n"i+1);
      ncomp = Mm->ip[Mt->elements[i].ipp[0][0]].ncompother;
      switch(selelem.st)
      {
        case sel_no:
          n=0;
          break;
        case sel_all:
        case sel_range:
        case sel_list:
          n=selelem.n;
          break;
        default:
          fprintf(stderr, "\n\nUnknown type of selection is required in function pointoutm::print_other\n");
          fprintf(stderr, " (file %s, line %d)", __FILE__, __LINE__);
          return;
      }
      for (j=0; j<n; j++)
      {
        for (k=0; k<npnt; k++)
        {
          if (selpnt[j].presence_id(k))
          {
//            fprintf(out, "    point %ld:", k+1);
            for (l=0; l<ncomp; l++)
            {
//              if(seloth[k].presence_id(l))
//                fprintf(out, "   other_%ld=% *e", l+1, proth, Mm->other.ev[i][k][l]);
            }
//            fprintf(out, "\n"); 
          }
        }
      }
//      fprintf(out, "\n");
    }
  } 
}

/**
  The function prints header, points and used elements to the VTK file

  Parameters :
  @param fname - pointer to the opened VTK file
  @param lcid  - load case id
  @param istep - integration step
  @param lambda - actual time in either s,h,days
*/
void outdriverm::print_default_vtk(FILE *fname, long lcid, long istep, double lambda){
  long int i,j,nn;
  int tr[20];
  fprintf(fname, "# vtk DataFile Version 1.0\n");
  fprintf(fname, "Time step %ld time %lf\n",istep,lambda);
  fprintf(fname, "ASCII\n\n");
  fprintf(fname, "DATASET UNSTRUCTURED_GRID\n");
  fprintf(fname, "POINTS %ld float\n", Mt->nn);
  for (i=0; i<Mt->nn; i++){
    fprintf(fname, "%e %e %e\n", Gtm->gnodes[i].x, Gtm->gnodes[i].y, Gtm->gnodes[i].z);
  }

  long int ncells = 0;
  for (i=0; i < Mt->ne; i++){
    ncells += Mt->give_nne (i);
  }
  fprintf(fname, "\nCELLS %ld %ld\n", Mt->ne, ncells+Mt->ne);
  for (i=0; i < Mt->ne; i++){
    fprintf(fname, "%ld ", Mt->give_nne(i));//number of nodes on element
    for(j=0;j<Mt->give_nne(i);j++){
      switch(Mt->elements[i].te){//for nodes renumbering
        case linearhex:
          tr[0] = 6; tr[1] = 7; tr[2] = 4; tr[3] = 5;//xy, z=0, xi=x,nu=y, theta=z
          tr[4] = 2; tr[5] = 3; tr[6] = 0; tr[7] = 1;//xy, z<>0
          nn=tr[j];
        case quadrhex:
          tr[0] = 6; tr[1] = 7; tr[2] = 4; tr[3] = 5;//xy, z=0, xi=x,nu=y, theta=z
          tr[4] = 2; tr[5] = 3; tr[6] = 0; tr[7] = 1;
          tr[8] = 18; tr[9] = 19; tr[10] = 16; tr[11] = 17;
          tr[12] = 10; tr[13] = 11; tr[14] = 8; tr[15] = 9;
          tr[16] = 14; tr[17] = 15; tr[18] = 12; tr[19] = 13;
          nn=tr[j];
          break;
        default:
          nn=j;
          break;
      }
      fprintf(fname, "%ld ", Gtm->gelements[i].nodes[nn]);
    }
    fprintf(fname, "\n");
  }
  
  //output cell types
  fprintf(fname, "\nCELL_TYPES %ld\n", Mt->ne);
  for (i=0; i < Mt->ne; i++){
    switch(Mt->elements[i].te){
      case bar2d:
      case beam2d:
      case bar3d:
      case beam3d:
        fprintf(fname, "3\n");//VTK_LINE
        break;
      case barq2d :
      case barq3d :
        fprintf(fname, "21\n");//VTK_QUADRATIC_EDGE
        break;
      case planeelementlt:
      case axisymmlt:
        fprintf(fname, "5\n");//VTK_TRIANGLE
        break;
      case planeelementlq:
      case axisymmlq:
        fprintf(fname, "9\n");//VTK_QUAD
        break;
      case planeelementqq:
      case axisymmqq:
        fprintf(fname, "23\n");//VTK_QUADRATIC_QUAD
        break;
      case lineartet:
        fprintf(fname, "10\n");//VTK_TETRA
        break;
      case linearhex:
        fprintf(fname, "12\n");//VTK_HEXAHEDRON - Renumbered nodes
        break;
      case quadrhex:
        fprintf(fname, "25\n");//VTK_QUADRATIC_HEXAHEDRON - Renumbered nodes
        break;
      default:
        fprintf(stderr, "\n\n Unknown element type in VTK export, file %s, line %d\n", __FILE__, __LINE__);
        abort();
        break;
    }
  }
}

/**
  The function prints data to the VTK file

  Parameters :
  @param fname - pointer to the opened VTK file
  @param lcid  - load case id
  @param desclcid - load case description
*/

//VTK can not handle integration points directly, results in nodes only
void outdriverm::write_vtk_unkn(FILE *fname, long lcid, const char *desclcid){
  double *r;
  long i, j, k, ncomp;
  fprintf(fname, "\nPOINT_DATA %ld\n", Mt->nn);
  //displacements at nodes
  if (nog.selndisp.st != sel_no){
    fprintf(fname, "VECTORS Nodal_displacements float\n");
    r = new double[3];
    for (i=0; i<Mt->nn; i++){
      r[0]=r[1]=r[2]=0;
      noddispl(lcid, r, i);
      fprintf(fname, "%e %e %e\n", r[0], r[1], r[2]);
    }
    delete [] r;
    fprintf(fname, "\n");
  }
    // strains at nodes
  if (nog.selnstra.st != sel_no){
    fprintf(fname, "TENSORS Strain float\n");
    for (i=0; i<Mt->nn; i++){
      ncomp = Mt->nodes[i].ncompstr;
      vector eps(4);
      matrix epst(3,3);
      fillv(0.0, eps);
      fillm(0.0, epst);
      for (j=0; j<ncomp; j++)
        eps[j] = Mt->nodes[i].strain[j];
      switch (ncomp){
        case 1:{
          vector_tensor (eps, epst, bar, strain);
          break;
        }
        case 4:{
          vector_tensor (eps, epst, planestrain, strain);//or planestress
          break;
        }
        case 6:{
          vector_tensor (eps, epst, spacestress, strain);
        }
      }
      for (k=0; k<3; k++){
        fprintf(fname, "%e %e %e\n", epst[k][0], epst[k][1], epst[k][2]);
      }
      fprintf(fname, "\n");
    }
  }
  // stresses at nodes
  if (nog.selnstre.st != sel_no){
    fprintf(fname, "TENSORS Stress float\n");
    for (i=0; i<Mt->nn; i++){
      ncomp = Mt->nodes[i].ncompstr;
      vector sig(4);
      matrix sigt(3,3);
      fillv(0.0, sig);
      fillm(0.0, sigt);
      for (j=0; j<ncomp; j++)
        sig[j] = Mt->nodes[i].stress[j];
      switch (ncomp){
        case 1:{
          vector_tensor (sig, sigt, bar, stress);
          break;
        }
        case 4:{
          vector_tensor (sig, sigt, planestrain, stress);//or planestress
          break;
        }
        case 6:{
          vector_tensor (sig, sigt, spacestress, stress);
        }
      }
      for (k=0; k<3; k++){
        fprintf(fname, "%e %e %e\n", sigt[k][0], sigt[k][1], sigt[k][2]);
      }
      fprintf(fname, "\n");
    }
  }
  // other values at nodes
  if (nog.selnoth.st != sel_no){
/*    fprintf(fname, "TENSORS Other float\n");
    for (i=0; i<Mt->nn; i++){
      ncomp = Mt->nodes[i].ncompother;
      vector oth(4);
      matrix otht(3,3);
      fillv(0.0, oth);
      fillm(0.0, otht);
      for (j=0; j<ncomp; j++)
        oth[j] = Mt->nodes[i].other[j];
      switch (ncomp){
        case 1:{
          vector_tensor (oth, otht, bar, stress);
          break;
        }
        case 4:{
          vector_tensor (oth, otht, planestrain, stress);//or planestress
          break;
        }
        case 6:{
          vector_tensor (oth, otht, spacestress, stress);
        }
      }
      for (k=0; k<3; k++){
        fprintf(fname, "%e %e %e\n", otht[k][0], otht[k][1], otht[k][2]);
      }
      fprintf(fname, "\n");
    }*/
  }
  // reactions - not implemented
}

